@@ -1,5 +1,5 @@
 /*
-Copyright (C) 2005, 2006, 2007 John E. Davis
+Copyright (C) 2005 John E. Davis
 
 This file is part of the S-Lang Library.
 
@@ -32,10 +32,9 @@ USA.
 # include <unistd.h>
 #endif
 #include <string.h>
-#include <signal.h>
 #include <slang.h>
 
-static char *Slsh_Version = "0.8.1-0";
+static char *Slsh_Version = "0.7.3-0";
 #define SLSHRC_FILE "slsh.rc"
 #include "slsh.h"
 
@@ -198,8 +197,6 @@ static void c_exit (int *code)
 	SLfree ((char *) AtExit_Hooks);
 	AtExit_Hooks = next;
      }
-   if (SLang_get_error ())
-     SLang_restart (1);
    exit (*code);
 }
 
@@ -406,6 +403,7 @@ static SLang_Intrin_Fun_Type Intrinsics [] =
 {
    MAKE_INTRINSIC_I("exit", c_exit, VOID_TYPE),
    MAKE_INTRINSIC_1("atexit", at_exit, VOID_TYPE, SLANG_REF_TYPE),
+   MAKE_INTRINSIC_S("slsh_readline", slsh_readline_intrinsic, VOID_TYPE),
    MAKE_INTRINSIC_0("stat_mode_to_string", stat_mode_to_string, VOID_TYPE),
    SLANG_END_INTRIN_FUN_TABLE
 };
@@ -417,21 +415,18 @@ static void usage (void)
 Usage: slsh [OPTIONS] [-|file [args...]]\n\
  --help           Print this help\n\
  --version        Show slsh version information\n\
- -e string        Execute 'string' as S-Lang code\n\
  -g               Compile with debugging code, tracebacks, etc\n\
  -n               Don't load personal init file\n\
- --init file      Use this file instead of ~/%s\n",
-	    USER_SLSHRC);
-   fprintf (stderr, "\
+ --init file      Use this file instead of ~/%s\n\
  --no-readline    Do not use readline\n\
  -i               Force interactive input\n\
- -t               Test mode.  If slsh_main exists, do not call it\n\
+ -t               Test mode.  If slsh_main exits, do not call it\n\
  -v               Show verbose loading messages\n\
- -Dname           Define \"name\" as a preprocessor symbol\n\
 \n\
   Note: - and -i are mutually exclusive\n\
 \n\
-"
+",
+	    USER_SLSHRC
 	    );
    libpath = SLpath_get_load_path ();
    fprintf (stderr, "Default search path: %s\n", (libpath == NULL) ? "" : libpath);
@@ -440,30 +435,16 @@ Usage: slsh [OPTIONS] [-|file [args...]]\n\
    exit (1);
 }
 
-static void output_version (void)
+static void version (void)
 {
-   fprintf (stdout, "slsh version %s; ", Slsh_Version);
-   fprintf (stdout, "S-Lang version: %s\n", SLang_Version_String);
+   fprintf (stdout, "slsh version %s\n", Slsh_Version);
+   fprintf (stdout, "S-Lang Library Version: %s\n", SLang_Version_String);
    if (SLANG_VERSION != SLang_Version)
      {
 	fprintf (stdout, "\t** Note: This program was compiled against version %s.\n",
 		 SLANG_VERSION_STRING);
      }
-}
-
-static int output_copyright (void)
-{
-   output_version ();
-   fprintf (stdout, "Copyright (C) 2005-2007 John E. Davis <jed@jedsoft.org>\r\n");
-   fprintf (stdout, "This is free software with ABSOLUTELY NO WARRANTY.\r\n");
-   fprintf (stdout, "\n");
    
-   return 0;
-}
-
-static void version (void)
-{
-   output_version ();
    exit (0);
 }
 
@@ -476,24 +457,25 @@ int main (int argc, char **argv)
    int is_interactive = 0;
    int use_readline = 1;
    int test_mode = 0;
-   char *exec_string = NULL;
+
+   if (SLang_Version < SLANG_VERSION)
+     {
+	fprintf (stderr, "***Warning: Executable compiled against S-Lang %s but linked to %s\n",
+		 SLANG_VERSION_STRING, SLang_Version_String);
+	fflush (stderr);
+     }
 
    (void) SLutf8_enable (-1);
 
    if ((-1 == SLang_init_all ())
        || (-1 == SLang_init_array_extra ())
        || (-1 == SLang_init_import ()) /* dynamic linking */
-       || (-1 == SLadd_intrin_fun_table (Intrinsics, NULL))
-       || (-1 == slsh_init_readline_intrinsics ()))
+       || (-1 == SLadd_intrin_fun_table (Intrinsics, NULL)))
      {
 	fprintf(stderr, "Unable to initialize S-Lang.\n");
 	return 1;
      }
 
-#ifdef SIGPIPE
-   (void) SLsignal (SIGPIPE, SIG_IGN);
-#endif
-
    /* FIXME for other systems */
 #ifdef __WIN32__
    init_file_dir = getenv ("USERPROFILE");
@@ -534,19 +516,10 @@ int main (int argc, char **argv)
 	     continue;
 	  }
 
-	if ((0 == strcmp (argv[1], "-e"))
-	    && (argc > 2))
-	  {
-	     argc -= 2;
-	     argv += 2;
-	     exec_string = *argv;
-	     continue;
-	  }
-
 	if (0 == strcmp (argv[1], "-g"))
 	  {
 	     SLang_generate_debug_info (1);
-	     SLang_Traceback = SL_TB_FULL;
+	     SLang_Traceback = 1;
 	     argc--;
 	     argv++;
 	     continue;
@@ -594,23 +567,12 @@ int main (int argc, char **argv)
 	     argv += 2;
 	     continue;
 	  }
-	
-	if (0 == strncmp (argv[1], "-D", 2))
-	  {
-	     char *prep = argv[1] + 2;
-	     if (*prep != 0)
-	       (void) SLdefine_for_ifdef (prep);
-	     argc--;
-	     argv++;
-	  }
-
 	break;
      }
 
    if (argc == 1)
      {
-	if ((exec_string == NULL) && (is_interactive == 0))
-	  is_interactive = (isatty (fileno(stdin)) && isatty (fileno(stdout)));
+	is_interactive = (isatty (fileno(stdin)) && isatty (fileno(stdout)));
 	file = NULL;
      }
    else
@@ -626,14 +588,8 @@ int main (int argc, char **argv)
 	argc--;
 	argv++;
      }
-
-   if ((is_interactive == 0)
-       && (SLang_Version < SLANG_VERSION))
-     {
-	fprintf (stderr, "***Warning: Executable compiled against S-Lang %s but linked to %s\n",
-		 SLANG_VERSION_STRING, SLang_Version_String);
-	fflush (stderr);
-     }
+   
+   (void) slsh_use_readline (use_readline);
 
    /* fprintf (stdout, "slsh: argv[0]=%s\n", argv[0]); */
    if (-1 == SLang_set_argc_argv (argc, argv))
@@ -645,18 +601,12 @@ int main (int argc, char **argv)
    if (-1 == load_startup_file (is_interactive))
      return SLang_get_error ();
 
-   /* Initializing the readline interface causes the .slrlinerc file
-    * to be loaded.  It is put here after the startup files have been loaded.
-    */
-   if (-1 == slsh_use_readline (SLpath_basename (argv[0]), use_readline, is_interactive))
-     return 1;
-
    if ((init_file != NULL)
        && (-1 == try_to_load_file (init_file_dir, init_file, NULL)))
      return SLang_get_error ();
 
-   if ((file != NULL)
-       || ((is_interactive == 0) && (exec_string == NULL)))
+   /* Now load an initialization file and exit */
+   if ((file != NULL) || (is_interactive == 0))
      {
 	if (0 == try_to_load_file (NULL, file, NULL))
 	  {
@@ -667,16 +617,8 @@ int main (int argc, char **argv)
 	  (void) SLang_run_hooks ("slsh_main", 0);
      }
 
-   if (exec_string != NULL)
-     {
-	(void) SLang_load_string (exec_string);
-     }
-
    if (is_interactive)
-     {
-	output_copyright ();
-	(void) slsh_interactive ();
-     }
+     (void) slsh_interactive ();
 
    exit_val = SLang_get_error ();
    c_exit (&exit_val);
