@@ -20,7 +20,9 @@
 
 #define SLANG_USE_INLINE_CODE		1
 
-/* Add extra information for tracking down errors. */
+/* This is experimental.  It adds extra information for tracking down
+ * errors.
+ */
 #define SLANG_HAS_DEBUG_CODE		1
 
 /* Experimental: Support for examining call frames */
@@ -62,10 +64,7 @@
  */
 #define SLTT_XTERM_ALWAYS_BCE		0
   
-/*
- * NOTE: Kanji is not supported by slang-2.  Consider using a UTF-8 aware 
- * terminal.  Setting this variable will have no effect.
- */
+/* Set this to 1 to enable Kanji support.  See above comment. */
 #define SLANG_HAS_KANJI_SUPPORT		0
 
 #define SLANG_HAS_SIGNALS		1
@@ -75,6 +74,3 @@
  * SLANG_HAS_DEBUG_CODE must be enabled for this to work.
  */
 #define SLANG_HAS_BOSEOS	SLANG_HAS_DEBUG_CODE
-
-/* Define this if you want the language to support qualifiers */
-#define SLANG_HAS_QUALIFIERS	1
