@@ -2,7 +2,7 @@
 #define DAVIS_SLANG_H_
 /* -*- mode: C; mode: fold; -*- */
 /*
-Copyright (C) 2004-2007 John E. Davis
+Copyright (C) 2004, 2005 John E. Davis
 
 This file is part of the S-Lang Library.
 
@@ -22,8 +22,8 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
 USA.  
 */
 
-#define SLANG_VERSION 20100
-#define SLANG_VERSION_STRING "2.1.0-0"
+#define SLANG_VERSION 20004
+#define SLANG_VERSION_STRING "2.0.4"
 /* #ifdef __DATE__ */
 /* # define SLANG_VERSION_STRING SLANG_VERSION_STRING0 " " __DATE__ */
 /* #else */
@@ -279,13 +279,9 @@ SL_EXTERN void SLstring_dump_stats (void);
 #if SLANG_SIZEOF_INT == 4
 typedef unsigned int SLwchar_Type;
 # define SLANG_WCHAR_TYPE SLANG_UINT_TYPE
-# define SLang_push_wchar SLang_push_uint
-# define SLang_pop_wchar SLang_pop_uint
 #else
 typedef unsigned long SLwchar_Type;
 # define SLANG_WCHAR_TYPE SLANG_ULONG_TYPE
-# define SLang_push_wchar SLang_push_ulong
-# define SLang_pop_wchar SLang_pop_ulong
 #endif
 typedef unsigned char SLuchar_Type;
 
@@ -322,14 +318,13 @@ SL_EXTERN int SLwchar_isspace (SLwchar_Type);
 SL_EXTERN int SLwchar_isupper (SLwchar_Type);
 SL_EXTERN int SLwchar_isxdigit (SLwchar_Type);
 
-#define SLWCWIDTH_SINGLE_WIDTH	0x0001
-SL_EXTERN int SLwchar_set_wcwidth_flags (int flags);
-
-/* The skip_char functions also skip over malformed strings counting each byte
- * in the malformed sequence.
- */
 SL_EXTERN SLuchar_Type *SLutf8_skip_char (SLuchar_Type *u, SLuchar_Type *umax);
 SL_EXTERN SLuchar_Type *SLutf8_bskip_char (SLuchar_Type *umin, SLuchar_Type *u);
+
+/* The SLutf8_strup/lo functions return slstrings -- free with SLang_free_slstring */
+SL_EXTERN SLuchar_Type *SLutf8_strup (SLuchar_Type *u, SLuchar_Type *umax);
+SL_EXTERN SLuchar_Type *SLutf8_strlo (SLuchar_Type *u, SLuchar_Type *umax);
+
 SL_EXTERN SLuchar_Type *SLutf8_skip_chars (SLuchar_Type *u, SLuchar_Type *umax,
 					unsigned int num, unsigned int *dnum,
 					int ignore_combining );
@@ -337,10 +332,6 @@ SL_EXTERN SLuchar_Type *SLutf8_bskip_chars (SLuchar_Type *umin, SLuchar_Type *u,
 					  unsigned int num, unsigned int *dnum,
 					 int ignore_combining);
 
-/* The SLutf8_strup/lo functions return slstrings -- free with SLang_free_slstring */
-SL_EXTERN SLuchar_Type *SLutf8_strup (SLuchar_Type *u, SLuchar_Type *umax);
-SL_EXTERN SLuchar_Type *SLutf8_strlo (SLuchar_Type *u, SLuchar_Type *umax);
-
 SL_EXTERN SLstr_Type *SLutf8_subst_wchar (SLuchar_Type *u, SLuchar_Type *umax,
 				       SLwchar_Type wch, unsigned int pos,
 				       int ignore_combining);
@@ -560,7 +551,7 @@ SL_EXTERN int SLadd_math_unary_table (SLang_Math_Unary_Type *, char *);
 SL_EXTERN int SLadd_iconstant_table (SLang_IConstant_Type *, char *);
 SL_EXTERN int SLadd_dconstant_table (SLang_DConstant_Type *, char *);
 SL_EXTERN int SLadd_fconstant_table (SLang_FConstant_Type *, char *);
-#ifdef HAVE_LONG_LONG
+#if HAVE_LONG_LONG
 SL_EXTERN int SLadd_llconstant_table (SLang_LLConstant_Type *, char *);
 #endif
 SL_EXTERN int SLadd_istruct_table (SLang_IStruct_Field_Type *, VOID_STAR, char *);
@@ -596,7 +587,6 @@ SL_EXTERN void SLns_delete_namespace (SLang_NameSpace_Type *);
 SL_EXTERN int SLns_load_file (char *, char *);
 SL_EXTERN int SLns_load_string (char *, char *);
 SL_EXTERN int (*SLns_Load_File_Hook) (char *, char *);
-
 SL_EXTERN int SLang_load_file_verbose (int);    
 /* if non-zero, display file loading messages */
 
@@ -721,9 +711,8 @@ SL_EXTERN int SLclass_set_acopy_function (SLang_Class_Type *, int (*)(SLtype, VO
 SL_EXTERN int SLclass_set_deref_function (SLang_Class_Type *, int (*)(SLtype, VOID_STAR));
 SL_EXTERN int SLclass_set_eqs_function (SLang_Class_Type *, int (*)(SLtype, VOID_STAR, SLtype, VOID_STAR));
 
-SL_EXTERN int SLclass_set_length_function (SLang_Class_Type *, int(*)(SLtype, VOID_STAR, unsigned int *));
-
 SL_EXTERN int SLclass_set_is_container (SLang_Class_Type *, int);
+
 SL_EXTERN int SLclass_set_foreach_functions (
   SLang_Class_Type *,
   SLang_Foreach_Context_Type *(*)(SLtype, unsigned int),   /* open method */
@@ -864,10 +853,10 @@ typedef struct _pSLang_Array_Type
 }
 SLang_Array_Type;
 
-SL_EXTERN int SLang_pop_array_of_type (SLang_Array_Type **atp, SLtype type);
-SL_EXTERN int SLang_pop_array (SLang_Array_Type **atp, int convert_scalar);
-SL_EXTERN int SLang_push_array (SLang_Array_Type *at, int do_free);
-SL_EXTERN void SLang_free_array (SLang_Array_Type *at);
+SL_EXTERN int SLang_pop_array_of_type (SLang_Array_Type **, SLtype);
+SL_EXTERN int SLang_pop_array (SLang_Array_Type **, int);
+SL_EXTERN int SLang_push_array (SLang_Array_Type *, int);
+SL_EXTERN void SLang_free_array (SLang_Array_Type *);
 SL_EXTERN SLang_Array_Type *SLang_create_array (SLtype, int, VOID_STAR, SLindex_Type *, unsigned int);
 SL_EXTERN SLang_Array_Type *SLang_duplicate_array (SLang_Array_Type *);
 SL_EXTERN int SLang_get_array_element (SLang_Array_Type *, SLindex_Type *, VOID_STAR);
@@ -906,9 +895,7 @@ SL_EXTERN int SLarray_map_array (SLCONST SLarray_Map_Type *);
 
 /*{{{ Interpreter Function Prototypes */
 
-SL_EXTERN int SLerr_throw (int err, char *msg, SLtype obj_type, VOID_STAR objptr);
 SL_EXTERN void SLang_verror (int, char *, ...) SLATTRIBUTE_PRINTF(2,3);
-SL_EXTERN void SLang_verror_va (int errcode, char *fmt, va_list va);
 SL_EXTERN int SLang_get_error (void);
 SL_EXTERN int SLang_set_error (int);
 SL_EXTERN char *SLerr_strerror (int errcode);
@@ -994,11 +981,6 @@ SL_EXTERN int SL_Import_Error;
 #define SL_INVALID_UTF8			SL_InvalidUTF8_Error
 
   SL_EXTERN int SLang_Traceback;
-#define SL_TB_NONE		0x0
-#define SL_TB_FULL		0x1    /* full traceback */
-#define SL_TB_OMIT_LOCALS	0x2    /* full, but omit local vars */
-#define SL_TB_PARTIAL		0x4    /* show just on line of traceback */
-
   /* If non-zero, dump an S-Lang traceback upon error.  Available as
      _traceback in S-Lang. */
 
@@ -1302,20 +1284,13 @@ SL_EXTERN int SLatoi(unsigned char *);
 SL_EXTERN long SLatol (unsigned char *);
 SL_EXTERN unsigned long SLatoul (unsigned char *);
 
-#ifdef HAVE_LONG_LONG
+#if HAVE_LONG_LONG
 SL_EXTERN long long SLatoll (unsigned char *s);
 SL_EXTERN unsigned long long SLatoull (unsigned char *s);
 #endif
 SL_EXTERN int SLang_pop_fileptr (SLang_MMT_Type **, FILE **);
 SL_EXTERN char *SLang_get_name_from_fileptr (SLang_MMT_Type *);
 
-SL_EXTERN int SLang_get_fileptr (SLang_MMT_Type *, FILE **);
-/* This function may be used to obtain the FILE* object associated with an MMT.
- * It returns 0 if no-errors were encountered, and -1 otherwise.  
- * If FILE* object has been closed, this function will return 0 and set the FILE* 
- * parameter to NULL.
- */
-
 typedef struct _pSLFile_FD_Type SLFile_FD_Type;
 SL_EXTERN SLFile_FD_Type *SLfile_create_fd (char *, int);
 SL_EXTERN void SLfile_free_fd (SLFile_FD_Type *);
@@ -1325,19 +1300,6 @@ SL_EXTERN int SLfile_get_fd (SLFile_FD_Type *, int *);
 SL_EXTERN SLFile_FD_Type *SLfile_dup_fd (SLFile_FD_Type *f0);
 SL_EXTERN int SLang_init_posix_io (void);
 
-/* These routines allow an application to attach an object to a file-descriptor
- * as well a provide read/write/close/etc methods.  To allow the coexistence of
- * different types of objects, a clientdata-id must be used.
- */
-SL_EXTERN int SLfile_set_getfd_method (SLFile_FD_Type *f, int (*func)(VOID_STAR, int *));
-SL_EXTERN int SLfile_set_close_method (SLFile_FD_Type *f, int (*func)(VOID_STAR));
-SL_EXTERN int SLfile_set_read_method (SLFile_FD_Type *f, int (*func)(VOID_STAR, char*, unsigned int));
-SL_EXTERN int SLfile_set_write_method (SLFile_FD_Type *f, int (*func)(VOID_STAR, char*, unsigned int));
-SL_EXTERN int SLfile_set_dup_method (SLFile_FD_Type *f, SLFile_FD_Type *(*func)(VOID_STAR));
-SL_EXTERN int SLfile_create_clientdata_id (int *id);
-SL_EXTERN int SLfile_set_clientdata (SLFile_FD_Type *f, void (*free_func)(VOID_STAR), VOID_STAR cd, int id);
-SL_EXTERN int SLfile_get_clientdata (SLFile_FD_Type *f, int id, VOID_STAR *cdp);
-
 typedef double (*SLang_To_Double_Fun_Type)(VOID_STAR);
 SL_EXTERN SLang_To_Double_Fun_Type SLarith_get_to_double_fun (SLtype, unsigned int *);
 
@@ -1494,8 +1456,6 @@ SL_EXTERN FVOID_STAR SLang_find_key_function(char *, SLkeymap_Type *);
 SL_EXTERN SLkeymap_Type *SLang_find_keymap(char *);
 
 SL_EXTERN int SLang_Last_Key_Char;
-/* Despite the name of this variable, its value is the last BYTE of a key-sequence */
-
 SL_EXTERN int SLang_Key_TimeOut_Flag;
 
 /*}}}*/
@@ -1504,7 +1464,6 @@ SL_EXTERN int SLang_Key_TimeOut_Flag;
 
 typedef struct _pSLrline_Type SLrline_Type;
 SL_EXTERN SLrline_Type *SLrline_open (unsigned int width, unsigned int flags);
-SL_EXTERN SLrline_Type *SLrline_open2 (char *, unsigned int width, unsigned int flags);
 #define SL_RLINE_NO_ECHO	1
 #define SL_RLINE_USE_ANSI	2
 #define SL_RLINE_BLINK_MATCH	4
@@ -1518,7 +1477,6 @@ SL_EXTERN int SLrline_bol (SLrline_Type *);
 SL_EXTERN int SLrline_eol (SLrline_Type *);
 SL_EXTERN int SLrline_del (SLrline_Type *, unsigned int len);
 SL_EXTERN int SLrline_ins (SLrline_Type *, char *s, unsigned int len);
-SL_EXTERN int SLrline_move (SLrline_Type *, int);
 
 SL_EXTERN int SLrline_set_echo (SLrline_Type *, int);
 SL_EXTERN int SLrline_set_tab (SLrline_Type *, unsigned int tabwidth);
@@ -1547,9 +1505,6 @@ SL_EXTERN void SLrline_redraw (SLrline_Type *);
 SL_EXTERN int SLrline_save_line (SLrline_Type *);
 SL_EXTERN int SLrline_add_to_history (SLrline_Type *, char *);
 
-/* Interpreter interface */
-SL_EXTERN int SLrline_init (char *appdef, char *user_initfile, char *sys_initfile);
-
 /* Compatibility */
 typedef SLrline_Type SLang_RLine_Info_Type;
 
@@ -2134,16 +2089,6 @@ SL_EXTERN int SLadd_intrinsic_function (char *, FVOID_STAR, SLtype, unsigned int
 SL_EXTERN int SLns_add_intrinsic_variable (SLang_NameSpace_Type *, char *, VOID_STAR, SLtype, int);
 SL_EXTERN int SLns_add_intrinsic_function (SLang_NameSpace_Type *, char *, FVOID_STAR, SLtype, unsigned int,...);
 
-/* These functions are used to patch intrinsic tables that make use of 
- * dynamically allocated types whose type id is determined at
- * run-time.  The second version is useful for the most common case of a 
- * single mapping.
- */
-extern int SLclass_patch_intrin_fun_table (SLang_Intrin_Fun_Type *table, 
-					 SLtype *from_types, SLtype *to_types, unsigned int num);
-extern int SLclass_patch_intrin_fun_table1 (SLang_Intrin_Fun_Type *table, 
-					  SLtype from_type, SLtype to_type);
-
 #define MAKE_INTRINSIC_N(n,f,out,in,a1,a2,a3,a4,a5,a6,a7) \
     {(n), NULL, SLANG_INTRINSIC, (FVOID_STAR) (f), \
       {a1,a2,a3,a4,a5,a6,a7}, (in), (out)}
@@ -2305,7 +2250,7 @@ SL_EXTERN int SLregexp_get_hints (SLRegexp_Type *, unsigned int *flagsp);
 #define SLREGEXP_HINT_BOL		0x01   /* pattern must match bol */
 #define SLREGEXP_HINT_OSEARCH		0x02   /* ordinary search will do */
 
-SL_EXTERN char *SLregexp_quote_string (char *pattern, char *buf, unsigned int buflen);
+SL_EXTERN char *SLregexp_quote_string (char *, char *, unsigned int);
 
 /*}}}*/
 
