@@ -1,5 +1,5 @@
 /*
-Copyright (C) 2004, 2005, 2006, 2007 John E. Davis
+Copyright (C) 2004, 2005 John E. Davis
 
 This file is part of the S-Lang Library.
 
@@ -187,7 +187,6 @@ static int resize_table (SLang_Assoc_Array_Type *a)
      }
    a->elements = new_es;
    a->table_len = new_table_len;
-   a->num_occupied -= a->num_deleted;
    a->num_deleted = 0;
    a->resize_num = 13*(new_table_len>>4);
    
@@ -264,12 +263,12 @@ static SLang_Assoc_Array_Type *alloc_assoc_array (SLtype type, int has_default_v
 }
 
 
-static int store_object (SLang_Assoc_Array_Type *a, SLstr_Type *s, SLang_Object_Type *obj)
+static int store_object (SLang_Assoc_Array_Type *a, char *s, SLang_Object_Type *obj)
 {
    unsigned long hash;
    _pSLAssoc_Array_Element_Type *e;
 
-   hash = _pSLstring_get_hash (s);
+   hash = _pSLcompute_string_hash (s);
    if (NULL != (e = find_element (a, s, hash)))
      SLang_free_object (&e->value);
    else
@@ -303,7 +302,7 @@ static void assoc_destroy (SLtype type, VOID_STAR ptr)
 static int pop_index (unsigned int num_indices,
 		      SLang_MMT_Type **mmt,
 		      SLang_Assoc_Array_Type **a,
-		      SLstr_Type **str)
+		      char **str)
 {
    if (NULL == (*mmt = SLang_pop_mmt (SLANG_ASSOC_TYPE)))
      {
@@ -331,8 +330,9 @@ static int pop_index (unsigned int num_indices,
 int _pSLassoc_aget (SLtype type, unsigned int num_indices)
 {
    unsigned long hash;
+   unsigned int len;
    SLang_MMT_Type *mmt;
-   SLstr_Type *str;
+   char *str;
    _pSLAssoc_Array_Element_Type *e;
    SLang_Assoc_Array_Type *a;
    SLang_Object_Type *obj;
@@ -343,7 +343,8 @@ int _pSLassoc_aget (SLtype type, unsigned int num_indices)
    if (-1 == pop_index (num_indices, &mmt, &a, &str))
      return -1;
 
-   hash = _pSLstring_get_hash (str);
+   len = strlen (str);
+   hash = _pSLstring_hash ((unsigned char *)str, (unsigned char *)str + len);
    e = find_element (a, str, hash);
 
    if (e == NULL)
@@ -370,7 +371,7 @@ int _pSLassoc_aget (SLtype type, unsigned int num_indices)
    
    free_and_return:
 
-   _pSLang_free_slstring (str);
+   _pSLfree_hashed_string (str, len, hash);
    SLang_free_mmt (mmt);
    return ret;
 }
@@ -378,7 +379,7 @@ int _pSLassoc_aget (SLtype type, unsigned int num_indices)
 int _pSLassoc_aput (SLtype type, unsigned int num_indices)
 {
    SLang_MMT_Type *mmt;
-   SLstr_Type *str;
+   char *str;
    SLang_Assoc_Array_Type *a;
    SLang_Object_Type obj;
    int ret;
@@ -411,7 +412,7 @@ int _pSLassoc_aput (SLtype type, unsigned int num_indices)
      }
 
    the_return:
-   _pSLang_free_slstring (str);
+   SLang_free_slstring (str);
    SLang_free_mmt (mmt);
    return ret;
 }
@@ -586,11 +587,11 @@ static void assoc_delete_key (SLang_Assoc_Array_Type *a, char *key)
 {
    _pSLAssoc_Array_Element_Type *e;
 
-   e = find_element (a, key, _pSLstring_get_hash (key));
+   e = find_element (a, key, _pSLcompute_string_hash (key));
    if (e == NULL)
      return;
    
-   _pSLang_free_slstring (e->key);
+   SLang_free_slstring (e->key);
    SLang_free_object (&e->value);
    e->key = Deleted_Key;
    a->num_deleted++;
@@ -666,12 +667,12 @@ cl_foreach_open (SLtype type, unsigned int num)
 	     SLang_verror (SL_NOT_IMPLEMENTED,
 			   "using '%s' not supported by SLassoc_Type",
 			   s);
-	     _pSLang_free_slstring (s);
+	     SLang_free_slstring (s);
 	     SLang_free_mmt (mmt);
 	     return NULL;
 	  }
 
-	_pSLang_free_slstring (s);
+	SLang_free_slstring (s);
      }
 
    if (NULL == (c = (SLang_Foreach_Context_Type *) SLmalloc (sizeof (SLang_Foreach_Context_Type))))
