@@ -1,6 +1,6 @@
 /* Complex Data Type definition for S-Lang */
 /*
-Copyright (C) 2004, 2005, 2006, 2007 John E. Davis
+Copyright (C) 2004, 2005 John E. Davis
 
 This file is part of the S-Lang Library.
 
@@ -44,7 +44,7 @@ int SLang_pop_complex (double *r, double *i)
    switch (SLang_peek_at_stack ())
      {
       case SLANG_COMPLEX_TYPE:
-	if (-1 == SLclass_pop_ptr_obj (SLANG_COMPLEX_TYPE, VOID_STAR_STAR(&c)))
+	if (-1 == SLclass_pop_ptr_obj (SLANG_COMPLEX_TYPE, (VOID_STAR *)&c))
 	  return -1;
 	*r = c[0];
 	*i = c[1];
