@@ -1,9 +1,26 @@
-/* Copyright (c) 1998, 1999 John E. Davis
- * This file is part of the S-Lang library.
- *
- * You may distribute under the terms of either the GNU General Public
- * License or the Perl Artistic License.
- */
+/*
+Copyright (C) 2004, 2005 John E. Davis
+
+This file is part of the S-Lang Library.
+
+The S-Lang Library is free software; you can redistribute it and/or
+modify it under the terms of the GNU General Public License as
+published by the Free Software Foundation; either version 2 of the
+License, or (at your option) any later version.
+
+The S-Lang Library is distributed in the hope that it will be useful,
+but WITHOUT ANY WARRANTY; without even the implied warranty of
+MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+General Public License for more details.
+
+You should have received a copy of the GNU General Public License
+along with this library; if not, write to the Free Software
+Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
+USA.  
+*/
+
+/* Support for UTF-8 combining characters added by Adrian Colley */
+
 #include "slinclud.h"
 
 #include <signal.h>
@@ -18,20 +35,42 @@
  * different that a 100% emulation is not possible.
  */
 
+#define SLCURSES_NULLATTR ((SLcurses_Char_Type)0)
+#define SLCURSES_NULLCHAR ((SLwchar_Type)0)
+
+#define SLCURSES_BUILD_CHAR(ch,wch,c) \
+   ((ch) = ((SLcurses_Char_Type)(wch)) | (((SLcurses_Char_Type)(c)) << 24))
+
+#define SLCURSES_BUILD_CELL(cellp,wch,c) \
+   do \
+     { \
+	int slcurses_build_cell_i; \
+	SLcurses_Cell_Type *slcurses_build_cell_c = (cellp); \
+	SLCURSES_BUILD_CHAR(slcurses_build_cell_c->main, wch, c); \
+	for (slcurses_build_cell_i = 0; \
+	     slcurses_build_cell_i < SLSMG_MAX_CHARS_PER_CELL-1; \
+	     slcurses_build_cell_i++) \
+	  slcurses_build_cell_c->combining[slcurses_build_cell_i] = SLCURSES_NULLCHAR; \
+     } \
+   while (0)
+
+#define SLCURSES_EXTRACT_COLOR(ch) (((ch >> 24)&0xFF))
+#define SLCURSES_EXTRACT_CHAR(ch) ((ch)&A_CHARTEXT)
+
+#define SLTT_EXTRACT_CHAR(ch) ((ch)&A_CHARTEXT)
+
 SLcurses_Window_Type *SLcurses_Stdscr;
 int SLcurses_Esc_Delay = 150;	       /* 0.15 seconds */
 SLtt_Char_Type SLcurses_Acs_Map [128];
 int SLcurses_Is_Endwin = 1;
 int SLcurses_Num_Colors = 8;
 
-static void blank_line (unsigned short *b, unsigned int len, unsigned short color)
+static void blank_line (SLcurses_Cell_Type *b, unsigned int len, SLsmg_Color_Type color)
 {
-   unsigned short *bmax;
+   SLcurses_Cell_Type *bmax = b + len;
 
-   bmax = b + len;
-   color = (color << 8) | 32;
-
-   while (b < bmax) *b++ = color;
+   while (b < bmax)
+     SLCURSES_BUILD_CELL(b++,' ',color);
 }
 
 static int va_mvprintw (SLcurses_Window_Type *w, int r, int c, int do_move,
@@ -41,7 +80,7 @@ static int va_mvprintw (SLcurses_Window_Type *w, int r, int c, int do_move,
 
    if (do_move) SLcurses_wmove (w, r, c);
 
-   (void) _SLvsnprintf (buf, sizeof(buf), fmt, ap);
+   (void) SLvsnprintf (buf, sizeof(buf), fmt, ap);
 
    SLcurses_waddnstr (w, buf, -1);
    return 0;
@@ -107,6 +146,62 @@ int SLcurses_nodelay (SLcurses_Window_Type *w, int onoff)
    return 0;
 }
 
+static unsigned char Keyboard_Buffer[256];
+static unsigned char *Keyboard_Buffer_Start = Keyboard_Buffer;
+static unsigned char *Keyboard_Buffer_Stop = Keyboard_Buffer;
+
+static int getkey_function (void)
+{
+   int ch;
+
+   ch = SLang_getkey ();
+   if (ch != SLANG_GETKEY_ERROR)
+     {
+	*Keyboard_Buffer_Stop++ = (unsigned char) ch;
+	if (Keyboard_Buffer_Stop == Keyboard_Buffer + sizeof (Keyboard_Buffer))
+	  Keyboard_Buffer_Stop = Keyboard_Buffer;
+     }
+   
+   return ch;
+}
+
+static int get_buffered_key (void)
+{
+   int ch;
+   
+   if (Keyboard_Buffer_Stop == Keyboard_Buffer_Start)
+     return SLANG_GETKEY_ERROR;
+
+   ch = *Keyboard_Buffer_Start++;
+   if (Keyboard_Buffer_Start == Keyboard_Buffer + sizeof (Keyboard_Buffer))
+     Keyboard_Buffer_Start = Keyboard_Buffer;
+
+   return ch;
+}
+
+static int get_keypad_key (void)
+{
+   int ch;
+   
+   if (Keyboard_Buffer_Stop != Keyboard_Buffer_Start)
+     return get_buffered_key ();
+
+   ch = SLang_getkey ();
+   if (ch == '\033')
+     {
+	if (0 == SLang_input_pending (ESCDELAY / 100))
+	  return ch;
+     }
+   else if (ch == SLANG_GETKEY_ERROR) return ERR;
+   SLang_ungetkey (ch);
+   ch = SLkp_getkey ();
+   if (ch == SL_KEY_ERR)
+     ch = get_buffered_key ();
+   else
+     Keyboard_Buffer_Stop = Keyboard_Buffer_Start;
+   return ch;
+}
+
 int SLcurses_wgetch (SLcurses_Window_Type *w)
 {
    if (w == NULL)
@@ -114,21 +209,13 @@ int SLcurses_wgetch (SLcurses_Window_Type *w)
 
    SLcurses_wrefresh (w);
 
-   if ((w->delay_off == -1) ||
-       SLang_input_pending (w->delay_off))
+   if ((Keyboard_Buffer_Start != Keyboard_Buffer_Stop)
+       || (w->delay_off == -1)
+       || SLang_input_pending (w->delay_off))
      {
 	if (w->use_keypad)
-	  {
-	     int ch = SLang_getkey ();
-	     if (ch == '\033')
-	       {
-		  if (0 == SLang_input_pending (ESCDELAY / 100))
-		    return ch;
-	       }
-	     else if (ch == 0xFFFF) return ERR;
-	     SLang_ungetkey (ch);
-	     return SLkp_getkey ();
-	  }
+	  return get_keypad_key ();
+
 	return SLang_getkey ();
      }
 
@@ -150,7 +237,7 @@ static unsigned int map_attr_to_object (SLtt_Char_Type attr)
    unsigned int obj;
    SLtt_Char_Type at;
 
-   obj = (attr >> 8) & 0xFF;
+   obj = SLCURSES_EXTRACT_COLOR(attr);
 
    if (SLtt_Use_Ansi_Colors)
      {
@@ -161,6 +248,7 @@ static unsigned int map_attr_to_object (SLtt_Char_Type attr)
 	if (attr & A_BOLD) at |= SLTT_BOLD_MASK;
 	if (attr & A_UNDERLINE) at |= SLTT_ULINE_MASK;
 	if (attr & A_REVERSE) at |= SLTT_REV_MASK;
+	if (attr & A_ALTCHARSET) at |= SLTT_ALTC_MASK;
 
 	SLtt_set_color_object (obj, at);
 
@@ -263,10 +351,10 @@ SLcurses_Window_Type *SLcurses_initscr (void)
        || (NULL == (SLcurses_Stdscr = SLcurses_newwin (0, 0, 0, 0)))
        || (-1 == SLsmg_init_smg ()))
      {
-	SLang_doerror (NULL);
 	SLang_exit_error ("SLcurses_initscr: init failed\n");
 	return NULL;
      }
+   SLkp_set_getkey_function (getkey_function);
 
 #ifdef SIGINT
    signal (SIGINT, sigint_handler);
@@ -276,12 +364,12 @@ SLcurses_Window_Type *SLcurses_initscr (void)
    signal (SIGTSTP, sigtstp_handler);
 #endif
 
-   SLtt_set_mono (A_BOLD >> 8, NULL, SLTT_BOLD_MASK);
-   SLtt_set_mono (A_UNDERLINE >> 8, NULL, SLTT_ULINE_MASK);
-   SLtt_set_mono (A_REVERSE >> 8, NULL, SLTT_REV_MASK);
-   /* SLtt_set_mono (A_BLINK >> 8, NULL, SLTT_BLINK_MASK); */
-   SLtt_set_mono ((A_BOLD|A_UNDERLINE) >> 8, NULL, SLTT_ULINE_MASK|SLTT_BOLD_MASK);
-   SLtt_set_mono ((A_REVERSE|A_UNDERLINE) >> 8, NULL, SLTT_ULINE_MASK|SLTT_REV_MASK);
+   SLtt_set_mono (SLCURSES_EXTRACT_COLOR(A_BOLD), NULL, SLTT_BOLD_MASK);
+   SLtt_set_mono (SLCURSES_EXTRACT_COLOR(A_UNDERLINE), NULL, SLTT_ULINE_MASK);
+   SLtt_set_mono (SLCURSES_EXTRACT_COLOR(A_REVERSE), NULL, SLTT_REV_MASK);
+   /* SLtt_set_mono (SLCURSES_EXTRACT_COLOR(A_BLINK), NULL, SLTT_BLINK_MASK); */
+   SLtt_set_mono ((SLCURSES_EXTRACT_COLOR(A_BOLD|A_UNDERLINE)), NULL, SLTT_ULINE_MASK|SLTT_BOLD_MASK);
+   SLtt_set_mono ((SLCURSES_EXTRACT_COLOR(A_REVERSE|A_UNDERLINE)), NULL, SLTT_ULINE_MASK|SLTT_REV_MASK);
 
    if (SLtt_Has_Alt_Charset)
      {
@@ -351,7 +439,7 @@ int SLcurses_delwin (SLcurses_Window_Type *w)
    if (w == NULL) return 0;
    if (w->lines != NULL)
      {
-	unsigned short **lines = w->lines;
+	SLcurses_Cell_Type **lines = w->lines;
 	if (w->is_subwin == 0)
 	  {
 	     unsigned int r, rmax;
@@ -376,7 +464,7 @@ SLcurses_Window_Type *SLcurses_newwin (unsigned int nrows, unsigned int ncols,
 				       unsigned int r, unsigned int c)
 {
    SLcurses_Window_Type *win;
-   unsigned short **lines;
+   SLcurses_Cell_Type **lines;
 
    if (r >= (unsigned int) SLtt_Screen_Rows)
      return NULL;
@@ -393,14 +481,14 @@ SLcurses_Window_Type *SLcurses_newwin (unsigned int nrows, unsigned int ncols,
    if (ncols == 0)
      ncols = (unsigned int) SLtt_Screen_Cols - c;
 
-   lines = (unsigned short **) SLmalloc (nrows * sizeof (unsigned short *));
+   lines = (SLcurses_Cell_Type **) SLmalloc (nrows * sizeof (SLcurses_Cell_Type *));
    if (lines == NULL)
      {
 	SLcurses_delwin (win);
 	return NULL;
      }
 
-   SLMEMSET ((char *) lines, 0, nrows * sizeof (unsigned short *));
+   SLMEMSET ((char *) lines, 0, nrows * sizeof (SLcurses_Cell_Type *));
 
    win->lines = lines;
    win->scroll_max = win->nrows = nrows;
@@ -414,9 +502,9 @@ SLcurses_Window_Type *SLcurses_newwin (unsigned int nrows, unsigned int ncols,
 
    for (r = 0; r < nrows; r++)
      {
-	unsigned short *b;
+	SLcurses_Cell_Type *b;
 
-	b = (unsigned short *) SLmalloc (ncols * sizeof (short));
+	b = (SLcurses_Cell_Type *) SLmalloc (ncols * sizeof (SLcurses_Cell_Type));
 	if (b == NULL)
 	  {
 	     SLcurses_delwin (win);
@@ -452,23 +540,102 @@ static int do_newline (SLcurses_Window_Type *w)
    return 0;
 }
 
+/* SLcurses_placechar(w, wch, width, color) places character wch at the current
+ * position in w.  If any multicolumn characters are partially overwritten as a
+ * result, the remains of those characters are blanked out.  The caller is
+ * responsible for ensuring that there is room at the current position for
+ * (width) columns.  The current position is not advanced by this function.  If
+ * width is zero, this function adds the character to the combining[] array of
+ * the current position, if there's room (the character is dropped if there's
+ * no room).
+ */
+static void SLcurses_placechar (SLcurses_Window_Type *w, SLwchar_Type wch,
+				int width, int color)
+{
+   SLcurses_Cell_Type *b;
+   unsigned int i;
+
+   if (width <= 0)
+     {
+	/* Backtrack to the start of any multicolumn character. */
+	for (i = w->_curx; i > 0; i--)
+	  if (w->lines[w->_cury][i].main != SLCURSES_NULLATTR)
+	    break;
+	b = &w->lines[w->_cury][i];
+	/* Find a place for the new combining character. */
+	for (i = 0; i < SLSMG_MAX_CHARS_PER_CELL-1; i++)
+	  if (b->combining[i] == SLCURSES_NULLCHAR)
+	    break;
+	/* Add character if it fits. */
+	if (i < SLSMG_MAX_CHARS_PER_CELL-1)
+	  b->combining[i] = wch;
+	return;
+     }
+
+   b = &w->lines[w->_cury][w->_curx];
+   if (b->main == SLCURSES_NULLATTR)
+     {
+	/* wch will overwrite part of a multicolumn character, so
+	 * blank out the left half of the character.
+	 */
+	int blankcolor = w->color;
+	i = w->_curx;
+	while (1)
+          {
+	     if (w->lines[w->_cury][i].main != SLCURSES_NULLATTR)
+	       {
+		  blankcolor = SLCURSES_EXTRACT_COLOR(w->lines[w->_cury][i].main);
+		  break;
+	       }
+	     if (i == 0)
+	       break;
+	     i--;
+	  }
+	while (i < w->_curx)
+	  {
+	     SLCURSES_BUILD_CELL(&w->lines[w->_cury][i], ' ', blankcolor);
+	     i++;
+	  }
+     }
+   /* Place wch itself. */
+   SLCURSES_BUILD_CELL(b, wch, w->color);
+   /* If wch is multicolumn, reserve the following cell(s). */
+   for (i = 1; i < (unsigned int) width; i++)
+     {
+	b[i].main = SLCURSES_NULLATTR;
+	/* It's unnecessary to clear b[i].combined when b[i].main==0. */
+     }
+   /* If wch is followed by the tail of a previous, overwritten, multicolumn
+    * character, then blank it out completely.
+    */
+   for (i = w->_curx + width; i < w->ncols; i++)
+     {
+	b = &w->lines[w->_cury][i];
+	if (b->main != SLCURSES_NULLATTR)
+	  break;
+	SLCURSES_BUILD_CELL(b, ' ', color);
+     }
+}
+
 int SLcurses_waddch (SLcurses_Window_Type *win, SLtt_Char_Type attr)
 {
-   unsigned short *b, ch;
-   unsigned short color;
+   SLtt_Char_Type ch;
+   SLsmg_Color_Type color;
+   int width;
 
    if (win == NULL) return -1;
 
    if (win->_cury >= win->nrows)
      {
-	/* Curses seems to move current postion to top of window. */
+	/* Curses seems to move current position to top of window. */
 	win->_cury = win->_curx = 0;
 	return -1;
      }
 
    win->modified = 1;
 
-   ch = attr & 0xFF;
+   ch = SLTT_EXTRACT_CHAR(attr);
+   if (ch == 0) return -1;
 
    if (attr == ch)
      color = win->color;
@@ -477,12 +644,12 @@ int SLcurses_waddch (SLcurses_Window_Type *win, SLtt_Char_Type attr)
 	/* hack to pick up the default color for graphics chars */
 	if (((attr & A_COLOR) == 0) && ((attr & A_ALTCHARSET) != 0))
 	  {
-	     attr |= win->color << 8;
+	     SLCURSES_BUILD_CHAR(attr, attr, win->color);
 	  }
 	color = map_attr_to_object (attr);
      }
 
-   if (ch < ' ')
+   if (SLwchar_iscntrl((SLwchar_Type)ch))
      {
 	if (ch == '\n')
 	  {
@@ -504,20 +671,63 @@ int SLcurses_waddch (SLcurses_Window_Type *win, SLtt_Char_Type attr)
 	     return 0;
 	  }
 
-	/* HACK HACK!!!! */
-	if (ch == '\t') ch = ' ';
+	if (ch == '\t')
+	  {
+	     int err;
+	     do
+	       err = SLcurses_waddch (win, (SLtt_Char_Type)' ');
+	     while (err == 0 && win->_curx % SLsmg_Tab_Width != 0);
+	     return err;
+	  }
      }
 
-   if (win->_curx >= win->ncols)
-     do_newline (win);
+   width = SLwchar_isprint (ch)
+	? (SLsmg_is_utf8_mode ()
+		? SLwchar_wcwidth (ch)
+		: 1)
+	: 0;
+   if (win->_curx + width > win->ncols)
+     {
+	SLcurses_wclrtoeol (win);
+	do_newline (win);
+     }
 
-   b = win->lines[win->_cury] + win->_curx;
-   *b = ch | (color << 8);
-   win->_curx++;
+   SLcurses_placechar (win, ch, width, color);
+   win->_curx += width;
 
    return 0;
 }
 
+static void write_color_chars (SLcurses_Cell_Type *p, unsigned int len)
+{
+   int color = -1;
+   unsigned int i;
+   
+   for (i = 0; i < len; i++)
+     {
+	SLwchar_Type ch;
+	int this_color;
+	int j;
+	if (p[i].main == SLCURSES_NULLATTR)
+	  continue;	/* prev char was multicolumn */
+	ch = SLCURSES_EXTRACT_CHAR(p[i].main);
+	this_color = SLCURSES_EXTRACT_COLOR(p[i].main);
+	if (this_color != color)
+	  {
+	     SLsmg_set_color (this_color);
+	     color = this_color;
+	  }
+	SLsmg_write_char (ch);
+	for (j = 0; j < SLSMG_MAX_CHARS_PER_CELL-1; j++)
+	  {
+	     SLwchar_Type combining = p[i].combining[j];
+	     if (combining == SLCURSES_NULLCHAR)
+	       break;	/* no more combining chars */
+	     SLsmg_write_char (combining);
+	  }
+     }
+}
+
 int SLcurses_wnoutrefresh (SLcurses_Window_Type *w)
 {
    unsigned int len;
@@ -549,7 +759,7 @@ int SLcurses_wnoutrefresh (SLcurses_Window_Type *w)
    for (i = 0; i < imax; i++)
      {
 	SLsmg_gotorc (r, c);
-	SLsmg_write_color_chars (w->lines[i], len);
+	write_color_chars (w->lines[i], len);
 	r++;
      }
 
@@ -576,8 +786,8 @@ int SLcurses_wrefresh (SLcurses_Window_Type *w)
 
 int SLcurses_wclrtoeol (SLcurses_Window_Type *w)
 {
-   unsigned short *b, *bmax;
-   unsigned short blank;
+   SLcurses_Cell_Type *b, *bmax;
+   int color;
 
    if (w == NULL) return -1;
    if (w->_cury >= w->nrows)
@@ -585,33 +795,35 @@ int SLcurses_wclrtoeol (SLcurses_Window_Type *w)
 
    w->modified = 1;
 
-   blank = (w->color << 8) | 0x20;
-
    b = w->lines[w->_cury];
    bmax = b + w->ncols;
    b += w->_curx;
 
-   while (b < bmax) *b++ = blank;
+   color = w->color;
+   while (b < bmax)
+     SLCURSES_BUILD_CELL(b++,' ',color);
+
    return 0;
 }
 
 int SLcurses_wclrtobot (SLcurses_Window_Type *w)
 {
-   unsigned short *b, *bmax;
-   unsigned short blank;
    unsigned int r;
+   int color;
 
    if (w == NULL) return -1;
 
    w->modified = 1;
-   blank = (w->color << 8) | 0x20;
+   color = w->color;
    SLcurses_wclrtoeol (w);
    for (r = w->_cury + 1; r < w->nrows; r++)
      {
+	SLcurses_Cell_Type *b, *bmax;
 	b = w->lines [r];
 	bmax = b + w->ncols;
 
-	while (b < bmax) *b++ = blank;
+	while (b < bmax)
+	  SLCURSES_BUILD_CELL(b++,' ',color);
      }
 
    return 0;
@@ -619,9 +831,9 @@ int SLcurses_wclrtobot (SLcurses_Window_Type *w)
 
 int SLcurses_wscrl (SLcurses_Window_Type *w, int n)
 {
-   unsigned short **lines;
-   unsigned int r, rmax, rmin, ncols;
-   unsigned short color;
+   SLcurses_Cell_Type **lines;
+   unsigned int r0, r1, rmax, rmin, ncols;
+   SLsmg_Color_Type color;
 
    if ((w == NULL) || (w->scroll_ok == 0))
      return -1;
@@ -644,35 +856,70 @@ int SLcurses_wscrl (SLcurses_Window_Type *w, int n)
    rmin = w->scroll_min;
    if (rmax > w->nrows)
      rmax = w->nrows;
-   if (rmin >= rmax)
+   if ((rmin >= rmax) || (rmax == 0))
      return 0;
 
-   while (n > 0)
+   if (n == 0)
+     return 0;
+
+   if (n > 0)
      {
-	for (r = rmin + 1; r < rmax; r++)
+	r0 = rmin;
+	r1 = rmin + n;
+
+	while (r1 < rmax)
 	  {
-	     /* lines[r - 1] = lines[r]; */
-	     memcpy ((char *)lines[r - 1], (char *)lines[r],
-		     sizeof (short) * ncols);
+	     if (w->is_subwin)
+	       memcpy ((char *)lines[r0], (char *)lines[r1],
+		       sizeof (SLcurses_Cell_Type) * ncols);
+	     else
+	       {
+		  SLcurses_Cell_Type *swap = lines[r0];
+		  lines[r0] = lines[r1];
+		  lines[r1] = swap;
+	       }
+	     r0++;
+	     r1++;
 	  }
-	blank_line (lines[rmax - 1], ncols, color);
-	n--;
+	while (r0 < rmax)
+	  {
+	     blank_line (lines[r0], ncols, color);
+	     r0++;
+	  }
+	return 0;
      }
 
-   rmax--;
-   while (n < 0)
+   /* else n < 0 */
+   n = -n;
+
+   r1 = rmax - 1;
+   if (r1 < (unsigned int) n)
+     n = r1;
+   r0 = r1 - n;
+
+   while (r0 >= rmin)
      {
-	for (r = rmax; r > rmin; r--)
+	if (w->is_subwin)
+	  memcpy ((char *)lines[r1], (char *)lines[r0],
+		  sizeof (SLcurses_Cell_Type) * ncols);
+	else
 	  {
-	     memcpy ((char *)lines[r], (char *)lines[r - 1],
-		     sizeof (short) * ncols);
+	     SLcurses_Cell_Type *swap = lines[r1];
+	     lines[r1] = lines[r0];
+	     lines[r0] = swap;
 	  }
-	blank_line (lines[rmin], ncols, color);
-	n++;
+	r1--;
+	if (r0 == 0)
+	  break;
+	r0--;
+     }
+   r0 = rmin;
+   while (r0 <= r1)
+     {
+	blank_line (lines[r0], ncols, color);
+	r0++;
      }
 
-   /* wmove (w, w->nrows - 1, 0); */
-   /* wclrtobot (w); */
    return 0;
 }
 
@@ -680,10 +927,9 @@ int SLcurses_wscrl (SLcurses_Window_Type *w, int n)
  */
 int SLcurses_waddnstr (SLcurses_Window_Type *w, char *str, int len)
 {
-   unsigned short *b;
-   unsigned short color;
-   unsigned char ch;
+   SLsmg_Color_Type color;
    unsigned int nrows, ncols, crow, ccol;
+   SLuchar_Type *u, *umax;
 
    if ((w == NULL)
        || (str == NULL))
@@ -694,7 +940,7 @@ int SLcurses_waddnstr (SLcurses_Window_Type *w, char *str, int len)
    ncols = w->ncols;
    crow = w->_cury;
    ccol = w->_curx;
-   color = (w->color << 8);
+   color = w->color;
 
    if (w->scroll_max <= nrows)
      nrows = w->scroll_max;
@@ -702,12 +948,39 @@ int SLcurses_waddnstr (SLcurses_Window_Type *w, char *str, int len)
    if (crow >= nrows)
      crow = 0;			       /* wrap back to top */
 
-   b = w->lines [crow] + ccol;
-
-   while (len && ((ch = (unsigned char) *str++) != 0))
+   u = (SLuchar_Type *)str;
+   umax = &u[len >= 0 ? (unsigned int)len : strlen(str)];
+   while (u < umax)
      {
-	len--;
+	SLwchar_Type ch;
+	unsigned int nconsumed;
+	int width = 1;
 
+	if (SLsmg_is_utf8_mode () && SLutf8_decode (u, umax, &ch, &nconsumed))
+	  {
+	     u += nconsumed;
+	     if ((ch & A_CHARTEXT) != ch)
+	       {
+		  ch = (SLwchar_Type)0xFFFDL;	/* Unicode replacement character */
+		  width = 1;
+	       }
+	     else if (SLwchar_isprint (ch))
+	       width = SLwchar_wcwidth (ch);
+	     else
+	       width = 0;	/* FIXME: cope with <%02X> printstrings. */
+	  }
+	else
+	  {
+	     ch = (SLwchar_Type)*u++;
+	     if (ch < 0x20 || (ch >= 0x7f && ch < 0xa0))
+	       width = 0;	/* FIXME: use display_8bit */
+	  }
+	if (ch == '\t')
+	  width = 1;	/* HACK forcing linewrap if ccol==ncols */
+	if (ch == 0)
+	  continue;	/* Avoid adding a literal SLCURSES_NULLCHAR. */
+
+	/* FIXME; should this function be defined in terms of waddch? */
 	if (ch == '\n')
 	  {
 	     w->_cury = crow;
@@ -716,40 +989,40 @@ int SLcurses_waddnstr (SLcurses_Window_Type *w, char *str, int len)
 	     do_newline (w);
 	     crow = w->_cury;
 	     ccol = w->_curx;
-	     b = w->lines[crow];
 	     continue;
 	  }
 
-	if (ccol >= ncols)
+	if (ccol + width > ncols)
 	  {
-	     ccol = 0;
-	     crow++;
+	     w->_curx = ccol;
+	     w->_cury = crow;
+	     SLcurses_wclrtoeol(w);	/* no-op if width<=1 */
+	     w->_curx = ccol = 0;
+	     w->_cury = ++crow;
 	     if (crow >= nrows)
 	       {
-		  w->_curx = 0;
-		  w->_cury = crow;
 		  do_newline (w);
 		  crow = w->_cury;
 		  ccol = w->_curx;
 	       }
-
-	     b = w->lines [crow];
 	  }
 
 	if (ch == '\t')
 	  {
-	     unsigned int n = ccol;
-	     n += SLsmg_Tab_Width;
-	     n = SLsmg_Tab_Width - (n % SLsmg_Tab_Width);
-	     if (ccol + n > ncols) n = ncols - len;
-	     ccol += n;
-	     while (n--)
-	       *b++ = color | ' ';
+	     /* assert (ccol < ncols); */
+	     w->_curx = ccol;
+	     w->_cury = crow;
+	     do
+	       {
+		  SLcurses_placechar (w, (SLwchar_Type)' ', 1, w->color);
+		  w->_curx = ++ccol;
+	       }
+	     while (ccol < ncols && ccol % SLsmg_Tab_Width != 0);
 	     continue;
 	  }
 
-	*b++ = color | ch;
-	ccol++;
+	SLcurses_placechar (w, ch, width, w->color);
+	w->_curx = (ccol += width);
      }
 
    w->_curx = ccol;
@@ -764,7 +1037,7 @@ int SLcurses_waddnstr (SLcurses_Window_Type *w, char *str, int len)
 #if 0
 int SLcurses_overlay (SLcurses_Window_Type *swin, SLcurses_Window_Type *dwin)
 {
-   unsigned short *s, *smax, *d, *dmax;
+   SLcurses_Char_Type *s, *smax, *d, *dmax;
 
    if ((swin == NULL) || (dwin == NULL))
      return -1;
@@ -776,8 +1049,8 @@ int SLcurses_overlay (SLcurses_Window_Type *swin, SLcurses_Window_Type *dwin)
 
    while ((s < smax) && (d < dmax))
      {
-	unsigned short ch = *s++;
-	if ((ch & 0xFF) != ' ')
+	SLcurses_Char_Type ch = *s++;
+	if (SLSMG_EXTRACT_CHAR(ch) != ' ')
 	  *d = ch;
 	d++;
      }
@@ -823,7 +1096,7 @@ SLcurses_Window_Type *SLcurses_subwin (SLcurses_Window_Type *orig,
    sw->_maxx = (begin_x + ncols) - 1;
    sw->_maxy = (begin_y + nlines) - 1;
 
-   sw->lines = (unsigned short **) SLmalloc (nlines * sizeof (unsigned short *));
+   sw->lines = (SLcurses_Cell_Type **) SLmalloc (nlines * sizeof (SLcurses_Cell_Type *));
    if (sw->lines == NULL)
      {
 	SLcurses_delwin (sw);
@@ -851,22 +1124,32 @@ int SLcurses_wclear (SLcurses_Window_Type *w)
 
 int SLcurses_wdelch (SLcurses_Window_Type *w)
 {
-   unsigned short *p, *p1, *pmax;
+   int src, dest, limit;
+   SLcurses_Cell_Type *line = w->lines[w->_cury];
 
-   p = w->lines[w->_cury];
-   pmax = p + w->ncols;
-   p += w->_curx;
-   p1 = p + 1;
+   dest = w->_curx;
+   /* Backtrack to start of this character. */
+   while (dest > 0 && line[dest].main == SLCURSES_NULLATTR)
+     dest--;
+   w->_curx = dest;	/* Point to start of character, not the middle. */
 
-   while (p1 < pmax)
-     {
-	*p = *p1;
-	p = p1;
-	p1++;
-     }
+   limit = w->ncols;
+   src = dest + 1;
+   /* Advance to start of next character. */
+   while (src < limit && line[src].main == SLCURSES_NULLATTR)
+     src++;
 
-   if (p < pmax)
-     *p = ' ' | (w->color << 8);
+   /* line[dest..dest+limit-1-src] = line[src..limit-1]; */
+#ifdef HAVE_MEMMOVE
+   memmove(&line[dest], &line[src], (limit-src) * sizeof SLcurses_Cell_Type);
+   dest += (limit-src);
+#else
+   while (src < limit)
+     line[dest++] = line[src++];
+#endif
+
+   while (dest < limit)
+     SLCURSES_BUILD_CELL(&line[dest++], ' ', w->color);
 
    w->modified = 1;
    return 0;
@@ -874,22 +1157,88 @@ int SLcurses_wdelch (SLcurses_Window_Type *w)
 
 int SLcurses_winsch (SLcurses_Window_Type *w, int ch)
 {
-   unsigned short *p, *p1, *pmax;
+   int colsneeded;
+#ifndef HAVE_MEMMOVE
+   int dest, src;
+#endif
+   SLcurses_Cell_Type *line = w->lines[w->_cury];
 
-   p = w->lines[w->_cury];
-   pmax = p + w->ncols;
-   p += w->_curx;
-   p1 = pmax - 1;
+   /* Backtrack to start of this character. */
+   while (w->_curx > 0 && line[w->_curx].main == SLCURSES_NULLATTR)
+     w->_curx--;
 
-   while (pmax > p)
+   if (ch == (SLcurses_Char_Type)'\t')
+     ch = ' ';	/* FIXME */
+
+   if (SLwchar_isprint (ch))
      {
-	*pmax = *p1;
-	pmax = p1;
-	p1--;
+	if (SLsmg_is_utf8_mode ())
+	  colsneeded = SLwchar_wcwidth (ch);
+	else
+	  colsneeded = 1;
+     }
+   else colsneeded = 0;
+
+   if (colsneeded == 0)
+     {
+	int x, i;
+	SLcurses_Cell_Type *b = NULL;
+	/* Back one character. */
+	for (x = w->_curx - 1; x >= 0; x--)
+	  {
+	     b = &line[x];
+	     if (b->main != SLCURSES_NULLATTR)
+	       break;
+	  }
+	if (x < 0)
+	  {	/* Back to end of previous line. */
+	     if (w->_cury == 0)
+	       return (-1);
+	     line = w->lines[w->_cury - 1];
+	     for (x = w->ncols - 1; x >= 0; x--)
+	       {
+		  b = &line[x];
+		  if (b->main != SLCURSES_NULLATTR)
+		    break;
+	       }
+	     if (x < 0)
+	       return (-1);
+	  }
+	/* b is previous character; insert combining character ch in *b. */
+	for (i = 0; i < SLSMG_MAX_CHARS_PER_CELL-1; i++)
+	  if (b->combining[i] == SLCURSES_NULLCHAR)
+	    break;
+	if (i < SLSMG_MAX_CHARS_PER_CELL-1)
+	  b->combining[i] = ch;
+	return 0;
      }
 
-   if (p < pmax)
-     *p = (unsigned short) ch | (w->color << 8);
+   /* Remove trailing character(s) from line. */
+   if ((colsneeded > 0) && ((unsigned int)colsneeded <= w->ncols))
+     {
+	int i;
+	for (i = w->ncols - colsneeded; i > 0; i--)
+	  if (line[i].main != SLCURSES_NULLATTR)
+	    break;
+	/* line[w->ncols-colsneeded..w->ncols-1] will be shifted off the end,
+	 * but if i < w->ncols-colsneeded, then line[i..w->ncols-colsneeded-1]
+	 * must be blanked because they are part of a multicolumn character
+	 * which is partially shifted off the end.
+	 */
+	while (i + colsneeded < (int)w->ncols)
+	  SLCURSES_BUILD_CELL(&line[i++], ' ', w->color);
+     }
+
+   /* line[_curx+colsneeded..ncols-1] = line[_curx..ncols-1-colsneeded]; */
+#ifdef HAVE_MEMMOVE
+   memmove(&line[w->_curx+colsneeded], &line[w->_curx], ncols-(w->_curx+colsneeded));
+#else
+   dest = w->ncols - 1;
+   for (src = dest - colsneeded; src >= (int)w->_curx; src--, dest--)
+     line[dest] = line[src];
+#endif
+   if (w->_curx + colsneeded <= w->ncols)
+     SLcurses_placechar (w, ch, colsneeded, w->color);
 
    w->modified = 1;
    return 0;
@@ -969,3 +1318,4 @@ int SLcurses_clearok (SLcurses_Window_Type *w, int bf)
      }
    return 0;
 }
+/* vim:set noautoindent cinoptions=>5,n-3,{2,^-2,+4 cindent ruler: */
