@@ -1,9 +1,23 @@
-/* Copyright (c) 1992, 1999, 2001 John E. Davis
- * This file is part of the S-Lang library.
- *
- * You may distribute under the terms of either the GNU General Public
- * License or the Perl Artistic License.
- */
+/*
+Copyright (C) 2004, 2005 John E. Davis
+
+This file is part of the S-Lang Library.
+
+The S-Lang Library is free software; you can redistribute it and/or
+modify it under the terms of the GNU General Public License as
+published by the Free Software Foundation; either version 2 of the
+License, or (at your option) any later version.
+
+The S-Lang Library is distributed in the hope that it will be useful,
+but WITHOUT ANY WARRANTY; without even the implied warranty of
+MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+General Public License for more details.
+
+You should have received a copy of the GNU General Public License
+along with this library; if not, write to the Free Software
+Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
+USA.  
+*/
 
 #include "slinclud.h"
 
@@ -415,11 +429,11 @@ void SLang_reset_tty (void)
 }
 
 /*----------------------------------------------------------------------*\
- *  Function:	int _SLsys_input_pending (int tsecs);
+ *  Function:	int _pSLsys_input_pending (int tsecs);
  *
  *  sleep for *tsecs tenths of a sec waiting for input
 \*----------------------------------------------------------------------*/
-int _SLsys_input_pending (int tsecs)
+int _pSLsys_input_pending (int tsecs)
 {
    if (keyWaiting()) return 1;
 
@@ -444,7 +458,7 @@ int _SLsys_input_pending (int tsecs)
 }
 
 /*----------------------------------------------------------------------*\
- *  Function:	unsigned int _SLsys_getkey (void);
+ *  Function:	unsigned int _pSLsys_getkey (void);
  *
  * Wait for and get the next available mouse-event / keystroke.
  * Also re-maps some useful keystrokes.
@@ -453,12 +467,12 @@ int _SLsys_input_pending (int tsecs)
  *	Ctrl-Space	=>	^@	(^@^3 - a pc NUL char)
  *	extended keys are prefixed by a null character
 \*----------------------------------------------------------------------*/
-unsigned int _SLsys_getkey (void)
+unsigned int _pSLsys_getkey (void)
 {
    unsigned int key, scan, ch, shift;
 
    while ((SLang_Input_Buffer_Len == 0)
-	  && (0 == _SLsys_input_pending (300)))
+	  && (0 == _pSLsys_input_pending (300)))
      ;
 
    if (SLang_Input_Buffer_Len)
@@ -482,7 +496,7 @@ unsigned int _SLsys_getkey (void)
 	/* drop */
       case 0xe0:
       case 0:			/* extended key code */
-	ch = _SLpc_convert_scancode (scan, 0, 1);
+	ch = _pSLpc_convert_scancode (scan, 0, 1);
      }
    return (ch);
 }
