@@ -1,6 +1,6 @@
 /* error handling common to all routines. */
 /*
-Copyright (C) 2004, 2005, 2006, 2007 John E. Davis
+Copyright (C) 2004, 2005 John E. Davis
 
 This file is part of the S-Lang Library.
 
@@ -379,19 +379,19 @@ typedef struct _Error_Message_Type
    int msg_type;
 #define _SLERR_MSG_ERROR	1
 #define _SLERR_MSG_WARNING	2
-#define _SLERR_MSG_TRACEBACK	3
+#define _SLERR_MSG_TRACEBACK	4
    struct _Error_Message_Type *next;
 }
 Error_Message_Type;
 
-struct _pSLerr_Error_Queue_Type
+typedef struct
 {
    Error_Message_Type *head;
    Error_Message_Type *tail;
-};
+}
+Error_Queue_Type;
 
-static _pSLerr_Error_Queue_Type *Default_Error_Queue;
-static _pSLerr_Error_Queue_Type *Active_Error_Queue;
+static Error_Queue_Type *Default_Error_Queue;
 
 static void free_error_msg (Error_Message_Type *m)
 {
@@ -418,7 +418,7 @@ static Error_Message_Type *allocate_error_msg (char *msg, int msg_type)
    return m;
 }
 
-static void free_queued_messages (_pSLerr_Error_Queue_Type *q)
+static void free_queued_messages (Error_Queue_Type *q)
 {
    Error_Message_Type *m;
 
@@ -436,7 +436,7 @@ static void free_queued_messages (_pSLerr_Error_Queue_Type *q)
    q->tail = NULL;
 }
 
-void _pSLerr_delete_error_queue (_pSLerr_Error_Queue_Type *q)
+static void delete_msg_queue (Error_Queue_Type *q)
 {
    if (q == NULL)
      return;
@@ -446,19 +446,17 @@ void _pSLerr_delete_error_queue (_pSLerr_Error_Queue_Type *q)
 }
 
    
-_pSLerr_Error_Queue_Type *_pSLerr_new_error_queue (int make_active)
+static Error_Queue_Type *create_msg_queue (void)
 {
-   _pSLerr_Error_Queue_Type *q;
+   Error_Queue_Type *q;
 
-   if (NULL == (q = (_pSLerr_Error_Queue_Type *)SLcalloc (1, sizeof(_pSLerr_Error_Queue_Type))))
+   if (NULL == (q = (Error_Queue_Type *)SLcalloc (1, sizeof(Error_Queue_Type))))
      return NULL;
    
-   if (make_active)
-     Active_Error_Queue = q;
    return q;
 }
 
-static int queue_message (_pSLerr_Error_Queue_Type *q, char *msg, int msg_type)
+static int queue_message (Error_Queue_Type *q, char *msg, int msg_type)
 {
    Error_Message_Type *m;
    
@@ -470,7 +468,7 @@ static int queue_message (_pSLerr_Error_Queue_Type *q, char *msg, int msg_type)
    if (q->head == NULL)
      q->head = m;
    q->tail = m;
-
+   
    return 0;
 }
 
@@ -517,9 +515,9 @@ static void print_queue (void)
    if (_pSLang_Error == 0)
      return;
 
-   if (Active_Error_Queue != NULL)
+   if (Default_Error_Queue != NULL)
      {
-	_pSLerr_Error_Queue_Type *q = Active_Error_Queue;
+	Error_Queue_Type *q = Default_Error_Queue;
 	Error_Message_Type *m = q->head;
 	while (m != NULL)
 	  {
@@ -542,14 +540,14 @@ static void print_queue (void)
 /* This function returns a pointer to the first error message in the queue.
  * Make no attempts to free the returned pointer.
  */
-char *_pSLerr_get_error_from_queue (_pSLerr_Error_Queue_Type *q)
+char *_pSLerr_get_error_from_queue (void)
 {
+   Error_Queue_Type *q;
    Error_Message_Type *m;
    unsigned int len;
    char *err, *err1, *err_max;
 
-   if ((q == NULL) 
-       && (NULL == (q = Default_Error_Queue)))
+   if (NULL == (q = Default_Error_Queue))
      return NULL;
 
    len = 0;
@@ -588,7 +586,6 @@ char *_pSLerr_get_error_from_queue (_pSLerr_Error_Queue_Type *q)
    return _pSLcreate_via_alloced_slstring (err, len);
 }
 
-
 void _pSLerr_print_message_queue (void)
 {
    print_queue ();
@@ -615,12 +612,13 @@ int _pSLerr_suspend_messages (void)
 
 void _pSLerr_free_queued_messages (void)
 {
-   free_queued_messages (Active_Error_Queue);
+   free_queued_messages (Default_Error_Queue);
 }
 
 
-void SLang_verror_va (int err_code, char *fmt, va_list ap)
+void SLang_verror (int err_code, char *fmt, ...)
 {
+   va_list ap;
    char err [4096];
 
    if (-1 == _pSLerr_init ())
@@ -638,23 +636,16 @@ void SLang_verror_va (int err_code, char *fmt, va_list ap)
    if (fmt == NULL)
      return;
 
+   va_start(ap, fmt);
    (void) SLvsnprintf (err, sizeof (err), fmt, ap);
+   va_end(ap);
 
    if (Suspend_Error_Messages)
-     (void) queue_message (Active_Error_Queue, err, _SLERR_MSG_ERROR);
+     (void) queue_message (Default_Error_Queue, err, _SLERR_MSG_ERROR);
    else
      print_error (_SLERR_MSG_ERROR, err);
 }
 
-void SLang_verror (int err_code, char *fmt, ...)
-{
-   va_list ap;
-
-   va_start(ap, fmt);
-   SLang_verror_va (err_code, fmt, ap);
-   va_end(ap);
-}
-
 int _pSLerr_traceback_msg (char *fmt, ...)
 {
    va_list ap;
@@ -664,7 +655,7 @@ int _pSLerr_traceback_msg (char *fmt, ...)
    (void) SLvsnprintf (msg, sizeof (msg), fmt, ap);
    va_end(ap);
 
-   return queue_message (Active_Error_Queue, msg, _SLERR_MSG_TRACEBACK);
+   return queue_message (Default_Error_Queue, msg, _SLERR_MSG_TRACEBACK);
 }
 
 void SLang_exit_error (char *fmt, ...)
@@ -725,7 +716,6 @@ void SLang_vmessage (char *fmt, ...)
      {
 	vfprintf (stdout, fmt, ap);
 	fputs ("\n", stdout);
-	(void) fflush (stdout);
      }
 
    va_end (ap);
@@ -751,24 +741,12 @@ void _pSLerr_dump_msg (char *fmt, ...)
    va_end (ap);
 }
 
-int _pSLerr_set_error_queue (_pSLerr_Error_Queue_Type *q)
-{
-   if (q == NULL)
-     {
-	q = Default_Error_Queue;
-	if (Default_Error_Queue == NULL)
-	  return _pSLerr_init ();
-     }
-   Active_Error_Queue = q;
-   return 0;
-}
-
 int _pSLerr_init (void)
 {
    if (Default_Error_Queue == NULL)
      {
 	Suspend_Error_Messages = 0;
-	if (NULL == (Default_Error_Queue = _pSLerr_new_error_queue (1)))
+	if (NULL == (Default_Error_Queue = create_msg_queue ()))
 	  return -1;
      }
 
@@ -781,9 +759,8 @@ int _pSLerr_init (void)
 void _pSLerr_deinit (void)
 {
    deinit_exceptions ();
-   _pSLerr_delete_error_queue (Default_Error_Queue);
+   delete_msg_queue (Default_Error_Queue);
    Suspend_Error_Messages = 0;
    Default_Error_Queue = NULL;
-   Active_Error_Queue = NULL;
 }
 
