@@ -1,6 +1,6 @@
 /* sin, cos, etc, for S-Lang */
 /*
-Copyright (C) 2004, 2005, 2006, 2007 John E. Davis
+Copyright (C) 2004, 2005 John E. Davis
 
 This file is part of the S-Lang Library.
 
@@ -216,6 +216,7 @@ static int double_math_op (int op,
 {
    double *a, *b;
    unsigned int i;
+   double (*fun) (double);
    char *c;
 
    (void) type;
@@ -228,52 +229,53 @@ static int double_math_op (int op,
 	return 0;
 
       case SLMATH_SINH:
-	for (i = 0; i < na; i++) b[i] = sinh(a[i]);
+	fun = sinh;
 	break;
       case SLMATH_COSH:
-	for (i = 0; i < na; i++) b[i] = cosh(a[i]);
+	fun = cosh;
 	break;
       case SLMATH_TANH:
-	for (i = 0; i < na; i++) b[i] = tanh(a[i]);
+	fun = tanh;
 	break;
       case SLMATH_TAN:
-	for (i = 0; i < na; i++) b[i] = tan(a[i]);
+	fun = tan;
 	break;
       case SLMATH_ASIN:
-	for (i = 0; i < na; i++) b[i] = asin(a[i]);
+	fun = asin;
 	break;
       case SLMATH_ACOS:
-	for (i = 0; i < na; i++) b[i] = acos(a[i]);
+	fun = acos;
 	break;
       case SLMATH_ATAN:
-	for (i = 0; i < na; i++) b[i] = atan(a[i]);
+	fun = atan;
 	break;
       case SLMATH_EXP:
-	for (i = 0; i < na; i++) b[i] = exp(a[i]);
+	fun = exp;
 	break;
       case SLMATH_LOG:
-	for (i = 0; i < na; i++) b[i] = log(a[i]);
+	fun = log;
 	break;
       case SLMATH_LOG10:
-	for (i = 0; i < na; i++) b[i] = log10(a[i]);
+	fun = log10;
 	break;
       case SLMATH_SQRT:
-	for (i = 0; i < na; i++) b[i] = sqrt (a[i]);
+	for (i = 0; i < na; i++)
+	  b[i] = sqrt ((double) a[i]);
 	return 1;
       case SLMATH_SIN:
-	for (i = 0; i < na; i++) b[i] = sin(a[i]);
+	fun = sin;
 	break;
       case SLMATH_COS:
-	for (i = 0; i < na; i++) b[i] = cos(a[i]);
+	fun = cos;
 	break;
       case SLMATH_ASINH:
-	for (i = 0; i < na; i++) b[i] = ASINH_FUN(a[i]);
+	fun = ASINH_FUN;
 	break;
       case SLMATH_ATANH:
-	for (i = 0; i < na; i++) b[i] = ATANH_FUN(a[i]);
+	fun = ATANH_FUN;
 	break;
       case SLMATH_ACOSH:
-	for (i = 0; i < na; i++) b[i] = ACOSH_FUN(a[i]);
+	fun = ACOSH_FUN;
 	break;
 
       case SLMATH_CONJ:
@@ -299,18 +301,21 @@ static int double_math_op (int op,
 	return 1;
 	
       case SLMATH_FLOOR:
-	for (i = 0; i < na; i++) b[i] = floor(a[i]);
+	fun = floor;
 	break;
 
       case SLMATH_CEIL:
-	for (i = 0; i < na; i++) b[i] = ceil(a[i]);
+	fun = ceil;
 	break;
 	
       case SLMATH_ROUND:
-	for (i = 0; i < na; i++) b[i] = ROUND_FUN(a[i]);
+	fun = ROUND_FUN;
 	break;
      }
-   
+
+   for (i = 0; i < na; i++)
+     b[i] = (*fun) (a[i]);
+
    return 1;
 }
 
@@ -320,6 +325,7 @@ static int float_math_op (int op,
 {
    float *a, *b;
    unsigned int i;
+   double (*fun) (double);
    char *c;
 
    (void) type;
@@ -333,34 +339,34 @@ static int float_math_op (int op,
 	return 0;
 
       case SLMATH_SINH:
-	for (i = 0; i < na; i++) b[i] = sinh(a[i]);
+	fun = sinh;
 	break;
       case SLMATH_COSH:
-	for (i = 0; i < na; i++) b[i] = cosh(a[i]);
+	fun = cosh;
 	break;
       case SLMATH_TANH:
-	for (i = 0; i < na; i++) b[i] = tanh(a[i]);
+	fun = tanh;
 	break;
       case SLMATH_TAN:
-	for (i = 0; i < na; i++) b[i] = tan(a[i]);
+	fun = tan;
 	break;
       case SLMATH_ASIN:
-	for (i = 0; i < na; i++) b[i] = asin(a[i]);
+	fun = asin;
 	break;
       case SLMATH_ACOS:
-	for (i = 0; i < na; i++) b[i] = acos(a[i]);
+	fun = acos;
 	break;
       case SLMATH_ATAN:
-	for (i = 0; i < na; i++) b[i] = atan(a[i]);
+	fun = atan;
 	break;
       case SLMATH_EXP:
-	for (i = 0; i < na; i++) b[i] = exp(a[i]);
+	fun = exp;
 	break;
       case SLMATH_LOG:
-	for (i = 0; i < na; i++) b[i] = log(a[i]);
+	fun = log;
 	break;
       case SLMATH_LOG10:
-	for (i = 0; i < na; i++) b[i] = log10(a[i]);
+	fun = log10;
 	break;
       case SLMATH_SQRT:
 	for (i = 0; i < na; i++)
@@ -368,20 +374,20 @@ static int float_math_op (int op,
 	return 1;
 
       case SLMATH_SIN:
-	for (i = 0; i < na; i++) b[i] = sin(a[i]);
+	fun = sin;
 	break;
       case SLMATH_COS:
-	for (i = 0; i < na; i++) b[i] = cos(a[i]);
+	fun = cos;
 	break;
 
       case SLMATH_ASINH:
-	for (i = 0; i < na; i++) b[i] = ASINH_FUN(a[i]);
+	fun = ASINH_FUN;
 	break;
       case SLMATH_ATANH:
-	for (i = 0; i < na; i++) b[i] = ATANH_FUN(a[i]);
+	fun = ATANH_FUN;
 	break;
       case SLMATH_ACOSH:
-	for (i = 0; i < na; i++) b[i] = ACOSH_FUN(a[i]);
+	fun = ACOSH_FUN;
 	break;
 
       case SLMATH_CONJ:
@@ -406,16 +412,19 @@ static int float_math_op (int op,
 	return 1;
 
       case SLMATH_FLOOR:
-	for (i = 0; i < na; i++) b[i] = floor(a[i]);
+	fun = floor;
 	break;
       case SLMATH_CEIL:
-	for (i = 0; i < na; i++) b[i] = ceil(a[i]);
+	fun = ceil;
 	break;
       case SLMATH_ROUND:
-	for (i = 0; i < na; i++) b[i] = ROUND_FUN(a[i]);
+	fun = ROUND_FUN;
 	break;
      }
 
+   for (i = 0; i < na; i++)
+     b[i] = (float) (*fun) ((double) a[i]);
+
    return 1;
 }
 
@@ -426,6 +435,7 @@ static int generic_math_op (int op,
    double *b;
    unsigned int i;
    SLang_To_Double_Fun_Type to_double;
+   double (*fun) (double);
    unsigned int da;
    char *a, *c;
 
@@ -441,117 +451,53 @@ static int generic_math_op (int op,
 	return 0;
 
       case SLMATH_SINH:
-	for (i = 0; i < na; i++) 
-	  {
-	     b[i] = sinh (to_double ((VOID_STAR) a));
-	     a += da;
-	  }
+	fun = sinh;
 	break;
       case SLMATH_COSH:
-	for (i = 0; i < na; i++) 
-	  {
-	     b[i] = cosh (to_double ((VOID_STAR) a));
-	     a += da;
-	  }
+	fun = cosh;
 	break;
       case SLMATH_TANH:
-	for (i = 0; i < na; i++) 
-	  {
-	     b[i] = tanh (to_double ((VOID_STAR) a));
-	     a += da;
-	  }
+	fun = tanh;
 	break;
       case SLMATH_TAN:
-	for (i = 0; i < na; i++) 
-	  {
-	     b[i] = tan (to_double ((VOID_STAR) a));
-	     a += da;
-	  }
+	fun = tan;
 	break;
       case SLMATH_ASIN:
-	for (i = 0; i < na; i++) 
-	  {
-	     b[i] = asin (to_double ((VOID_STAR) a));
-	     a += da;
-	  }
+	fun = asin;
 	break;
       case SLMATH_ACOS:
-	for (i = 0; i < na; i++) 
-	  {
-	     b[i] = acos (to_double ((VOID_STAR) a));
-	     a += da;
-	  }
+	fun = acos;
 	break;
       case SLMATH_ATAN:
-	for (i = 0; i < na; i++) 
-	  {
-	     b[i] = atan (to_double ((VOID_STAR) a));
-	     a += da;
-	  }
+	fun = atan;
 	break;
       case SLMATH_EXP:
-	for (i = 0; i < na; i++) 
-	  {
-	     b[i] = exp (to_double ((VOID_STAR) a));
-	     a += da;
-	  }
+	fun = exp;
 	break;
       case SLMATH_LOG:
-	for (i = 0; i < na; i++) 
-	  {
-	     b[i] = log (to_double ((VOID_STAR) a));
-	     a += da;
-	  }
+	fun = log;
 	break;
       case SLMATH_LOG10:
-	for (i = 0; i < na; i++) 
-	  {
-	     b[i] = log10 (to_double ((VOID_STAR) a));
-	     a += da;
-	  }
+	fun = log10;
 	break;
       case SLMATH_SQRT:
-	for (i = 0; i < na; i++) 
-	  {
-	     b[i] = sqrt (to_double ((VOID_STAR) a));
-	     a += da;
-	  }
+	fun = sqrt;
 	break;
       case SLMATH_SIN:
-	for (i = 0; i < na; i++) 
-	  {
-	     b[i] = sin (to_double ((VOID_STAR) a));
-	     a += da;
-	  }
+	fun = sin;
 	break;
       case SLMATH_COS:
-	for (i = 0; i < na; i++) 
-	  {
-	     b[i] = cos (to_double ((VOID_STAR) a));
-	     a += da;
-	  }
+	fun = cos;
 	break;
 
       case SLMATH_ASINH:
-	for (i = 0; i < na; i++) 
-	  {
-	     b[i] = ASINH_FUN (to_double ((VOID_STAR) a));
-	     a += da;
-	  }
+	fun = ASINH_FUN;
 	break;
       case SLMATH_ATANH:
-	for (i = 0; i < na; i++) 
-	  {
-	     b[i] = ATANH_FUN (to_double ((VOID_STAR) a));
-	     a += da;
-	  }
+	fun = ATANH_FUN;
 	break;
       case SLMATH_ACOSH:
-	for (i = 0; i < na; i++) 
-	  {
-	     b[i] = ACOSH_FUN (to_double ((VOID_STAR) a));
-	     a += da;
-	  }
+	fun = ACOSH_FUN;
 	break;
 
 
@@ -586,27 +532,21 @@ static int generic_math_op (int op,
 	  }
 	return 1;
       case SLMATH_FLOOR:
-	for (i = 0; i < na; i++) 
-	  {
-	     b[i] = floor (to_double ((VOID_STAR) a));
-	     a += da;
-	  }
+	fun = floor;
 	break;
       case SLMATH_CEIL:
-	for (i = 0; i < na; i++) 
-	  {
-	     b[i] = ceil (to_double ((VOID_STAR) a));
-	     a += da;
-	  }
+	fun = ceil;
 	break;
       case SLMATH_ROUND:
-	for (i = 0; i < na; i++) 
-	  {
-	     b[i] = ROUND_FUN (to_double ((VOID_STAR) a));
-	     a += da;
-	  }
+	fun = ROUND_FUN;
 	break;
      }
+
+   for (i = 0; i < na; i++)
+     {
+	b[i] = (*fun) (to_double ((VOID_STAR) a));
+	a += da;
+     }
    
    return 1;
 }
@@ -749,103 +689,113 @@ static int complex_math_op (int op,
 }
 #endif
 
-typedef struct
+static int do_dd_fun (double (*f)(double, double),
+		      double *a, unsigned int a_inc,
+		      double *b, unsigned int b_inc, 
+		      double *c, unsigned int n)
 {
-   SLang_Array_Type *at;
-   int is_float;		       /* if non-zero, use fptr */
-   float f;
-   double d;
-   char c;
-   float *fptr;			       /* points at at->data or &f */
-   double *dptr;
-   char *cptr;
-   unsigned int inc;		       /* inc = 0, if at==NULL */
-   unsigned int num;
-}
-Array_Or_Scalar_Type;
-
-
-static void free_array_or_scalar (Array_Or_Scalar_Type *ast)
-{
-   if (ast->at != NULL)
-     SLang_free_array (ast->at);
+   unsigned int i;
+   for (i = 0; i < n; i++)
+     {
+	c[i] = (*f)(*a, *b);
+	a += a_inc;
+	b += b_inc;
+     }
+   return 0;
 }
 
-static int pop_array_or_scalar (Array_Or_Scalar_Type *ast)
+static int do_ff_fun (double (*f)(double, double),
+		      float *a, unsigned int a_inc,
+		      float *b, unsigned int b_inc, 
+		      float *c, unsigned int n)
+{
+   unsigned int i;
+   for (i = 0; i < n; i++)
+     {
+	c[i] = (float) (*f)(*a, *b);
+	a += a_inc;
+	b += b_inc;
+     }
+   return 0;
+}
+
+static int do_fd_fun (double (*f)(double, double),
+		      float *a, unsigned int a_inc,
+		      double *b, unsigned int b_inc, 
+		      double *c, unsigned int n)
+{
+   unsigned int i;
+   for (i = 0; i < n; i++)
+     {
+	c[i] = (*f)(*a, *b);
+	a += a_inc;
+	b += b_inc;
+     }
+   return 0;
+}
+
+static int do_df_fun (double (*f)(double, double),
+		      double *a, unsigned int a_inc,
+		      float *b, unsigned int b_inc, 
+		      double *c, unsigned int n)
+{
+   unsigned int i;
+   for (i = 0; i < n; i++)
+     {
+	c[i] = (*f)(*a, *b);
+	a += a_inc;
+	b += b_inc;
+     }
+   return 0;
+}
+
+static int pop_array_or_scalar (SLang_Array_Type **atp, 
+				float *f, float **fp, 
+				double *d, double **dp,
+				int *is_floatp)
 {
    SLang_Array_Type *at;
 
-   ast->at = NULL;
-   ast->inc = 0;
-   ast->num = 1;
+   *atp = NULL;
    switch (SLang_peek_at_stack1 ())
      {
       case -1:
 	return -1;
 
       case SLANG_FLOAT_TYPE:
-	ast->is_float = 1;
+	*is_floatp = 1;
 	if (SLang_peek_at_stack () == SLANG_ARRAY_TYPE)
 	  {
 	     if (-1 == SLang_pop_array_of_type (&at, SLANG_FLOAT_TYPE))
 	       return -1;
-	     ast->fptr = (float *) at->data;
-	     ast->inc = 1;
-	     ast->num = at->num_elements;
-	     ast->at = at;
+	     *fp = (float *) at->data;
+	     *atp = at;
 	     return 0;
 	  }
 
-	ast->fptr = &ast->f;
-	if (-1 == SLang_pop_float (ast->fptr))
+	if (-1 == SLang_pop_float (f))
 	  return -1;
+	*fp = f;
 	return 0;
 
       default:
-	ast->is_float = 0;
+	*is_floatp = 0;
 	if (SLang_peek_at_stack () == SLANG_ARRAY_TYPE)
 	  {
 	     if (-1 == SLang_pop_array_of_type (&at, SLANG_DOUBLE_TYPE))
 	       return -1;
-	     ast->dptr = (double *) at->data;
-	     ast->inc = 1;
-	     ast->num = at->num_elements;
-	     ast->at = at;
+	     *dp = (double *) at->data;
+	     *atp = at;
 	     return 0;
 	  }
 
-	ast->dptr = &ast->d;
-	if (-1 == SLang_pop_double (ast->dptr))
+	if (-1 == SLang_pop_double (d))
 	  return -1;
+	*dp = d;
 	return 0;
      }
 }
 
-static int pop_2_arrays_or_scalar (Array_Or_Scalar_Type *a, Array_Or_Scalar_Type *b)
-{
-   if (-1 == pop_array_or_scalar (b))
-     return -1;
-
-   if (-1 == pop_array_or_scalar (a))
-     {
-	free_array_or_scalar (b);
-	return -1;
-     }
-
-   if ((a->at != NULL) && (b->at != NULL))
-     {
-	if (a->num != b->num)
-	  {
-	     SLang_verror (SL_TypeMismatch_Error, "Array sizes do not match");
-	     free_array_or_scalar (a);
-	     free_array_or_scalar (b);
-	     return -1;
-	  }
-     }
-   return 0;
-}
-
-
 static SLang_Array_Type *create_from_tmp_array (SLang_Array_Type *a, SLang_Array_Type *b, SLtype type)
 {
    SLang_Array_Type *c;
@@ -869,149 +819,100 @@ static SLang_Array_Type *create_from_tmp_array (SLang_Array_Type *a, SLang_Array
    return SLang_create_array1 (type, 0, NULL, c->dims, c->num_dims, 1);
 }
 
-static int do_dd_fun (double (*f)(double, double),
-		      Array_Or_Scalar_Type *a_ast,
-		      Array_Or_Scalar_Type *b_ast,
-		      Array_Or_Scalar_Type *c_ast)
-{
-   double *a = a_ast->dptr;
-   double *b = b_ast->dptr;
-   double *c = c_ast->dptr;
-   unsigned int a_inc = a_ast->inc;
-   unsigned int b_inc = b_ast->inc;
-   unsigned int i, n = c_ast->num;
-
-   for (i = 0; i < n; i++)
-     {
-	c[i] = (*f)(*a, *b);
-	a += a_inc;
-	b += b_inc;
-     }
-   return 0;
-}
-
-static int do_ff_fun (double (*f)(double, double),
-		      Array_Or_Scalar_Type *a_ast,
-		      Array_Or_Scalar_Type *b_ast,
-		      Array_Or_Scalar_Type *c_ast)
-{
-   float *a = a_ast->fptr;
-   float *b = b_ast->fptr;
-   float *c = c_ast->fptr;
-   unsigned int a_inc = a_ast->inc;
-   unsigned int b_inc = b_ast->inc;
-   unsigned int i, n = c_ast->num;
-
-   for (i = 0; i < n; i++)
-     {
-	c[i] = (float) (*f)(*a, *b);
-	a += a_inc;
-	b += b_inc;
-     }
-   return 0;
-}
-
-static int do_fd_fun (double (*f)(double, double),
-		      Array_Or_Scalar_Type *a_ast,
-		      Array_Or_Scalar_Type *b_ast,
-		      Array_Or_Scalar_Type *c_ast)
-{
-   float *a = a_ast->fptr;
-   double *b = b_ast->dptr;
-   double *c = c_ast->dptr;
-   unsigned int a_inc = a_ast->inc;
-   unsigned int b_inc = b_ast->inc;
-   unsigned int i, n = c_ast->num;
-
-   for (i = 0; i < n; i++)
-     {
-	c[i] = (*f)(*a, *b);
-	a += a_inc;
-	b += b_inc;
-     }
-   return 0;
-}
-
-static int do_df_fun (double (*f)(double, double),
-		      Array_Or_Scalar_Type *a_ast,
-		      Array_Or_Scalar_Type *b_ast,
-		      Array_Or_Scalar_Type *c_ast)
-{
-   double *a = a_ast->dptr;
-   float *b = b_ast->fptr;
-   double *c = c_ast->dptr;
-   unsigned int a_inc = a_ast->inc;
-   unsigned int b_inc = b_ast->inc;
-   unsigned int i, n = c_ast->num;
-
-   for (i = 0; i < n; i++)
-     {
-	c[i] = (*f)(*a, *b);
-	a += a_inc;
-	b += b_inc;
-     }
-   return 0;
-}
-
 static int do_binary_function (double (*f)(double, double))
 {
+   int a_is_float, b_is_float, c_is_float;
+   SLang_Array_Type *at = NULL;
+   SLang_Array_Type *bt = NULL;
+   SLang_Array_Type *ct;
+   float *af, *bf, *cf, afbuf, bfbuf, cfbuf;
+   double *ad, *bd, *cd, adbuf, bdbuf, cdbuf;
+   unsigned int ainc, binc;
+   unsigned int na, nb, nc;
    SLtype type;
-   Array_Or_Scalar_Type a_ast, b_ast, c_ast;
+   int is_scalar;
 
-   if (-1 == pop_2_arrays_or_scalar (&a_ast, &b_ast))
+   if (-1 == pop_array_or_scalar (&bt, &bfbuf, &bf, &bdbuf, &bd, &b_is_float))
      return -1;
 
-   c_ast.is_float = (a_ast.is_float && b_ast.is_float);
-   c_ast.at = NULL;
-   c_ast.num = 1;
-   c_ast.inc = 0;
-   if (c_ast.is_float)
+   if (-1 == pop_array_or_scalar (&at, &afbuf, &af, &adbuf, &ad, &a_is_float))
      {
-	type = SLANG_FLOAT_TYPE;
-	c_ast.fptr = &c_ast.f;
-     }
-   else
-     {
-	type = SLANG_DOUBLE_TYPE;
-	c_ast.dptr = &c_ast.d;
+	if (bt != NULL)
+	  SLang_free_array (bt);
+	return -1;
      }
 
-   if ((a_ast.at != NULL) || (b_ast.at != NULL))
+   c_is_float = (a_is_float && b_is_float);
+   if (c_is_float) 
+     type = SLANG_FLOAT_TYPE;
+   else
+     type = SLANG_DOUBLE_TYPE;
+
+   is_scalar = 1;
+   ainc = binc = 0;
+   na = nb = nc = 1;
+
+   if (at != NULL)
      {
-	if (NULL == (c_ast.at = create_from_tmp_array (a_ast.at, b_ast.at, type)))
+	ainc = 1;
+	na = at->num_elements;
+	is_scalar = 0;
+     }
+
+   if (bt != NULL)
+     {
+	binc = 1;
+	nb = bt->num_elements;
+	is_scalar = 0;
+	if ((at != NULL) && (nb != na))
 	  {
-	     free_array_or_scalar (&a_ast);
-	     free_array_or_scalar (&b_ast);
+	     SLang_verror (SL_TypeMismatch_Error, "Array sizes do not match");
+	     SLang_free_array (at);
+	     SLang_free_array (bt);
 	     return -1;
 	  }
-	c_ast.fptr = (float *) c_ast.at->data;
-	c_ast.dptr = (double *) c_ast.at->data;
-	c_ast.num = c_ast.at->num_elements;
-	c_ast.inc = 1;
      }
 
-   if (a_ast.is_float)
+   if (is_scalar == 0)
      {
-	if (b_ast.is_float)
-	  (void) do_ff_fun (f, &a_ast, &b_ast, &c_ast);
-	else
-	  (void) do_fd_fun (f, &a_ast, &b_ast, &c_ast);
+	if (NULL == (ct = create_from_tmp_array (at, bt, type)))
+	  {
+	     SLang_free_array (at);    /* NULL ok */
+	     SLang_free_array (bt);    /* NULL ok */
+	     return -1;
+	  }
+	cf = (float *) ct->data;
+	cd = (double *) ct->data;
+	nc = ct->num_elements;
      }
-   else if (b_ast.is_float)
-     (void) do_df_fun (f, &a_ast, &b_ast, &c_ast);
    else
-     (void) do_dd_fun (f, &a_ast, &b_ast, &c_ast);
+     {
+	cf = (float *) &cfbuf;
+	cd = (double *) &cdbuf;
+	ct = NULL;
+     }
+
+   if (a_is_float)
+     {
+	if (b_is_float)
+	  (void) do_ff_fun (f, af, ainc, bf, binc, cf, nc);
+	else
+	  (void) do_fd_fun (f, af, ainc, bd, binc, cd, nc);
+     }
+   else if (b_is_float)
+     (void) do_df_fun (f, ad, ainc, bf, binc, cd, nc);
+   else
+     (void) do_dd_fun (f, ad, ainc, bd, binc, cd, nc);
    
-   free_array_or_scalar (&a_ast);
-   free_array_or_scalar (&b_ast);
+   SLang_free_array (at);
+   SLang_free_array (bt);
+   if (ct != NULL)
+     return SLang_push_array (ct, 1);
 
-   if (c_ast.at != NULL)
-     return SLang_push_array (c_ast.at, 1);
+   if (c_is_float)
+     return SLang_push_float (cfbuf);
 
-   if (c_ast.is_float)
-     return SLang_push_float (c_ast.f);
-
-   return SLang_push_double (c_ast.d);
+   return SLang_push_double (cdbuf);
 }
 
 static void hypot_fun (void)
@@ -1052,331 +953,6 @@ static void diff_fun (void)
 {
    (void) do_binary_function (do_diff);
 }
-
-#if 1
-
-static int do_c_dd_fun (int (*f)(double, double, VOID_STAR),
-			VOID_STAR cd,
-			Array_Or_Scalar_Type *a_ast,
-			Array_Or_Scalar_Type *b_ast,
-			Array_Or_Scalar_Type *c_ast)
-{
-   double *a = a_ast->dptr;
-   double *b = b_ast->dptr;
-   char *c = c_ast->cptr;
-   unsigned int a_inc = a_ast->inc;
-   unsigned int b_inc = b_ast->inc;
-   unsigned int i, n = c_ast->num;
-
-   for (i = 0; i < n; i++)
-     {
-	c[i] = (char) (*f)(*a, *b, cd);
-	a += a_inc;
-	b += b_inc;
-     }
-   return 0;
-}
-
-static int do_c_ff_fun (int (*f)(double, double, VOID_STAR),
-			VOID_STAR cd,
-			Array_Or_Scalar_Type *a_ast,
-			Array_Or_Scalar_Type *b_ast,
-			Array_Or_Scalar_Type *c_ast)
-{
-   float *a = a_ast->fptr;
-   float *b = b_ast->fptr;
-   char *c = c_ast->cptr;
-   unsigned int a_inc = a_ast->inc;
-   unsigned int b_inc = b_ast->inc;
-   unsigned int i, n = c_ast->num;
-
-   for (i = 0; i < n; i++)
-     {
-	c[i] = (char) (*f)(*a, *b, cd);
-	a += a_inc;
-	b += b_inc;
-     }
-   return 0;
-}
-
-static int do_c_fd_fun (int (*f)(double, double, VOID_STAR),
-			VOID_STAR cd,
-			Array_Or_Scalar_Type *a_ast,
-			Array_Or_Scalar_Type *b_ast,
-			Array_Or_Scalar_Type *c_ast)
-{
-   float *a = a_ast->fptr;
-   double *b = b_ast->dptr;
-   char *c = c_ast->cptr;
-   unsigned int a_inc = a_ast->inc;
-   unsigned int b_inc = b_ast->inc;
-   unsigned int i, n = c_ast->num;
-
-   for (i = 0; i < n; i++)
-     {
-	c[i] = (char) (*f)(*a, *b, cd);
-	a += a_inc;
-	b += b_inc;
-     }
-   return 0;
-}
-
-static int do_c_df_fun (int (*f)(double, double, VOID_STAR),
-			VOID_STAR cd,
-			Array_Or_Scalar_Type *a_ast,
-			Array_Or_Scalar_Type *b_ast,
-			Array_Or_Scalar_Type *c_ast)
-{
-   double *a = a_ast->dptr;
-   float *b = b_ast->fptr;
-   char *c = c_ast->cptr;
-   unsigned int a_inc = a_ast->inc;
-   unsigned int b_inc = b_ast->inc;
-   unsigned int i, n = c_ast->num;
-
-   for (i = 0; i < n; i++)
-     {
-	c[i] = (char) (*f)(*a, *b, cd);
-	a += a_inc;
-	b += b_inc;
-     }
-   return 0;
-}
-
-static int do_binary_function_c (int (*f)(double, double,VOID_STAR), VOID_STAR cd)
-{
-   Array_Or_Scalar_Type a_ast, b_ast, c_ast;
-
-   if (-1 == pop_2_arrays_or_scalar (&a_ast, &b_ast))
-     return -1;
-
-   c_ast.at = NULL;
-   c_ast.num = 1;
-   c_ast.inc = 0;
-   c_ast.cptr = &c_ast.c;
-
-   if ((a_ast.at != NULL) || (b_ast.at != NULL))
-     {
-	if (NULL == (c_ast.at = SLang_create_array1 (SLANG_CHAR_TYPE, 0, NULL, a_ast.at->dims, a_ast.at->num_dims, 1)))
-	  {
-	     free_array_or_scalar (&a_ast);
-	     free_array_or_scalar (&b_ast);
-	     return -1;
-	  }
-	c_ast.cptr = (char *) c_ast.at->data;
-	c_ast.num = c_ast.at->num_elements;
-	c_ast.inc = 1;
-     }
-
-   if (a_ast.is_float)
-     {
-	if (b_ast.is_float)
-	  (void) do_c_ff_fun (f, cd, &a_ast, &b_ast, &c_ast);
-	else
-	  (void) do_c_fd_fun (f, cd, &a_ast, &b_ast, &c_ast);
-     }
-   else if (b_ast.is_float)
-     (void) do_c_df_fun (f, cd, &a_ast, &b_ast, &c_ast);
-   else
-     (void) do_c_dd_fun (f, cd, &a_ast, &b_ast, &c_ast);
-
-   free_array_or_scalar (&a_ast);
-   free_array_or_scalar (&b_ast);
-
-   if (c_ast.at != NULL)
-     return SLang_push_array (c_ast.at, 1);
-
-   return SLang_push_char (c_ast.c);
-}
-
-typedef struct 
-{
-   double relerr;
-   double abserr;
-}
-Feqs_Err_Type;
-
-static int get_tolorances (int nargs, Feqs_Err_Type *ep)
-{
-   switch (nargs)
-     {
-      case 2:
-	if ((-1 == SLang_pop_double (&ep->abserr))
-	    || (-1 == SLang_pop_double (&ep->relerr)))
-	  return -1;
-	break;
-	
-      case 1:
-	if (-1 == SLang_pop_double (&ep->relerr))
-	  return -1;
-	ep->abserr = 0.0;
-	break;
-	
-      default:
-	ep->relerr = 0.01;
-	ep->abserr = 1e-6;
-	break;
-     }
-   return 0;
-}
-	  
-static int do_feqs (double a, double b, VOID_STAR cd)
-{   
-   Feqs_Err_Type *e = (Feqs_Err_Type *)cd;
-   
-   if (fabs (a-b) <= e->abserr)
-     return 1;
-   
-   if (fabs(a) > fabs(b))
-     return fabs ((a-b)/a) <= e->relerr;
-   
-   return fabs ((b-a)/b) <= e->relerr;
-}
-
-static int do_fneqs (double a, double b, VOID_STAR cd)
-{
-   return !do_feqs (a, b, cd);
-}
-
-static int do_fleqs (double a, double b, VOID_STAR cd)
-{
-   return (a < b) || do_feqs (a, b, cd);
-}
-
-static int do_fgeqs (double a, double b, VOID_STAR cd)
-{
-   return (a > b) || do_feqs (a, b, cd);
-}
-
-static void do_an_feqs_fun (int (*f)(double, double, VOID_STAR))
-{
-   Feqs_Err_Type e;
-   if (-1 == get_tolorances (SLang_Num_Function_Args-2, &e))
-     return;
-   
-   do_binary_function_c (f, (VOID_STAR)&e);
-}
-
-static void feqs_fun (void)
-{
-   do_an_feqs_fun (do_feqs);
-}
-
-static void fgeqs_fun (void)
-{
-   do_an_feqs_fun (do_fgeqs);
-}
-
-static void fleqs_fun (void)
-{
-   do_an_feqs_fun (do_fleqs);
-}
-
-static void fneqs_fun (void)
-{
-   do_an_feqs_fun (do_fneqs);
-}
-#endif
-
-
-static int do_nint (double x)
-{
-   double xf, xi;
-   
-   xf = modf (x, &xi);		       /* x = xi + xf */
-   if (x >= 0)
-     {
-	if (xf >= 0.5)
-	  return xi + 1;
-     }
-   else 
-     {
-	if (xf <= -0.5)
-	  return xi - 1;
-     }
-   return xi;
-}
-
-static int float_to_nint (SLang_Array_Type *at, SLang_Array_Type *bt)
-{
-   unsigned int n, i;
-   int *ip;
-   float *fp;
-   
-   fp = (float *) at->data;
-   ip = (int *) bt->data;
-   n = at->num_elements;
-
-   for (i = 0; i < n; i++)
-     ip[i] = do_nint ((double) fp[i]);
-
-   return 0;
-}
-
-static int double_to_nint (SLang_Array_Type *at, SLang_Array_Type *bt)
-{
-   unsigned int n, i;
-   int *ip;
-   double *dp;
-   
-   dp = (double *) at->data;
-   ip = (int *) bt->data;
-   n = at->num_elements;
-
-   for (i = 0; i < n; i++)
-     ip[i] = do_nint (dp[i]);
-
-   return 0;
-}
-
-static void nint_intrin (void)
-{
-   double x;
-   SLang_Array_Type *at, *bt;
-   int (*at_to_int_fun)(SLang_Array_Type *, SLang_Array_Type *);
-
-   if (SLang_peek_at_stack () != SLANG_ARRAY_TYPE)
-     {
-	if (-1 == SLang_pop_double (&x))
-	  return;
-	(void) SLang_push_int (do_nint (x));
-	return;
-     }
-   switch (SLang_peek_at_stack1 ())
-     {
-      case -1:
-	return;
-
-      case SLANG_INT_TYPE:
-	return;
-
-      case SLANG_FLOAT_TYPE:
-	if (-1 == SLang_pop_array_of_type (&at, SLANG_FLOAT_TYPE))
-	  return;
-	at_to_int_fun = float_to_nint;
-	break;
-
-      case SLANG_DOUBLE_TYPE:
-      default:
-	if (-1 == SLang_pop_array_of_type (&at, SLANG_DOUBLE_TYPE))
-	  return;
-	at_to_int_fun = double_to_nint;
-	break;
-     }
-   
-   if (NULL == (bt = SLang_create_array1 (SLANG_INT_TYPE, 0, NULL, at->dims, at->num_dims, 1)))
-     {
-	SLang_free_array (at);
-	return;
-     }
-   if (0 == (*at_to_int_fun) (at, bt))
-     (void) SLang_push_array (bt, 0);
-   
-   SLang_free_array (bt);
-   SLang_free_array (at);
-}
-
-
 static void fpu_clear_except_bits (void)
 {
    SLfpu_clear_except_bits ();
@@ -1428,17 +1004,12 @@ static SLang_Math_Unary_Type SLmath_Table [] =
 
 static SLang_Intrin_Fun_Type SLang_Math_Table [] =
 {
-   MAKE_INTRINSIC_0("nint", nint_intrin, SLANG_VOID_TYPE),
    MAKE_INTRINSIC_0("polynom", math_poly, SLANG_DOUBLE_TYPE),
    MAKE_INTRINSIC_0("hypot", hypot_fun, SLANG_VOID_TYPE),
    MAKE_INTRINSIC_0("atan2", atan2_fun, SLANG_VOID_TYPE),
    MAKE_INTRINSIC_0("_min", min_fun, SLANG_VOID_TYPE),
    MAKE_INTRINSIC_0("_max", max_fun, SLANG_VOID_TYPE),
    MAKE_INTRINSIC_0("_diff", diff_fun, SLANG_VOID_TYPE),
-   MAKE_INTRINSIC_0("feqs", feqs_fun, SLANG_VOID_TYPE),
-   MAKE_INTRINSIC_0("fneqs", fneqs_fun, SLANG_VOID_TYPE),
-   MAKE_INTRINSIC_0("flteqs", fleqs_fun, SLANG_VOID_TYPE),
-   MAKE_INTRINSIC_0("fgteqs", fgeqs_fun, SLANG_VOID_TYPE),
    MAKE_INTRINSIC_0("fpu_clear_except_bits", fpu_clear_except_bits, SLANG_VOID_TYPE),
    MAKE_INTRINSIC_1("fpu_test_except_bits", fpu_test_except_bits, _pSLANG_LONG_TYPE, _pSLANG_LONG_TYPE),
    SLANG_END_INTRIN_FUN_TABLE
