@@ -1,19 +1,31 @@
-/* Copyright (c) 1992, 1995 John E. Davis
- * All rights reserved.
- * 
- * You may distribute under the terms of either the GNU General Public
- * License or the Perl Artistic License.
- */
+/*
+Copyright (C) 2004, 2005 John E. Davis
 
+This file is part of the S-Lang Library.
+
+The S-Lang Library is free software; you can redistribute it and/or
+modify it under the terms of the GNU General Public License as
+published by the Free Software Foundation; either version 2 of the
+License, or (at your option) any later version.
+
+The S-Lang Library is distributed in the hope that it will be useful,
+but WITHOUT ANY WARRANTY; without even the implied warranty of
+MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+General Public License for more details.
+
+You should have received a copy of the GNU General Public License
+along with this library; if not, write to the Free Software
+Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
+USA.  
+*/
 
 /* These routines are fast memcpy, memset routines.  When available, I
    use system rouines.  For msdos, I use inline assembly. */
 
 /* The current versions only work in the forward direction only!! */
-      
-#include "config.h"
 
-#include <stdio.h>
+#include "slinclud.h"
+
 #include "slang.h"
 #include "_slang.h"
 
@@ -24,9 +36,9 @@ int SLmemcmp(register char *s1, register char *s2, int n)
 {
    register int cmp;
    register char *s1max;
-   
+
    s1max = s1 + (n - 32);
-	
+
    while (s1 <= s1max)
      {
 	if (*s1 != *s2) return ((unsigned char) *s1 - (unsigned char) *s2);
@@ -63,7 +75,7 @@ int SLmemcmp(register char *s1, register char *s2, int n)
 	if (*(s1 + 31) != *(s2 + 31)) return ((unsigned char) *(s1 + 31) - (unsigned char) *(s2 + 31));
 	s1 += 32; s2 += 32;
      }
-   
+
    s1max = s1 + (n % 32);
 
    while (s1 < s1max)
@@ -73,6 +85,6 @@ int SLmemcmp(register char *s1, register char *s2, int n)
 	s1++;
 	s2++;
      }
-   
+
    return(0);
 }
