@@ -1,5 +1,5 @@
 /*
-Copyright (C) 2004, 2005, 2006, 2007 John E. Davis
+Copyright (C) 2004, 2005 John E. Davis
 
 This file is part of the S-Lang Library.
 
@@ -124,7 +124,7 @@ typedef struct
 }
 Token_List_Type;
 
-#define MAX_TOKEN_LISTS 256
+#define MAX_TOKEN_LISTS 16
 static Token_List_Type Token_List_Stack [MAX_TOKEN_LISTS];
 static unsigned int Token_List_Stack_Depth = 0;
 static Token_List_Type *Token_List = NULL;
@@ -249,36 +249,6 @@ static int append_token (_pSLang_Token_Type *t)
    return 0;
 }
 
-static int append_copy_of_string_token (_pSLang_Token_Type *t)
-{
-   _pSLang_Token_Type *t1;
-
-   if (-1 == check_token_list_space (Token_List, 1))
-     return -1;
-
-   t1 = Token_List->stack + Token_List->len;
-   *t1 = *t;
-   
-   if (NULL == (t1->v.s_val = SLang_create_slstring (t->v.s_val)))
-     return -1;
-   
-   t1->free_sval_flag = 1;
-   t1->num_refs = 1;
-
-   Token_List->len += 1;
-   return 0;
-}
-
-static int append_int_token (int n)
-{     
-   _pSLang_Token_Type num_tok;
-
-   init_token (&num_tok);
-   num_tok.type = INT_TOKEN;
-   num_tok.v.long_val = n;
-   return append_token (&num_tok);
-}
-
 static int append_token_of_type (unsigned char t)
 {
    _pSLang_Token_Type *tok;
@@ -289,23 +259,6 @@ static int append_token_of_type (unsigned char t)
    /* The memset when the list was created ensures that the other fields
     * are properly initialized.
     */
-   if ((t == CHS_TOKEN) && Token_List->len)
-     {
-	tok = Token_List->stack + (Token_List->len-1);
-	if (IS_INTEGER_TOKEN(tok->type))
-	  {
-	     tok->v.long_val = -tok->v.long_val;
-	     return 0;
-	  }
-#ifdef HAVE_LONG_LONG
-	if ((tok->type == LLONG_TOKEN) || (tok->type == ULLONG_TOKEN))
-	  {
-	     tok->v.llong_val = -tok->v.llong_val;
-	     return 0;
-	  }
-#endif
-     }
-
    tok = Token_List->stack + Token_List->len;
    init_token (tok);
    tok->type = t;
@@ -504,10 +457,10 @@ static void handle_for_statement (_pSLang_Token_Type *);
 static void handle_foreach_statement (_pSLang_Token_Type *);
 static void statement_list (_pSLang_Token_Type *);
 static void variable_list (_pSLang_Token_Type *, unsigned char);
-static void struct_declaration (_pSLang_Token_Type *, int);
+static void struct_declaration (_pSLang_Token_Type *);
 static void define_function_args (_pSLang_Token_Type *);
 static void typedef_definition (_pSLang_Token_Type *);
-static void function_args_expression (_pSLang_Token_Type *, int, int);
+static void function_args_expression (_pSLang_Token_Type *, int);
 static void expression (_pSLang_Token_Type *);
 static void expression_with_commas (_pSLang_Token_Type *, int);
 static void simple_expression (_pSLang_Token_Type *);
@@ -601,43 +554,6 @@ static void define_function (_pSLang_Token_Type *ctok, unsigned char type)
    free_token (&fname);
 }
 
-/* This is called from "statement", which is expected to return no
- * new token.
- */
-static void check_for_loop_then_else (_pSLang_Token_Type *ctok)
-{
-   int b = 0;
-   
-   get_token (ctok);
-
-   while (1)
-     {
-	unsigned char type = ctok->type;
-#ifdef LOOP_ELSE_TOKEN
-	if ((type == ELSE_TOKEN) && ((b & 1) == 0))
-	  {
-	     get_token (ctok);
-	     block (ctok);
-	     compile_token_of_type (LOOP_ELSE_TOKEN);
-	     get_token (ctok);
-	     b |= 1;
-	     continue;
-	  }
-#endif
-	if ((type == THEN_TOKEN) && ((b & 2) == 0))
-	  {
-	     get_token (ctok);
-	     block (ctok);
-	     compile_token_of_type (LOOP_THEN_TOKEN);
-	     get_token (ctok);
-	     b |= 2;
-	     continue;
-	  }
-	break;
-     }
-   unget_token (ctok);
-}
-
 /* statement:
  *	 compound-statement
  *	 if ( expression ) statement
@@ -725,7 +641,6 @@ static void statement (_pSLang_Token_Type *ctok)
       case _FOR_TOKEN:
 	get_token (ctok);
 	handle_for_statement (ctok);
-	check_for_loop_then_else (ctok);
 	break;
 
       case LOOP_TOKEN:
@@ -739,13 +654,11 @@ static void statement (_pSLang_Token_Type *ctok)
 #endif
 	block (ctok);
 	compile_token_of_type (LOOP_TOKEN);
-	check_for_loop_then_else (ctok);
 	break;
 
       case FOREACH_TOKEN:
 	get_token (ctok);
 	handle_foreach_statement (ctok);
-	check_for_loop_then_else (ctok);
 	break;
 
       case WHILE_TOKEN:
@@ -761,7 +674,6 @@ static void statement (_pSLang_Token_Type *ctok)
 	compile_token_of_type (CBRACE_TOKEN);
 	block (ctok);
 	compile_token_of_type (WHILE_TOKEN);
-	check_for_loop_then_else (ctok);
 	break;
 
       case DO_TOKEN:
@@ -787,7 +699,6 @@ static void statement (_pSLang_Token_Type *ctok)
 	compile_token_of_type (CBRACE_TOKEN);
 	compile_token_of_type (DOWHILE_TOKEN);
 	handle_semicolon (ctok);
-	check_for_loop_then_else (ctok);
 	break;
 
       case FOR_TOKEN:
@@ -862,7 +773,6 @@ static void statement (_pSLang_Token_Type *ctok)
 	get_token (ctok);
 	block (ctok);
 	compile_token_of_type (FOR_TOKEN);
-	check_for_loop_then_else (ctok);
 	break;
 
       case FOREVER_TOKEN:
@@ -878,8 +788,6 @@ static void statement (_pSLang_Token_Type *ctok)
 	get_token (ctok);
 	block (ctok);
 	compile_token_of_type (type);
-	if (type == FOREVER_TOKEN)
-	  check_for_loop_then_else (ctok);
 	break;
 
       case BREAK_TOKEN:
@@ -1132,7 +1040,7 @@ static int init_identifier_token (_pSLang_Token_Type *t, char *name, int alloc)
  *       exception-list , exception
  * 
  *    finally_block_opt:
- *       FINALLY[:] block
+ *       FINALLY block
  * 
  * The above gets compiled into the form:
  * 
@@ -1227,8 +1135,6 @@ static void handle_try_statement (_pSLang_Token_Type *ctok)
    if (ctok->type == FINALLY_TOKEN)
      {
 	get_token (ctok);
-	if (ctok->type == COLON_TOKEN)
-	  get_token (ctok);
 	block (ctok);
      }
    else
@@ -1245,7 +1151,6 @@ static void handle_try_statement (_pSLang_Token_Type *ctok)
 
 /*
  * throw-statement:
- *   throw ;
  *   throw exception_expr ;
  *   throw exception_expr , simple_expression
  *   throw exception_expr , simple_expression ;
@@ -1259,47 +1164,30 @@ static void handle_throw_statement (_pSLang_Token_Type *ctok)
 
    push_token_list ();
 
-   if (ctok->type == SEMICOLON_TOKEN)
-     append_token_of_type (ARG_TOKEN);
-   else
-     {
 #if SLANG_HAS_BOSEOS
-	eos = append_bos (ctok, 2);
+   eos = append_bos (ctok, 2);
 #endif
-	append_token_of_type (ARG_TOKEN);
+   append_token_of_type (ARG_TOKEN);
+   simple_expression (ctok);
+   if (ctok->type == COMMA_TOKEN)
+     {
+	get_token (ctok);
 	simple_expression (ctok);
-	if (ctok->type == COMMA_TOKEN)
-	  {
-	     get_token (ctok);
-	     simple_expression (ctok);
-	  }
-	if (ctok->type == COMMA_TOKEN)
-	  {
-	     get_token (ctok);
-	     simple_expression (ctok);
-	  }
-	handle_semicolon (ctok);
+     }
+   if (ctok->type == COMMA_TOKEN)
+     {
+	get_token (ctok);
+	simple_expression (ctok);
+     }
+   handle_semicolon (ctok);
    /* not-necessary -- append_token_of_type (EARG_TOKEN); */
 #if SLANG_HAS_BOSEOS
-	if (eos) append_eos ();
+   if (eos) append_eos ();
 #endif
-     }
    compile_token_list ();
    compile_token_of_type (THROW_TOKEN);
 }
 
-static _pSLang_Token_Type *allocate_token (void)
-{
-   _pSLang_Token_Type *v;
-
-   v = (_pSLang_Token_Type *)SLmalloc (sizeof (_pSLang_Token_Type));
-   if (v == NULL)
-     return NULL;
-   
-   init_token (v);
-   return v;
-}
-
 static void handle_foreach_statement (_pSLang_Token_Type *ctok)
 {
    _pSLang_Token_Type *var_tokens = NULL;
@@ -1345,7 +1233,7 @@ static void handle_foreach_statement (_pSLang_Token_Type *ctok)
 	     goto free_return;
 	  }
 	get_token (ctok);
-	function_args_expression (ctok, 0, 0);
+	function_args_expression (ctok, 0);
      }
    append_token_of_type (EARG_TOKEN);
    
@@ -1554,176 +1442,53 @@ static void variable_list (_pSLang_Token_Type *name_token, unsigned char variabl
    if (declaring) compile_token_of_type (CBRACKET_TOKEN);
 }
 
-static void free_token_linked_list (_pSLang_Token_Type *tok)
-{
-   while (tok != NULL)
-     {
-	_pSLang_Token_Type *next = tok->next;
-	free_token (tok);
-	if (tok->num_refs != 0)
-	  {
-	     SLang_verror (SL_INTERNAL_ERROR, "Cannot free token in linked list");
-	  }
-	else
-	  SLfree ((char *) tok);
-
-	tok = next;
-     }
-}
-
-/*
- * This function parses a structure definition block.  It returns the names
- * of the structure fields in the form of a linked list of tokens.
- *
- * If the structure contains assignments, the function parses the expressions and returns the
- * number of such assignments.  So:
- *
- *  foo = foo_expr, bar = bar_expr, ...
- *
- * generates:  foo_expr_tokens "foo" bar_expr_tokens "bar" ...
- */
-static _pSLang_Token_Type *
-  handle_struct_assign_list (_pSLang_Token_Type *ctok, int assign_ok, unsigned int *nassignp)
-{
-   _pSLang_Token_Type *name_list_root = NULL;
-   _pSLang_Token_Type *name_list_tail = NULL;
-   unsigned int n, m;
-
-   n = m = 0;
-   while ((_pSLang_Error == 0) 
-	  && (IDENT_TOKEN == ctok->type))
-     {
-	_pSLang_Token_Type *new_tok = allocate_token ();
-	if (new_tok == NULL)
-	  break;
-
-	*new_tok = *ctok;
-	new_tok->type = STRING_TOKEN;
-
-	init_token (ctok);
-
-	if (name_list_root == NULL)
-	  name_list_tail = name_list_root = new_tok;
-	else
-	  name_list_tail->next = new_tok;
-	name_list_tail = new_tok;
-
-	n++;
-
-	if (COMMA_TOKEN == get_token (ctok))
-	  {
-	     get_token (ctok);
-	     continue;
-	  }
-
-	if (assign_ok && (ASSIGN_TOKEN == ctok->type))
-	  {
-	     /* name = ... */
-#if SLANG_HAS_BOSEOS
-	     int eos = compile_bos (ctok, 1);
-#endif
-	     get_token (ctok);
-	     simple_expression (ctok);
-#if SLANG_HAS_BOSEOS
-	     if (eos) compile_eos ();
-#endif
-	     if (-1 == append_copy_of_string_token (new_tok))
-	       break;
-	     
-	     m++;
-
-	     if (ctok->type == COMMA_TOKEN)
-	       {
-		  get_token (ctok);
-		  continue;
-	       }
-	  }	
-	break;
-     }
-
-   if (_pSLang_Error)
-     {
-	free_token_linked_list (name_list_root);
-	return NULL;
-     }
-   
-   if (n == 0)
-     {
-	_pSLparse_error (SL_SYNTAX_ERROR, "Expecting an identifier", ctok, 0);
-	return NULL;
-     }
-
-   *nassignp = m;
-   return name_list_root;
-}
-
-static int handle_struct_fields (_pSLang_Token_Type *ctok, int assign_ok)
-{
-   _pSLang_Token_Type *name_list, *next;
-   unsigned int n, m;
-
-   if (NULL == (name_list = handle_struct_assign_list (ctok, assign_ok, &m)))
-     return -1;
-
-   n = 0;
-   next = name_list;
-   while (next != NULL)
-     {
-	if (-1 == append_token (next))
-	  break;
-	next = next->next;
-	n++;
-     }
-   free_token_linked_list (name_list);
-
-   if (_pSLang_Error)
-     return -1;
-
-   append_int_token (n);
-   if (m == 0)
-     append_token_of_type (STRUCT_TOKEN);
-   else
-     {
-	append_int_token (m);
-	append_token_of_type (STRUCT_WITH_ASSIGN_TOKEN);
-     }
-   
-   if (_pSLang_Error)
-     return -1;
-
-   return 0;
-}
-
 /* struct-declaration:
  * 	struct { struct-field-list };
  *
  * struct-field-list:
- * 	struct-field-name [= simple_expr], struct-field-list
- * 	struct-field-name [= simple_expr]
+ * 	struct-field-name , struct-field-list
+ * 	struct-field-name
  *
- * Generates code: 
- *    "field-name-1" ... "field-name-N" N STRUCT_TOKEN
- * - OR -
- *    expr-k1 "field-name-k1" ... expr-kM "field-name-kM"  
- *        "name-1" ... "field-name-N" N M STRUCT_DEF_ASSIGN_TOKEN
+ * Generates code: "field-name-1" ... "field-name-N" N STRUCT_TOKEN
  */
-static void struct_declaration (_pSLang_Token_Type *ctok, int assign_ok)
+static void struct_declaration (_pSLang_Token_Type *ctok)
 {
+   int n;
+   _pSLang_Token_Type num_tok;
+
    if (ctok->type != OBRACE_TOKEN)
      {
 	_pSLparse_error (SL_SYNTAX_ERROR, "Expecting {", ctok, 0);
 	return;
      }
-   get_token (ctok);
 
-   if (-1 == handle_struct_fields (ctok, assign_ok))
-     return;
-   
+   n = 0;
+   while (IDENT_TOKEN == get_token (ctok))
+     {
+	n++;
+	ctok->type = STRING_TOKEN;
+	append_token (ctok);
+	if (COMMA_TOKEN != get_token (ctok))
+	  break;
+     }
+
    if (ctok->type != CBRACE_TOKEN)
      {
 	_pSLparse_error (SL_SYNTAX_ERROR, "Expecting }", ctok, 0);
 	return;
      }
+   if (n == 0)
+     {
+	_pSLparse_error (SL_SYNTAX_ERROR, "struct requires at least 1 field", ctok, 0);
+	return;
+     }
+
+   init_token (&num_tok);
+   num_tok.type = INT_TOKEN;
+   num_tok.v.long_val = n;
+   append_token (&num_tok);
+   append_token_of_type (STRUCT_TOKEN);
+
    get_token (ctok);
 }
 
@@ -1746,7 +1511,7 @@ static void typedef_definition (_pSLang_Token_Type *t)
      }
    get_token (t);
 
-   struct_declaration (t, 0);
+   struct_declaration (t);
    if (t->type != IDENT_TOKEN)
      {
 	_pSLparse_error (SL_SYNTAX_ERROR, "Expecting identifier", t, 0);
@@ -1945,9 +1710,6 @@ static void expression (_pSLang_Token_Type *ctok)
 /* priority levels of binary operations */
 static unsigned char Binop_Level[] =
 {
-/* SC_AND_TOKEN */	10,
-/* SC_OR_TOKEN */	12,
-/* POW_TOKEN */		0,  /* This case is handled in unary expression so that -x^2 == -(x^2) */
 /* ADD_TOKEN */		2,
 /* SUB_TOKEN */		2,
 /* MUL_TOKEN */		1,
@@ -1956,10 +1718,10 @@ static unsigned char Binop_Level[] =
 /* LE_TOKEN */		4,
 /* GT_TOKEN */		4,
 /* GE_TOKEN */		4,
-/* EQ_TOKEN */		4,
-/* NE_TOKEN */		4,
+/* EQ_TOKEN */		5,
+/* NE_TOKEN */		5,
 /* AND_TOKEN */		9,
-/* OR_TOKEN */		11,
+/* OR_TOKEN */		10,
 /* MOD_TOKEN */		1,
 /* BAND_TOKEN */	6,
 /* SHL_TOKEN */		3,
@@ -1969,112 +1731,6 @@ static unsigned char Binop_Level[] =
 /* POUND_TOKEN */	1  /* Matrix Multiplication */
 };
 
-static void handle_binary_sequence (_pSLang_Token_Type *, unsigned char);
-static void handle_sc_sequence (_pSLang_Token_Type *ctok, unsigned char level)
-{
-   unsigned char type = ctok->type;
-   
-   while ((ctok->type == type) && (_pSLang_Error == 0))
-     {
-	append_token_of_type (OBRACE_TOKEN);
-	get_token (ctok);
-	unary_expression (ctok);
-	handle_binary_sequence (ctok, level);
-	append_token_of_type (CBRACE_TOKEN);
-     }
-   append_token_of_type (type);
-}
-
-/* unary0 OP1 unary1 OP2 unary2 ... OPN unaryN
- * ==> unary0 unary1 ... unaryN {OP1 OP2 ... OPN} COMPARE_BLOCK
- */
-static void handle_compare_sequence (_pSLang_Token_Type *ctok, unsigned char level)
-{
-   unsigned char op_stack [64];
-   unsigned int op_num = 0;
-   unsigned int i;
-
-   do
-     {
-	if (op_num >= sizeof(op_stack))
-	  {
-	     _pSLparse_error (SL_BUILTIN_LIMIT_EXCEEDED, "Too many comparison operators", ctok, 0);
-	     return;
-	  }
-	op_stack[op_num++] = ctok->type;
-	get_token (ctok);
-	unary_expression (ctok);
-	handle_binary_sequence (ctok, level);
-     }
-   while (IS_COMPARE_OP(ctok->type) && (_pSLang_Error == 0));
-
-   if (op_num == 1)
-     {
-	append_token_of_type (op_stack[0]);
-	return;
-     }
-
-   append_token_of_type (OBRACE_TOKEN);
-   for (i = 0; i < op_num; i++)
-     append_token_of_type (op_stack[i]);
-   append_token_of_type (CBRACE_TOKEN);
-   append_token_of_type (_COMPARE_TOKEN);
-}
-
-
-static void handle_binary_sequence (_pSLang_Token_Type *ctok, unsigned char max_level)
-{
-   unsigned char op_stack [64];
-   unsigned char level_stack [64];
-   unsigned char level;
-   unsigned int op_num;
-   unsigned char type;
-
-   op_num = 0;
-   type = ctok->type;
-
-   while ((_pSLang_Error == 0)
-	  && (IS_BINARY_OP(type)))
-     {
-	level = Binop_Level[type - FIRST_BINARY_OP];
-	if (level >= max_level)
-	  break;
-
-	while ((op_num > 0) && (level_stack [op_num - 1] <= level))
-	  append_token_of_type (op_stack [--op_num]);
-
-	if ((type == SC_AND_TOKEN) || (type == SC_OR_TOKEN))
-	  {
-	     handle_sc_sequence (ctok, level);
-	     type = ctok->type;
-	     continue;
-	  }
-
-	if (IS_COMPARE_OP(type))
-	  {
-	     handle_compare_sequence (ctok, level);
-	     type = ctok->type;
-	     continue;
-	  }
-	if (op_num >= sizeof (op_stack) - 1)
-	  {
-	     _pSLparse_error (SL_BUILTIN_LIMIT_EXCEEDED, "Binary op stack overflow", ctok, 0);
-	     return;
-	  }
-
-	op_stack [op_num] = type;
-	level_stack [op_num] = level;
-	op_num++;
-
-	get_token (ctok);
-	unary_expression (ctok);
-	type = ctok->type;
-     }
-
-   while (op_num > 0)
-     append_token_of_type(op_stack[--op_num]);
-}
-
 /* % Note: simple-expression groups operators OP1 at same level.  The
  * % actual implementation will not do this.
  * simple-expression:
@@ -2093,6 +1749,10 @@ static void handle_binary_sequence (_pSLang_Token_Type *ctok, unsigned char max_
 static void simple_expression (_pSLang_Token_Type *ctok)
 {
    unsigned char type;
+   unsigned char op_stack [64];
+   unsigned char level_stack [64];
+   unsigned char level;
+   unsigned int op_num;
 
    switch (ctok->type)
      {
@@ -2135,7 +1795,33 @@ static void simple_expression (_pSLang_Token_Type *ctok)
    if (SEMICOLON_TOKEN == (type = ctok->type))
      return;
 
-   handle_binary_sequence (ctok, 0xFF);
+   op_num = 0;
+
+   while ((_pSLang_Error == 0)
+	  && (IS_BINARY_OP(type)))
+     {
+	level = Binop_Level[type - FIRST_BINARY_OP];
+
+	while ((op_num > 0) && (level_stack [op_num - 1] <= level))
+	  append_token_of_type (op_stack [--op_num]);
+
+	if (op_num >= sizeof (op_stack) - 1)
+	  {
+	     _pSLparse_error (SL_BUILTIN_LIMIT_EXCEEDED, "Binary op stack overflow", ctok, 0);
+	     return;
+	  }
+
+	op_stack [op_num] = type;
+	level_stack [op_num] = level;
+	op_num++;
+
+	get_token (ctok);
+	unary_expression (ctok);
+	type = ctok->type;
+     }
+
+   while (op_num > 0)
+     append_token_of_type(op_stack[--op_num]);
 }
 
 /* unary-expression:
@@ -2175,7 +1861,6 @@ static void unary_expression (_pSLang_Token_Type *ctok)
 
 	   case SUB_TOKEN:
 	     (void) get_token (ctok);
-#if 0
 	     if (IS_INTEGER_TOKEN (ctok->type))
 	       {
 		  ctok->v.long_val = -ctok->v.long_val;
@@ -2188,7 +1873,7 @@ static void unary_expression (_pSLang_Token_Type *ctok)
 		  break;
 	       }
 #endif
-#endif
+
 	     if (num_unary_ops == 16)
 	       goto stack_overflow_error;
 	     save_unary_ops [num_unary_ops++] = CHS_TOKEN;
@@ -2221,6 +1906,28 @@ static void unary_expression (_pSLang_Token_Type *ctok)
      }
 
    out_of_switch:
+   if (num_unary_ops == 0)
+     return;
+#if 0
+   if ((DEREF_TOKEN == save_unary_ops[num_unary_ops - 1])
+       && (NULL != (last_token = get_last_token ()))
+       && (IS_ASSIGN_TOKEN(last_token->type)))
+     {
+	/* FIXME: Priority=medium
+	 * This needs generalized so that things like @a.y = 1 will work properly.
+	 */
+	if ((num_unary_ops != 1)
+	    || (last_token->type != _SCALAR_ASSIGN_TOKEN))
+	  {
+	     SLang_verror (SL_NOT_IMPLEMENTED, 
+			   "Only derefence assignments to simple variables are possible");
+	     return;
+	  }
+
+	last_token->type += (_DEREF_ASSIGN_TOKEN - _SCALAR_ASSIGN_TOKEN);
+	return;
+     }
+#endif
    while (num_unary_ops)
      {
 	num_unary_ops--;
@@ -2357,7 +2064,6 @@ static void postfix_expression (_pSLang_Token_Type *ctok)
 {
    unsigned int start_pos, end_pos;
    unsigned char type;
-   _pSLang_Token_Type *last_token;
 
    if (Token_List == NULL)
      return;
@@ -2396,7 +2102,7 @@ static void postfix_expression (_pSLang_Token_Type *ctok)
       case COMPLEX_TOKEN:
 #endif
       case STRING_DOLLAR_TOKEN:
-#ifdef HAVE_LONG_LONG
+#if HAVE_LONG_LONG
       case LLONG_TOKEN:
       case ULLONG_TOKEN:
 #endif
@@ -2415,7 +2121,6 @@ static void postfix_expression (_pSLang_Token_Type *ctok)
 	break;
 
       case BAND_TOKEN:
-#if 0
 	if (IDENT_TOKEN != get_identifier_expr_token (ctok))
 	  break;
 
@@ -2424,25 +2129,6 @@ static void postfix_expression (_pSLang_Token_Type *ctok)
 	get_token (ctok);
 	if ((ctok->type == OBRACKET_TOKEN) || (ctok->type == DOT_TOKEN))
 	  _pSLparse_error (SL_NOT_IMPLEMENTED, "& of an array or structure element is not currently supported", ctok, 0);
-#else
-	get_token (ctok);
-	postfix_expression (ctok);
-	last_token = get_last_token ();
-	switch (last_token->type)
-	  {
-	   case IDENT_TOKEN:
-	     last_token->type = _REF_TOKEN;
-	     break;
-	   case ARRAY_TOKEN:
-	     last_token->type = _ARRAY_ELEM_REF_TOKEN;
-	     break;
-	   case DOT_TOKEN:
-	     last_token->type = _STRUCT_FIELD_REF_TOKEN;
-	     break;
-	   default:
-	     _pSLparse_error (SL_NOT_IMPLEMENTED, "& not supported in this context", last_token, 0);
-	  }
-#endif
 	break;
 
       case OBRACKET_TOKEN:
@@ -2468,7 +2154,7 @@ static void postfix_expression (_pSLang_Token_Type *ctok)
 
       case STRUCT_TOKEN:
 	get_token (ctok);
-	struct_declaration (ctok, 1);
+	struct_declaration (ctok);
 	break;
 
       case TMP_TOKEN:
@@ -2527,21 +2213,9 @@ static void postfix_expression (_pSLang_Token_Type *ctok)
 
 	   case OPAREN_TOKEN:
 	     /* f(args) ==> args f */
-	     if ((NULL != (last_token = get_last_token ()))
-		 && (last_token->type == DEREF_TOKEN))
-	       {
-		  /* Expressions such as (@A[i])(args...) */
-		  last_token->type = _DEREF_FUNCALL_TOKEN;
-		  append_token_of_type (ARG_TOKEN);
-		  (void) get_token (ctok);
-		  function_args_expression (ctok, 0, 1);
-		  token_list_element_exchange (start_pos, end_pos);
-		  break;
-	       } 
-	     
 	     if (CPAREN_TOKEN != get_token (ctok))
 	       {
-		  function_args_expression (ctok, 1, 1);
+		  function_args_expression (ctok, 1);
 		  token_list_element_exchange (start_pos, end_pos);
 	       }
 	     else get_token (ctok);
@@ -2562,13 +2236,15 @@ static void postfix_expression (_pSLang_Token_Type *ctok)
 #ifdef DOT_METHOD_CALL_TOKEN
 	     if (ctok->type == OPAREN_TOKEN)
 	       {
+		  _pSLang_Token_Type *last_token;
+
 		  if (NULL == (last_token = get_last_token ()))
 		    return;
 		  last_token->type = DOT_METHOD_CALL_TOKEN;
 		  end_pos = Token_List->len;
 		  append_token_of_type (ARG_TOKEN);
 		  get_token (ctok);
-		  function_args_expression (ctok, 0, 1);
+		  function_args_expression (ctok, 0);
 		  token_list_element_exchange (start_pos, end_pos);
 	       }
 #endif
@@ -2593,20 +2269,19 @@ static void postfix_expression (_pSLang_Token_Type *ctok)
 	     token_list_element_exchange (start_pos, end_pos);
 	     break;
 #endif
-#if 1
 	   case POW_TOKEN:
 	     get_token (ctok);
 	     unary_expression (ctok);
 	     append_token_of_type (POW_TOKEN);
 	     break;
-#endif
+
 	   default:
 	     return;
 	  }
      }
 }
 
-static void function_args_expression (_pSLang_Token_Type *ctok, int handle_num_args, int handle_qualifiers)
+static void function_args_expression (_pSLang_Token_Type *ctok, int handle_num_args)
 {
    unsigned char last_type, this_type;
    
@@ -2633,37 +2308,10 @@ static void function_args_expression (_pSLang_Token_Type *ctok, int handle_num_a
 	     get_token (ctok);
 	     return;
 
-	   case SEMICOLON_TOKEN:
-	     if (handle_qualifiers)
-	       {
-		  if (last_type == COMMA_TOKEN)
-		    append_token_of_type (_NULL_TOKEN);
-
-		  if (SEMICOLON_TOKEN == get_token (ctok))
-		    {
-		       /* foo (args... ;; q) form */
-		       if (CPAREN_TOKEN == get_token (ctok))
-			 break;  /* foo (args ;;) */
-		       simple_expression (ctok);
-		    }
-		  else if (ctok->type == CPAREN_TOKEN)
-		    break;	       /* foo (args;) */
-		  else if (-1 == handle_struct_fields (ctok, 1))
-		    return;
-
-		  append_token_of_type (QUALIFIER_TOKEN);
-		  if (ctok->type != CPAREN_TOKEN)
-		    _pSLparse_error (SL_SYNTAX_ERROR, "Expecting ')'", ctok, 0);
-		  break;
-	       }
-	     /* drop */
-		  
 	   default:
 	     simple_expression (ctok);
 	     if ((ctok->type != COMMA_TOKEN)
-		 && (ctok->type != CPAREN_TOKEN)
-		 && ((handle_qualifiers == 0) 
-		     || (ctok->type != SEMICOLON_TOKEN)))
+		 && (ctok->type != CPAREN_TOKEN))
 	       {
 		  _pSLparse_error (SL_SYNTAX_ERROR, "Expecting ')'", ctok, 0);
 		  break;
@@ -2679,26 +2327,22 @@ static int check_for_lvalue (unsigned char eqs_type, _pSLang_Token_Type *ctok)
 
    if ((ctok == NULL)
        && (NULL == (ctok = get_last_token ())))
-     type = ILLEGAL_TOKEN;
-   else 
-     type = ctok->type;
+     return -1;
+
+   type = ctok->type;
 
    eqs_type -= ASSIGN_TOKEN;
-   switch (type)
+
+   if (type == IDENT_TOKEN)
+     eqs_type += _SCALAR_ASSIGN_TOKEN;
+   else if (type == ARRAY_TOKEN)
+     eqs_type += _ARRAY_ASSIGN_TOKEN;
+   else if (type == DOT_TOKEN)
+     eqs_type += _STRUCT_ASSIGN_TOKEN;
+   else if (type == DEREF_TOKEN)
+     eqs_type += _DEREF_ASSIGN_TOKEN;
+   else
      {
-      case IDENT_TOKEN:
-	eqs_type += _SCALAR_ASSIGN_TOKEN;
-	break;
-      case ARRAY_TOKEN:
-	eqs_type += _ARRAY_ASSIGN_TOKEN;
-	break;
-      case DOT_TOKEN:
-	eqs_type += _STRUCT_ASSIGN_TOKEN;
-	break;
-      case DEREF_TOKEN:
-	eqs_type += _DEREF_ASSIGN_TOKEN;
-	break;
-      default:
 	_pSLparse_error (SL_SYNTAX_ERROR, "Expecting LVALUE", ctok, 0);
 	return -1;
      }
@@ -2724,8 +2368,6 @@ static void inline_list_expression (_pSLang_Token_Type *ctok)
 	     if (ctok->type != COMMA_TOKEN)
 	       break;
 	     get_token (ctok);
-	     if (ctok->type == CBRACE_TOKEN)   /* trailing comma: {a,b,} */
-	       break;
 	  }
 	if (ctok->type != CBRACE_TOKEN)
 	  {
@@ -2769,8 +2411,6 @@ static void array_index_expression (_pSLang_Token_Type *ctok)
 	  return;
 	num_commas++;
 	get_token (ctok);
-	if (ctok->type == CBRACKET_TOKEN)   /* allow trailing comma */
-	  return;
      }
 }
 
@@ -2778,12 +2418,10 @@ static void array_index_expression (_pSLang_Token_Type *ctok)
  *    array_index_expression
  *    simple_expression : simple_expression
  *    simple_expression : simple_expression : simple_expression
- *    simple_expression : simple_expression : # simple_expression
  */
 static void inline_array_expression (_pSLang_Token_Type *ctok)
 {
    int num_colons = 0;
-   int has_pound = 0;
 
    append_token_of_type (ARG_TOKEN);
 
@@ -2804,11 +2442,7 @@ static void inline_array_expression (_pSLang_Token_Type *ctok)
 	if (ctok->type == COLON_TOKEN)
 	  {
 	     num_colons++;
-	     if (POUND_TOKEN == get_token (ctok))
-	       {
-		  has_pound = 1;
-		  get_token(ctok);
-	       }
+	     get_token (ctok);
 	     simple_expression (ctok);
 	  }
      }
@@ -2821,12 +2455,7 @@ static void inline_array_expression (_pSLang_Token_Type *ctok)
 
    /* append_token_of_type (EARG_TOKEN); */
    if (num_colons)
-     {
-	if (has_pound)
-	  append_token_of_type (_INLINE_IMPLICIT_ARRAYN_TOKEN);
-	else
-	  append_token_of_type (_INLINE_IMPLICIT_ARRAY_TOKEN);
-     }
+     append_token_of_type (_INLINE_IMPLICIT_ARRAY_TOKEN);
    else
      append_token_of_type (_INLINE_ARRAY_TOKEN);
    get_token (ctok);
