@@ -1,10 +1,24 @@
 /* file intrinsics for S-Lang */
-/* Copyright (c) 1992, 1999, 2001 John E. Davis
- * This file is part of the S-Lang library.
- *
- * You may distribute under the terms of either the GNU General Public
- * License or the Perl Artistic License.
- */
+/*
+Copyright (C) 2004, 2005 John E. Davis
+
+This file is part of the S-Lang Library.
+
+The S-Lang Library is free software; you can redistribute it and/or
+modify it under the terms of the GNU General Public License as
+published by the Free Software Foundation; either version 2 of the
+License, or (at your option) any later version.
+
+The S-Lang Library is distributed in the hope that it will be useful,
+but WITHOUT ANY WARRANTY; without even the implied warranty of
+MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+General Public License for more details.
+
+You should have received a copy of the GNU General Public License
+along with this library; if not, write to the Free Software
+Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
+USA.  
+*/
 
 #include "slinclud.h"
 
@@ -61,6 +75,7 @@
 
 #if defined(__WIN32__)
 # include <windows.h>
+# include <direct.h>
 #else
 # if defined(__OS2__) && defined(__IBMC__)
 #  define INCL_DOS
@@ -98,35 +113,37 @@
 #include "slang.h"
 #include "_slang.h"
 
+typedef struct
+{
+   struct stat st;
+   int st_opt_attrs;
+}
+Stat_Type;
+
+static SLang_CStruct_Field_Type Stat_Struct [] =
+{
+   MAKE_CSTRUCT_INT_FIELD(Stat_Type, st.st_dev, "st_dev", 0),
+   MAKE_CSTRUCT_INT_FIELD(Stat_Type, st.st_ino, "st_ino", 0),
+   MAKE_CSTRUCT_INT_FIELD(Stat_Type, st.st_mode, "st_mode", 0),
+   MAKE_CSTRUCT_INT_FIELD(Stat_Type, st.st_nlink, "st_nlink", 0),
+   MAKE_CSTRUCT_UINT_FIELD(Stat_Type, st.st_uid, "st_uid", 0),
+   MAKE_CSTRUCT_UINT_FIELD(Stat_Type, st.st_gid, "st_gid", 0),
+   MAKE_CSTRUCT_INT_FIELD(Stat_Type, st.st_rdev, "st_rdev", 0),
+   MAKE_CSTRUCT_UINT_FIELD(Stat_Type, st.st_size, "st_size", 0),
+   MAKE_CSTRUCT_UINT_FIELD(Stat_Type, st.st_atime, "st_atime", 0),
+   MAKE_CSTRUCT_UINT_FIELD(Stat_Type, st.st_mtime, "st_mtime", 0),
+   MAKE_CSTRUCT_UINT_FIELD(Stat_Type, st.st_ctime, "st_ctime", 0),
+   MAKE_CSTRUCT_INT_FIELD(Stat_Type, st_opt_attrs, "st_opt_attrs", 0),
+   SLANG_END_CSTRUCT_TABLE
+};
+
 static int push_stat_struct (struct stat *st, int opt_attrs)
 {
-   char *field_names [12];
-   unsigned char field_types[12];
-   VOID_STAR field_values [12];
-   int int_values [12];
-   unsigned int i;
-
-   field_names [0] = "st_dev"; int_values [0] = (int) st->st_dev;
-   field_names [1] = "st_ino"; int_values [1] = (int) st->st_ino;
-   field_names [2] = "st_mode"; int_values [2] = (int) st->st_mode;
-   field_names [3] = "st_nlink"; int_values [3] = (int) st->st_nlink;
-   field_names [4] = "st_uid"; int_values [4] = (int) st->st_uid;
-   field_names [5] = "st_gid"; int_values [5] = (int) st->st_gid;
-   field_names [6] = "st_rdev"; int_values [6] = (int) st->st_rdev;
-   field_names [7] = "st_size"; int_values [7] = (int) st->st_size;
-   field_names [8] = "st_atime"; int_values [8] = (int) st->st_atime;
-   field_names [9] = "st_mtime"; int_values [9] = (int) st->st_mtime;
-   field_names [10] = "st_ctime"; int_values [10] = (int) st->st_ctime;
-
-   field_names [11] = "st_opt_attrs"; int_values[11] = opt_attrs;
-
-   for (i = 0; i < 12; i++)
-     {
-	field_types [i] = SLANG_INT_TYPE;
-	field_values [i] = (VOID_STAR) (int_values + i);
-     }
-
-   return SLstruct_create_struct (12, field_names, field_types, field_values);
+   Stat_Type s;
+   
+   s.st = *st;
+   s.st_opt_attrs = opt_attrs;
+   return SLang_push_cstruct ((VOID_STAR) &s, Stat_Struct);
 }
 
 static void stat_cmd (char *file)
@@ -154,7 +171,7 @@ static void stat_cmd (char *file)
 #endif
    if (status == -1)
      {
-	_SLerrno_errno = errno;
+	_pSLerrno_errno = errno;
 	SLang_push_null ();
 	return;
      }
@@ -176,7 +193,7 @@ static void lstat_cmd (char *file)
 
    if (-1 == lstat (file, &st))
      {
-	_SLerrno_errno = errno;
+	_pSLerrno_errno = errno;
 	SLang_push_null ();
 	return;
      }
@@ -290,7 +307,7 @@ static void readlink_cmd (char *s)
    n = readlink (s, buf, sizeof (buf)-1);
    if (n == -1)
      {
-	_SLerrno_errno = errno;
+	_pSLerrno_errno = errno;
 	s = NULL;
      }
    else
@@ -303,11 +320,35 @@ static void readlink_cmd (char *s)
 }
 #endif
 
+#ifdef HAVE_SYMLINK
+static int symlink_cmd (char *oldpath, char *newpath)
+{
+   int s;
+   
+   if (-1 == (s = symlink (oldpath, newpath)))
+     _pSLerrno_errno = errno;
+   
+   return s;
+}
+#endif
+
+#ifdef HAVE_LINK
+static int hardlink_cmd (char *oldpath, char *newpath)
+{
+   int s;
+   
+   if (-1 == (s = link (oldpath, newpath)))
+     _pSLerrno_errno = errno;
+   
+   return s;
+}
+#endif
+
 static int chmod_cmd (char *file, int *mode)
 {
    if (-1 == chmod(file, (mode_t) *mode))
      {
-	_SLerrno_errno = errno;
+	_pSLerrno_errno = errno;
 	return -1;
      }
    return 0;
@@ -319,7 +360,7 @@ static int chown_cmd (char *file, int *owner, int *group)
    int ret;
 
    if (-1 == (ret = chown(file, (uid_t) *owner, (gid_t) *group)))
-     _SLerrno_errno = errno;
+     _pSLerrno_errno = errno;
    return ret;
 }
 #endif
@@ -361,7 +402,7 @@ static void slget_cwd (void)
 
    if (p == NULL)
      {
-	_SLerrno_errno = errno;
+	_pSLerrno_errno = errno;
 	SLang_push_null ();
 	return;
      }
@@ -394,7 +435,7 @@ static int chdir_cmd (char *s)
 	if (errno == EINTR)
 	  continue;
 #endif
-	_SLerrno_errno = errno;
+	_pSLerrno_errno = errno;
 	break;
      }
    return ret;
@@ -502,7 +543,7 @@ static int rmdir_cmd (char *s)
 	if (errno == EINTR)
 	  continue;
 #endif
-	_SLerrno_errno = errno;
+	_pSLerrno_errno = errno;
 	break;
      }
    return ret;
@@ -528,7 +569,7 @@ static int remove_cmd (char *s)
 	if (errno == EINTR)
 	  continue;
 #endif
-	_SLerrno_errno = errno;
+	_pSLerrno_errno = errno;
 	break;
      }
    return ret;
@@ -543,17 +584,26 @@ static int rename_cmd (char *oldpath, char *newpath)
 	if (errno == EINTR)
 	  continue;
 #endif
-	_SLerrno_errno = errno;
+	_pSLerrno_errno = errno;
 	break;
      }
    return ret;
 }
 
-static int mkdir_cmd (char *s, int *mode_ptr)
+static int mkdir_cmd (void)
 {
    int ret;
+   char *dir;
+   int mode = 0777;
+   
+   if (SLang_Num_Function_Args != 1)
+     {
+	if (-1 == SLang_pop_integer (&mode))
+	  return -1;
+     }
+   if (-1 == SLang_pop_slstring (&dir))
+     return -1;
 
-   (void) mode_ptr;
    errno = 0;
 
 #if defined (__MSDOS__) && !defined(__GO32__)
@@ -562,7 +612,7 @@ static int mkdir_cmd (char *s, int *mode_ptr)
 # if defined (__os2__) && !defined (__EMX__)
 #  define MKDIR(x,y) mkdir(x)
 # else
-#  if defined (__WIN32__) && !defined (__CYGWIN32__)
+#  if (defined (__WIN32__) || defined(__WINDOWS__)) && !defined (__CYGWIN32__)
 #   define MKDIR(x,y) mkdir(x)
 #  else
 #   define MKDIR mkdir
@@ -570,15 +620,17 @@ static int mkdir_cmd (char *s, int *mode_ptr)
 # endif
 #endif
 
-   while (-1 == (ret = MKDIR(s, *mode_ptr)))
+   while (-1 == (ret = MKDIR(dir, mode)))
      {
 #ifdef EINTR
 	if (errno == EINTR)
 	  continue;
 #endif
-	_SLerrno_errno = errno;
+	_pSLerrno_errno = errno;
 	break;
      }
+   
+   SLang_free_slstring (dir);
    return ret;
 }
 
@@ -587,7 +639,7 @@ static int mkfifo_cmd (char *path, int *mode)
 {
    if (-1 == mkfifo (path, *mode))
      {
-	_SLerrno_errno = errno;
+	_pSLerrno_errno = errno;
 	return -1;
      }
    return 0;
@@ -644,19 +696,19 @@ static int build_dirlist (char *file, char *opt, char ***listp, unsigned int *nu
 # ifdef __WIN32__
    if (status == (DWORD)-1)
      {
-	_SLerrno_errno = ENOENT;
+	_pSLerrno_errno = ENOENT;
 	return -1;
      }
    if (0 == (status & FILE_ATTRIBUTE_DIRECTORY))
      {
-	_SLerrno_errno = ENOTDIR;
+	_pSLerrno_errno = ENOTDIR;
 	return -1;
      }
 # else
    if ((rc != 0) || (status.attrFile & FILE_DIRECTORY) == 0)
      {
 	/* ENOTDIR isn't defined in VA3. */
-	_SLerrno_errno = ENOENT;
+	_pSLerrno_errno = ENOENT;
 	return -1;
      }
 # endif
@@ -766,7 +818,7 @@ static int build_dirlist (char *file, char *opt, char ***listp, unsigned int *nu
 		  break;
 	       }
 
-	     _SLerrno_errno = errno;
+	     _pSLerrno_errno = errno;
 	     FindClose (h);
 	     goto return_error;
 	  }
@@ -781,7 +833,7 @@ static int build_dirlist (char *file, char *opt, char ***listp, unsigned int *nu
 		  break;
 	       }
 
-	     _SLerrno_errno = errno;
+	     _pSLerrno_errno = errno;
 	     DosFindClose (h);
 	     goto return_error;
 	  }
@@ -814,7 +866,7 @@ static int build_dirlist (char *dir, char *opt, char ***listp, unsigned int *num
 
    if (NULL == (dp = opendir (dir)))
      {
-	_SLerrno_errno = errno;
+	_pSLerrno_errno = errno;
 	return -1;
      }
 
@@ -873,7 +925,7 @@ static void listdir_cmd (char *dir, char *opt)
    SLang_Array_Type *at;
    unsigned int num_files;
    unsigned int max_num_files;
-   int inum_files;
+   SLindex_Type inum_files;
    char **list;
 
    if (-1 == build_dirlist (dir, opt, &list, &num_files, &max_num_files))
@@ -896,7 +948,7 @@ static void listdir_cmd (char *dir, char *opt)
 	list = new_list;
      }
 
-   inum_files = (int) num_files;
+   inum_files = (SLindex_Type) num_files;
    if (NULL == (at = SLang_create_array (SLANG_STRING_TYPE, 0, (VOID_STAR) list, &inum_files, 1)))
      {
 	free_dir_list (list, num_files);
@@ -949,6 +1001,12 @@ static SLang_Intrin_Fun_Type PosixDir_Name_Table [] =
 {
 #ifdef HAVE_READLINK
    MAKE_INTRINSIC_S("readlink", readlink_cmd, SLANG_VOID_TYPE),
+#endif
+#ifdef HAVE_SYMLINK
+   MAKE_INTRINSIC_SS("symlink", symlink_cmd, SLANG_INT_TYPE),
+#endif
+#ifdef HAVE_LINK
+   MAKE_INTRINSIC_SS("hardlink", hardlink_cmd, SLANG_INT_TYPE),
 #endif
    MAKE_INTRINSIC_S("lstat_file", lstat_cmd, SLANG_VOID_TYPE),
    MAKE_INTRINSIC_S("stat_file", stat_cmd, SLANG_VOID_TYPE),
@@ -964,7 +1022,7 @@ static SLang_Intrin_Fun_Type PosixDir_Name_Table [] =
    MAKE_INTRINSIC_I("umask", umask_cmd, SLANG_INT_TYPE),
 #endif
    MAKE_INTRINSIC_0("getcwd", slget_cwd, SLANG_VOID_TYPE),
-   MAKE_INTRINSIC_SI("mkdir", mkdir_cmd, SLANG_INT_TYPE),
+   MAKE_INTRINSIC_0("mkdir", mkdir_cmd, SLANG_INT_TYPE),
    MAKE_INTRINSIC_S("chdir", chdir_cmd, SLANG_INT_TYPE),
    MAKE_INTRINSIC_S("rmdir", rmdir_cmd, SLANG_INT_TYPE),
    MAKE_INTRINSIC_S("remove", remove_cmd, SLANG_INT_TYPE),
@@ -1047,7 +1105,7 @@ int SLang_init_posix_dir (void)
 
    if ((-1 == SLadd_intrin_fun_table(PosixDir_Name_Table, "__POSIX_DIR__"))
        || (-1 == SLadd_iconstant_table (PosixDir_Consts, NULL))
-       || (-1 == _SLerrno_init ()))
+       || (-1 == _pSLerrno_init ()))
      return -1;
 
    Initialized = 1;
