@@ -1,15 +1,53 @@
 /* ed style regular expressions */
-/* Copyright (c) 1992, 1999 John E. Davis
- *
- * You may distribute under the terms of either the GNU General Public
- * License or the Perl Artistic License.
- */
+/*
+Copyright (C) 2004, 2005 John E. Davis
+
+This file is part of the S-Lang Library.
+
+The S-Lang Library is free software; you can redistribute it and/or
+modify it under the terms of the GNU General Public License as
+published by the Free Software Foundation; either version 2 of the
+License, or (at your option) any later version.
+
+The S-Lang Library is distributed in the hope that it will be useful,
+but WITHOUT ANY WARRANTY; without even the implied warranty of
+MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+General Public License for more details.
+
+You should have received a copy of the GNU General Public License
+along with this library; if not, write to the Free Software
+Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
+USA.  
+*/
 
 #include "slinclud.h"
 
 #include "slang.h"
 #include "_slang.h"
 
+struct _pSLRegexp_Type
+{
+   /* These must be set by calling routine. */
+   unsigned char *pat;		       /* regular expression pattern */
+   unsigned char *buf;		       /* buffer for compiled regexp */
+   unsigned int buf_len;	       /* length of buffer */
+   int case_sensitive;		       /* 1 if match is case sensitive  */
+
+   /* The rest are set by SLang_regexp_compile */
+
+   int must_match;		       /* 1 if line must contain substring */
+   int must_match_bol;		       /* true if it must match beginning of line */
+   unsigned char must_match_str[16];   /* 15 char null term substring */
+   int osearch;			       /* 1 if ordinary search suffices */
+   unsigned int min_length;	       /* minimum length the match must be */
+   int beg_matches[10];		       /* offset of start of \( */
+   unsigned int end_matches[10];       /* length of nth submatch
+					* Note that the entire match corresponds
+					* to \0
+					*/
+   int offset;			       /* offset to be added to beg_matches */
+};
+
 #define SET_BIT(b, n) b[(unsigned int) (n) >> 3] |= 1 << ((unsigned int) (n) % 8)
 #define TEST_BIT(b, n) (b[(unsigned int)(n) >> 3] & (1 << ((unsigned int) (n) % 8)))
 #define LITERAL 1
@@ -40,21 +78,33 @@
 static unsigned char Word_Chars[256];
 #define IS_WORD_CHAR(x) Word_Chars[(unsigned int) (x)]
 
-static int Open_Paren_Number;
+#if 0
+static int ctx->open_paren_number;
 static char Closed_Paren_Matches[10];
 
 static SLRegexp_Type *This_Reg;
 static unsigned char *This_Str;
+#endif
 
-static unsigned char *do_nth_match(int n, unsigned char *str, unsigned char *estr)
+typedef struct
+{
+   SLRegexp_Type *reg;
+   unsigned char *str;
+   unsigned int len;
+   char closed_paren_matches[10];
+   int open_paren_number;
+}
+Re_Context_Type;
+
+static unsigned char *do_nth_match (Re_Context_Type *ctx, int n, unsigned char *str, unsigned char *estr)
 {
    unsigned char *bpos;
 
-   if (Closed_Paren_Matches[n] == 0)
+   if (ctx->closed_paren_matches[n] == 0)
      return NULL;
 
-   bpos = This_Reg->beg_matches[n] + This_Str;
-   n = This_Reg->end_matches[n];
+   bpos = ctx->reg->beg_matches[n] + ctx->str;
+   n = ctx->reg->end_matches[n];
    if (n == 0) return(str);
    if (n > (int) (estr - str)) return (NULL);
 
@@ -65,7 +115,10 @@ static unsigned char *do_nth_match(int n, unsigned char *str, unsigned char *est
 }
 
 /* returns pointer to the end of regexp or NULL */
-static unsigned char *regexp_looking_at(register unsigned char *str, unsigned char *estr, unsigned char *buf, register int cs)
+static unsigned char *regexp_looking_at (Re_Context_Type *ctx, 
+					 unsigned char *str, unsigned char *estr, 
+					 unsigned char *regexp, 
+					 int cs)
 {
    register unsigned char p, p1;
    unsigned char *save_str, *tmpstr;
@@ -73,17 +126,17 @@ static unsigned char *regexp_looking_at(register unsigned char *str, unsigned ch
    int save_num_open;
    char save_closed_matches[10];
 
-   p = *buf++;
+   p = *regexp++;
 
    while (p != 0)
      {
-	/* p1 = UPPERCASE(*buf); */
+	/* p1 = UPPERCASE(*regexp); */
 	/* if (str < estr) c = UPPERCASE(*str); */
 
 	switch((unsigned char) p)
 	  {
 	   case BOW:
-	     if ((str != This_Str)
+	     if ((str != ctx->str)
 		 && ((str >= estr)
 		     || IS_WORD_CHAR(*(str - 1))
 		     || (0 == IS_WORD_CHAR(*str)))) return NULL;
@@ -98,55 +151,55 @@ static unsigned char *regexp_looking_at(register unsigned char *str, unsigned ch
 	   case NO_CASE: cs = 0; break;
 
 	   case OPAREN:
-	     Open_Paren_Number++;
-	     This_Reg->beg_matches[Open_Paren_Number] = (int) (str - This_Str);
+	     ctx->open_paren_number++;
+	     ctx->reg->beg_matches[ctx->open_paren_number] = (int) (str - ctx->str);
 	     break;
 	   case CPAREN:
-	     n = Open_Paren_Number;
+	     n = ctx->open_paren_number;
 	     while (n > 0)
 	       {
-		  if (Closed_Paren_Matches[n] != 0)
+		  if (ctx->closed_paren_matches[n] != 0)
 		    {
 		       n--;
 		       continue;
 		    }
-		  Closed_Paren_Matches[n] = 1;
-		  This_Reg->end_matches[n] = (unsigned int) (str - (This_Str + This_Reg->beg_matches[n]));
+		  ctx->closed_paren_matches[n] = 1;
+		  ctx->reg->end_matches[n] = (unsigned int) (str - (ctx->str + ctx->reg->beg_matches[n]));
 		  break;
 	       }
 	     break;
 #ifdef NOT_LITERAL
 	   case NOT_LITERAL:
-	     if ((str >= estr) || (*buf == UPPERCASE(*str))) return (NULL);
-	     str++; buf++;
+	     if ((str >= estr) || (*regexp == UPPERCASE(*str))) return (NULL);
+	     str++; regexp++;
 	     break;
 
 	   case MAYBE_ONCE | NOT_LITERAL:
 	     save_str = str;
-	     if ((str < estr) && (*buf != UPPERCASE(*str))) str++;
-	     buf++;
+	     if ((str < estr) && (*regexp != UPPERCASE(*str))) str++;
+	     regexp++;
 	     goto match_rest;
 
 	   case NOT_LITERAL | LEAST_ONCE:   /* match at least once */
-	     if ((str >= estr) || (UPPERCASE(*str) == UPPERCASE(*buf))) return (NULL);
+	     if ((str >= estr) || (UPPERCASE(*str) == UPPERCASE(*regexp))) return (NULL);
 	     str++;
 	     /* drop */
 	   case STAR | NOT_LITERAL:
-	     save_str = str;  p1 = *buf;
+	     save_str = str;  p1 = *regexp;
 	     while ((str < estr) && (UPPERCASE(*str) != p1)) str++;
-	     buf++;
+	     regexp++;
 	     goto match_rest;
 
 	     /* this type consists of the expression + two bytes that
 	        determine number of matches to perform */
 	   case MANY | NOT_LITERAL:
-	     p1 = *buf; buf++;
-	     n = n0 = (int) (unsigned char) *buf++;
+	     p1 = *regexp; regexp++;
+	     n = n0 = (int) (unsigned char) *regexp++;
 	     /* minimum number to match--- could be 0 */
-	     n1 = (int) (unsigned char) *buf++;
+	     n1 = (int) (unsigned char) *regexp++;
 	     /* maximum number to match */
 
-	     while (n && (str < estr) && (p1 != *str))
+	     while (n && (str < estr) && (p1 != UPPERCASE(*str)))
 	       {
 		  n--;
 		  str++;
@@ -155,7 +208,7 @@ static unsigned char *regexp_looking_at(register unsigned char *str, unsigned ch
 
 	     save_str = str;
 	     n = n1 - n0;
-	     while (n && (str < estr) && (p1 != *str))
+	     while (n && (str < estr) && (p1 != UPPERCASE(*str)))
 	       {
 		  n--;
 		  str++;
@@ -163,36 +216,36 @@ static unsigned char *regexp_looking_at(register unsigned char *str, unsigned ch
 	     goto match_rest;
 #endif				       /* NOT_LITERAL */
 	   case LITERAL:
-	     if ((str >= estr) || (*buf != UPPERCASE(*str))) return (NULL);
-	     str++; buf++;
+	     if ((str >= estr) || (*regexp != UPPERCASE(*str))) return (NULL);
+	     str++; regexp++;
 	     break;
 
 	   case MAYBE_ONCE | LITERAL:
 	     save_str = str;
-	     if ((str < estr) && (*buf == UPPERCASE(*str))) str++;
-	     buf++;
+	     if ((str < estr) && (*regexp == UPPERCASE(*str))) str++;
+	     regexp++;
 	     goto match_rest;
 
 	   case LITERAL | LEAST_ONCE:   /* match at least once */
-	     if ((str >= estr) || (UPPERCASE(*str) != UPPERCASE(*buf))) return (NULL);
+	     if ((str >= estr) || (UPPERCASE(*str) != UPPERCASE(*regexp))) return (NULL);
 	     str++;
 	     /* drop */
 	   case STAR | LITERAL:
-	     save_str = str;  p1 = *buf;
+	     save_str = str;  p1 = *regexp;
 	     while ((str < estr) && (UPPERCASE(*str) == p1)) str++;
-	     buf++;
+	     regexp++;
 	     goto match_rest;
 
 	     /* this type consists of the expression + two bytes that
 	        determine number of matches to perform */
 	   case MANY | LITERAL:
-	     p1 = *buf; buf++;
-	     n = n0 = (int) (unsigned char) *buf++;
+	     p1 = *regexp; regexp++;
+	     n = n0 = (int) (unsigned char) *regexp++;
 	     /* minimum number to match--- could be 0 */
-	     n1 = (int) (unsigned char) *buf++;
+	     n1 = (int) (unsigned char) *regexp++;
 	     /* maximum number to match */
 
-	     while (n && (str < estr) && (p1 == *str))
+	     while (n && (str < estr) && (p1 == UPPERCASE(*str)))
 	       {
 		  n--;
 		  str++;
@@ -201,7 +254,7 @@ static unsigned char *regexp_looking_at(register unsigned char *str, unsigned ch
 
 	     save_str = str;
 	     n = n1 - n0;
-	     while (n && (str < estr) && (p1 == *str))
+	     while (n && (str < estr) && (p1 == UPPERCASE(*str)))
 	       {
 		  n--;
 		  str++;
@@ -209,14 +262,14 @@ static unsigned char *regexp_looking_at(register unsigned char *str, unsigned ch
 	     goto match_rest;
 
 	   case NTH_MATCH:
-	     if ((str = do_nth_match((int) (unsigned char) *buf, str, estr)) == NULL) return(NULL);
-	     buf++;
+	     if ((str = do_nth_match(ctx, (int) (unsigned char) *regexp, str, estr)) == NULL) return(NULL);
+	     regexp++;
 	     break;
 
 	   case MAYBE_ONCE | NTH_MATCH:
 	     save_str = str;
-	     tmpstr = do_nth_match((int) (unsigned char) *buf, str, estr);
-	     buf++;
+	     tmpstr = do_nth_match (ctx, (int) (unsigned char) *regexp, str, estr);
+	     regexp++;
 	     if (tmpstr != NULL)
 	       {
 		  str = tmpstr;
@@ -225,15 +278,15 @@ static unsigned char *regexp_looking_at(register unsigned char *str, unsigned ch
 	     continue;
 
 	   case LEAST_ONCE | NTH_MATCH:
-	     if ((str = do_nth_match((int) (unsigned char) *buf, str, estr)) == NULL) return(NULL);
+	     if ((str = do_nth_match(ctx, (int) (unsigned char) *regexp, str, estr)) == NULL) return(NULL);
 	     /* drop */
 	   case STAR | NTH_MATCH:
 	     save_str = str;
-	     while (NULL != (tmpstr = do_nth_match((int) (unsigned char) *buf, str, estr)))
+	     while (NULL != (tmpstr = do_nth_match(ctx, (int) (unsigned char) *regexp, str, estr)))
 	       {
 		  str = tmpstr;
 	       }
-	     buf++;
+	     regexp++;
 	     goto match_rest;
 
 	   case MANY | NTH_MATCH: return(NULL);
@@ -241,24 +294,24 @@ static unsigned char *regexp_looking_at(register unsigned char *str, unsigned ch
 
 	   case RANGE:
 	     if (str >= estr) return (NULL);
-	     if (TEST_BIT(buf, UPPERCASE(*str)) == 0) return (NULL);
-	     buf += 32; str++;
+	     if (TEST_BIT(regexp, UPPERCASE(*str)) == 0) return (NULL);
+	     regexp += 32; str++;
 	     break;
 
 	   case MAYBE_ONCE | RANGE:
 	     save_str = str;
-	     if ((str < estr) && TEST_BIT(buf, UPPERCASE(*str))) str++;
-	     buf += 32;
+	     if ((str < estr) && TEST_BIT(regexp, UPPERCASE(*str))) str++;
+	     regexp += 32;
 	     goto match_rest;
 
 	   case LEAST_ONCE | RANGE:
-	     if ((str >= estr) || (0 == TEST_BIT(buf, UPPERCASE(*str)))) return NULL;
+	     if ((str >= estr) || (0 == TEST_BIT(regexp, UPPERCASE(*str)))) return NULL;
 	     str++;
 	     /* drop */
 	   case STAR | RANGE:
 	     save_str = str;
-	     while ((str < estr) && TEST_BIT(buf, UPPERCASE(*str))) str++;
-	     buf += 32;
+	     while ((str < estr) && TEST_BIT(regexp, UPPERCASE(*str))) str++;
+	     regexp += 32;
 	     goto match_rest;
 
 	     /* The first 32 bytes correspond to the range and the two
@@ -266,11 +319,11 @@ static unsigned char *regexp_looking_at(register unsigned char *str, unsigned ch
 	      */
 	   case MANY | RANGE:
 	     /* minimum number to match--- could be 0 */
-	     n = n0 = (int) (unsigned char) *(buf + 32);
+	     n = n0 = (int) (unsigned char) *(regexp + 32);
 	     /* maximum number to match */
-	     n1 = (int) (unsigned char) *(buf + 33);
+	     n1 = (int) (unsigned char) *(regexp + 33);
 
-	     while (n && (str < estr) && (TEST_BIT(buf, UPPERCASE(*str))))
+	     while (n && (str < estr) && (TEST_BIT(regexp, UPPERCASE(*str))))
 	       {
 		  n--;
 		  str++;
@@ -278,12 +331,12 @@ static unsigned char *regexp_looking_at(register unsigned char *str, unsigned ch
 	     if (n) return (NULL);
 	     save_str = str;
 	     n = n1 - n0;
-	     while (n && (str < estr) && (TEST_BIT(buf, UPPERCASE(*str))))
+	     while (n && (str < estr) && (TEST_BIT(regexp, UPPERCASE(*str))))
 	       {
 		  n--;
 		  str++;
 	       }
-	     buf += 34;		       /* 32 + 2 */
+	     regexp += 34;		       /* 32 + 2 */
 	     goto match_rest;
 
 	   case ANY_DIGIT:
@@ -297,7 +350,7 @@ static unsigned char *regexp_looking_at(register unsigned char *str, unsigned ch
 	     goto match_rest;
 
 	   case LEAST_ONCE | ANY_DIGIT:
-	     if ((str < estr) && ((*str > '9') || (*str < '0'))) return NULL;
+	     if ((str >= estr) || ((*str > '9') || (*str < '0'))) return NULL;
 	     str++;
 	     /* drop */
 	   case STAR | ANY_DIGIT:
@@ -338,13 +391,13 @@ static unsigned char *regexp_looking_at(register unsigned char *str, unsigned ch
 
 	   default: return (NULL);
 	  }
-	p = *buf++;
+	p = *regexp++;
 	continue;
 
 	match_rest:
 	if (save_str == str)
 	  {
-	     p = *buf++;
+	     p = *regexp++;
 	     continue;
 	  }
 
@@ -354,14 +407,14 @@ static unsigned char *regexp_looking_at(register unsigned char *str, unsigned ch
 	 * }
 	 */
 
-	SLMEMCPY(save_closed_matches, Closed_Paren_Matches, sizeof(save_closed_matches));
-	save_num_open = Open_Paren_Number;
+	SLMEMCPY(save_closed_matches, ctx->closed_paren_matches, sizeof(save_closed_matches));
+	save_num_open = ctx->open_paren_number;
 	while (str >= save_str)
 	  {
-	     tmpstr = regexp_looking_at(str, estr, buf, cs);
+	     tmpstr = regexp_looking_at (ctx, str, estr, regexp, cs);
 	     if (tmpstr != NULL) return(tmpstr);
-	     SLMEMCPY(Closed_Paren_Matches, save_closed_matches, sizeof(Closed_Paren_Matches));
-	     Open_Paren_Number = save_num_open;
+	     SLMEMCPY(ctx->closed_paren_matches, save_closed_matches, sizeof(ctx->closed_paren_matches));
+	     ctx->open_paren_number = save_num_open;
 	     str--;
 	  }
 	return NULL;
@@ -370,7 +423,7 @@ static unsigned char *regexp_looking_at(register unsigned char *str, unsigned ch
 }
 
 static void
-fixup_beg_end_matches (SLRegexp_Type *r, unsigned char *str, unsigned char *epos)
+fixup_beg_end_matches (Re_Context_Type *ctx, SLRegexp_Type *r, unsigned char *str, unsigned char *epos)
 {
    int i;
 
@@ -378,17 +431,17 @@ fixup_beg_end_matches (SLRegexp_Type *r, unsigned char *str, unsigned char *epos
      {
 	r->beg_matches[0] = -1;
 	r->end_matches[0] = 0;
-	SLMEMSET(Closed_Paren_Matches, 0, sizeof(Closed_Paren_Matches));
+	SLMEMSET(ctx->closed_paren_matches, 0, sizeof(ctx->closed_paren_matches));
      }
    else
      {
-	r->beg_matches[0] = (int) (str - This_Str);
+	r->beg_matches[0] = (int) (str - ctx->str);
 	r->end_matches[0] = (unsigned int) (epos - str);
      }
 
    for (i = 1; i < 10; i++)
      {
-	if (Closed_Paren_Matches [i] == 0)
+	if (ctx->closed_paren_matches [i] == 0)
 	  {
 	     r->beg_matches[i] = -1;
 	     r->end_matches[i] = 0;
@@ -396,26 +449,33 @@ fixup_beg_end_matches (SLRegexp_Type *r, unsigned char *str, unsigned char *epos
      }
 }
 
-unsigned char *SLang_regexp_match(unsigned char *str,
+static void init_re_context (Re_Context_Type *ctx, SLRegexp_Type *reg, 
+			     unsigned char *str, unsigned int len)
+{
+   memset ((char *) ctx, 0, sizeof (Re_Context_Type));
+   ctx->reg = reg;
+   ctx->str = str;
+   ctx->len = len;
+}
+
+static unsigned char *regexp_match(unsigned char *str,
 				  unsigned int len, SLRegexp_Type *reg)
 {
    register unsigned char c = 0, *estr = str + len;
    int cs = reg->case_sensitive, lit = 0;
    unsigned char *buf = reg->buf, *epos = NULL;
+   Re_Context_Type ctx_buf;
 
    if (reg->min_length > len) return NULL;
 
-   Open_Paren_Number = 0;
-   SLMEMSET(Closed_Paren_Matches, 0, sizeof(Closed_Paren_Matches));
-   This_Reg = reg;
-   This_Str = str;
+   init_re_context (&ctx_buf, reg, str, len);
 
    if (*buf == BOL)
      {
-	if (NULL == (epos = regexp_looking_at(str, estr, buf + 1, cs)))
+	if (NULL == (epos = regexp_looking_at (&ctx_buf, str, estr, buf + 1, cs)))
 	  str = NULL;
 
-	fixup_beg_end_matches (reg, str, epos);
+	fixup_beg_end_matches (&ctx_buf, reg, str, epos);
 	return str;
      }
 
@@ -440,10 +500,10 @@ unsigned char *SLang_regexp_match(unsigned char *str,
 	c = *(buf + 2);
      }
 
-   while (str < estr)
+   while (1)
      {
-	Open_Paren_Number = 0;
-	SLMEMSET(Closed_Paren_Matches, 0, sizeof(Closed_Paren_Matches));
+	ctx_buf.open_paren_number = 0;
+	memset (ctx_buf.closed_paren_matches, 0, sizeof(ctx_buf.closed_paren_matches));
 	/* take care of leading chars */
 	if (lit)
 	  {
@@ -452,17 +512,24 @@ unsigned char *SLang_regexp_match(unsigned char *str,
 	       break;		       /* failed */
 	  }
 
-	if (NULL != (epos = regexp_looking_at(str, estr, buf, cs)))
+	if (NULL != (epos = regexp_looking_at(&ctx_buf, str, estr, buf, cs)))
 	  {
-	     fixup_beg_end_matches (reg, str, epos);
+	     fixup_beg_end_matches (&ctx_buf, reg, str, epos);
 	     return str;
 	  }
+	if (str >= estr)
+	  break;
 	str++;
      }
-   fixup_beg_end_matches (reg, NULL, epos);
+   fixup_beg_end_matches (&ctx_buf, reg, NULL, epos);
    return NULL;
 }
 
+char *SLregexp_match (SLRegexp_Type *reg, char *str, unsigned int len)
+{
+   return (char *) regexp_match ((unsigned char *)str, len, reg);
+}
+
 static unsigned char *convert_digit(unsigned char *pat, int *nn)
 {
    int n = 0, m = 0;
@@ -484,7 +551,7 @@ static unsigned char *convert_digit(unsigned char *pat, int *nn)
 #define ERROR  return (int) (pat - reg->pat)
 
 /* Returns 0 if successful or offset in pattern of error */
-int SLang_regexp_compile (SLRegexp_Type *reg)
+static int regexp_compile (SLRegexp_Type *reg)
 {
    register unsigned char *buf, *ebuf, *pat;
    unsigned char *last = NULL, *tmppat;
@@ -521,6 +588,22 @@ int SLang_regexp_compile (SLRegexp_Type *reg)
 	i++;
      }
 
+   if (*pat == '\\')
+     {
+	if (pat[1] == 'c')
+	  {
+	     cs = 1;
+	     pat += 2;
+	     no_osearch = 1;
+	  }
+	else if (pat[1] == 'C')
+	  {
+	     cs = 0;
+	     pat += 2;
+	     no_osearch = 1;
+	  }
+     }
+
    if (*pat == '^')
      {
 	pat++;
@@ -529,6 +612,12 @@ int SLang_regexp_compile (SLRegexp_Type *reg)
      }
    else reg->must_match_bol = 0;
 
+   if (cs != reg->case_sensitive)
+     {
+	if (cs) *buf++ = YES_CASE; 
+	else *buf++ = NO_CASE; 
+     }
+
    *buf = 0;
 
    last_count = count = 0;
@@ -536,7 +625,7 @@ int SLang_regexp_compile (SLRegexp_Type *reg)
      {
 	if (buf >= ebuf - 3)
 	  {
-	     SLang_doerror ("Pattern too large to be compiled.");
+	     SLang_verror (SL_BUILTIN_LIMIT_EXCEEDED, "Pattern too large to be compiled.");
 	     ERROR;
 	  }
 
@@ -814,6 +903,41 @@ int SLang_regexp_compile (SLRegexp_Type *reg)
    return(0);
 }
 
+void SLregexp_free (SLRegexp_Type *reg)
+{
+   if (reg == NULL)
+     return;
+   if (reg->buf != NULL)
+     SLfree ((char *) reg->buf);
+   SLfree ((char *) reg);
+}
+
+SLRegexp_Type *SLregexp_compile (char *pattern, unsigned int flags)
+{
+   SLRegexp_Type *reg;
+
+   reg = (SLRegexp_Type *)SLcalloc (1, sizeof (SLRegexp_Type));
+   if (reg == NULL)
+     return NULL;
+   
+   if (NULL == (reg->buf = (unsigned char *)SLmalloc (1024)))
+     {
+	SLfree ((char *) reg);
+	return NULL;
+     }
+   reg->buf_len = 1024;
+   reg->case_sensitive = (0 == (flags & SLREGEXP_CASELESS));
+   reg->pat = (unsigned char *)pattern;
+
+   if (-1 == regexp_compile (reg))
+     {
+	SLregexp_free (reg);
+	return NULL;
+     }
+   
+   return reg;
+}
+
 char *SLregexp_quote_string (char *re, char *buf, unsigned int buflen)
 {
    char ch;
@@ -852,6 +976,39 @@ char *SLregexp_quote_string (char *re, char *buf, unsigned int buflen)
    return NULL;
 }
 
+int SLregexp_nth_match (SLRegexp_Type *reg, unsigned int nth,
+			unsigned int *ofsp, unsigned int *lenp)
+{
+   if (nth >= 10)
+     {
+	SLang_set_error (SL_InvalidParm_Error);
+	return -1;
+     }
+   if (reg->beg_matches[nth] < 0)
+     return -1;
+   
+   if (ofsp != NULL)
+     *ofsp = (unsigned int) reg->beg_matches[nth];
+   if (lenp != NULL)
+     *lenp = (unsigned int) reg->end_matches[nth];
+
+   return 0;
+}
+
+int SLregexp_get_hints (SLRegexp_Type *reg, unsigned int *hintsp)
+{
+   unsigned int hints = 0;
+   
+   if (reg == NULL)
+     return -1;
+   
+   if (reg->osearch) hints |= SLREGEXP_HINT_OSEARCH;
+   if (reg->must_match_bol) hints |= SLREGEXP_HINT_BOL;
+   
+   *hintsp = hints;
+   return 0;
+}
+
 #if 0
 #define MAX_EXP 4096
 int main(int argc, char **argv)
