@@ -1,24 +1,4 @@
 /* SLang Screen management routines */
-/*
-Copyright (C) 2004, 2005, 2006, 2007 John E. Davis
-
-This file is part of the S-Lang Library.
-
-The S-Lang Library is free software; you can redistribute it and/or
-modify it under the terms of the GNU General Public License as
-published by the Free Software Foundation; either version 2 of the
-License, or (at your option) any later version.
-
-The S-Lang Library is distributed in the hope that it will be useful,
-but WITHOUT ANY WARRANTY; without even the implied warranty of
-MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-General Public License for more details.
-
-You should have received a copy of the GNU General Public License
-along with this library; if not, write to the Free Software
-Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
-USA.  
-*/
 #include "slinclud.h"
 
 #include <stdio.h>
@@ -41,7 +21,7 @@ Screen_Type;
 #define TRASHED 0x2
 static int Screen_Trashed;
 
-static Screen_Type SL_Screen[SLTT_MAX_SCREEN_ROWS];
+Screen_Type SL_Screen[SLTT_MAX_SCREEN_ROWS];
 static int Start_Col, Start_Row;
 static unsigned int Screen_Cols, Screen_Rows;
 static int This_Row, This_Col;
@@ -457,13 +437,10 @@ static int point_visible (int col_too)
 
 #define NEXT_CHAR_CELL \
    { \
-      if (p < pmax) \
-	{ \
-	   if ((p->nchars != i) || (p->color != color)) flags |= TOUCHED; \
-	   p->nchars = i; \
-	   p->color = color; \
-	   p++; \
-	} \
+      if ((p->nchars != i) || (p->color != color)) flags |= TOUCHED; \
+      p->nchars = i; \
+      p->color = color; \
+      p++; \
       i = 0; \
       col++; \
    } (void) 0
@@ -471,7 +448,7 @@ static int point_visible (int col_too)
 
 #define ADD_TO_CHAR_CELL(wc) \
    { \
-      if ((p < pmax) && (p->wchars[i] != wc)) \
+      if (p->wchars[i] != wc) \
 	{ \
 	   p->wchars[i] = wc; \
 	   flags |= TOUCHED; \
@@ -496,7 +473,7 @@ static int point_visible (int col_too)
 
 void SLsmg_write_chars (unsigned char *u, unsigned char *umax)
 {
-   SLsmg_Char_Type *p, *pmax;
+   SLsmg_Char_Type *p;
    SLsmg_Color_Type color;
    int flags;
    int col, start_col, max_col;
@@ -510,16 +487,13 @@ void SLsmg_write_chars (unsigned char *u, unsigned char *umax)
    SLsmg_Color_Type default_color;
 #endif
    if (Smg_Inited == 0) return;
-
-   if (u >= umax)
-     return;
-
+   
    display_8bit = (unsigned char) SLsmg_Display_Eight_Bit;
    if (utf8_mode)
      display_8bit = 0xA0;
 
    color = This_Color;
-   /* If we are using unicode characters for the line drawing characters, then
+   /* If we are unicode characters for the line drawing characters, then
     * do not attempt to use the terminals alternate character set
     */
    alt_char_set_flag = (color & SLSMG_ACS_MASK);
@@ -541,24 +515,19 @@ void SLsmg_write_chars (unsigned char *u, unsigned char *umax)
    max_col = start_col + Screen_Cols;
 
    p = SL_Screen[This_Row - Start_Row].neew;
-   pmax = p + Screen_Cols;
-
    if (col >= start_col)
      {
 	p += (col - start_col);
-	if ((p < pmax) && (p->nchars == 0))
+	if (p->nchars == 0)
 	  {
 	     /* It looks like we are about to overwrite the right side of a 
-	      * double width character.  Let's see...
+	      * double width character.
 	      */
 	     if (col > start_col)
 	       {
 		  p--;
-		  if (p->nchars != 0)
-		    {
-		       p->nchars = 1;
-		       p->wchars[0] = ' ';
-		    }
+		  p->nchars = 1;
+		  p->wchars[0] = ' ';
 		  p++;
 	       }
 	  }
@@ -808,6 +777,34 @@ void SLsmg_write_string (char *str)
 		      (unsigned char *)str + strlen (str));
 }
 
+void SLsmg_write_nstring (char *str, unsigned int n)
+{
+   unsigned int width;
+   unsigned char *blank = (unsigned char *)" ";
+   unsigned char *u = (unsigned char *)str;
+
+   /* Avoid a problem if a user accidently passes a negative value */
+   if ((int) n < 0)
+     return;
+
+   if (u == NULL) width = 0;
+   else
+     {
+	unsigned char *umax;
+	
+	width = strlen ((char *)u);
+	if (UTF8_Mode)
+	  umax = SLutf8_skip_chars (u, u+width, n, &width, 0);
+	else 
+	  {
+	     if (width > n) width = n;
+	     umax = u + width;
+	  }
+	SLsmg_write_chars (u, umax);
+     }
+
+   while (width++ < n) SLsmg_write_chars (blank, blank+1);
+}
 
 void SLsmg_write_wrapped_string (SLuchar_Type *u, int r, int c,
 				 unsigned int dr, unsigned int dc,
@@ -816,26 +813,15 @@ void SLsmg_write_wrapped_string (SLuchar_Type *u, int r, int c,
    int maxc = (int) dc;
    unsigned char *p, *pmax;
    int utf8_mode = UTF8_Mode;
-   unsigned char display_8bit = (unsigned char) SLsmg_Display_Eight_Bit;
-
-   if (utf8_mode)
-     display_8bit = 0xA0;
 
    if ((dr == 0) || (dc == 0)) return;
-   if (u == NULL)
-     u = (unsigned char *)"";
-
    p = u;
    pmax = u + strlen ((char *)u);
    
    dc = 0;
    while (1)
      {
-	SLwchar_Type wc;
-	unsigned int nconsumed;
 	unsigned char ch = *p;
-	unsigned int ddc;
-
 	if ((ch == 0) || (ch == '\n'))
 	  {
 	     int diff;
@@ -859,82 +845,27 @@ void SLsmg_write_wrapped_string (SLuchar_Type *u, int r, int c,
 	     continue;
 	  }
 
-	/* If the width of the characters buffered so far extend to or beyond
-	 * the width of the box, then write them out and goto the
-	 * next line. Note that dc > maxc if the displayable width of
-	 * the last character to be written is greater than the width
-	 * of the box.  This will be the case if (maxc<ddc) -- see below
-	 */
-	if ((int) dc >= maxc)
-	  goto write_chars_and_reset;
-
-	nconsumed = 1;
-	if (ch < 0x80)
+	if ((int) dc == maxc)
 	  {
-	     p++;
-	     if ((ch >= 0x20) && (ch != 0x7F))
-	       {
-		  dc++;
-		  continue;
-	       }
-	     /* Otherwise display as ^X */
-	     dc += 2;
+	     SLsmg_gotorc (r, c);
+	     SLsmg_write_chars (u, p);
+	     if (dr == 1) break;
+
+	     r++;
+	     dc = 0;
+	     dr--;
+	     u = p;
 	     continue;
 	  }
 
-	nconsumed = 1;
-	if ((utf8_mode == 0)
-	    || (NULL == SLutf8_decode (p, pmax, &wc, &nconsumed)))
-	  {	     
-	     if ((utf8_mode == 0)
-		 && (display_8bit && (*p >= display_8bit)))
-	       {
-		  dc++;
-		  p += nconsumed;
-		  continue;
-	       }
-
-	     ddc = 4*nconsumed;      /* <XX> */
-	  }
-	else if (wc < (SLwchar_Type)display_8bit)
-	  ddc = 4;		       /* displays as <XX> */
+	dc++;
+	if (utf8_mode)
+	  p = SLutf8_skip_chars (p, pmax, 1, NULL, 0);
 	else
-	  ddc = SLwchar_wcwidth (wc);
-
-	dc += ddc;
-	if (((int)dc > maxc) && (maxc > (int)ddc))
-	  {
-	     dc -= ddc;
-	     goto write_chars_and_reset;
-	  }
-	p += nconsumed;
-	continue;
-
-	write_chars_and_reset:	
-	SLsmg_gotorc (r, c);
-	SLsmg_write_chars (u, p);
-	while ((int)dc < maxc)
-	  {
-	     SLsmg_write_char (' ');
-	     dc++;
-	  }
-	if (dr == 1) break;
-	r++;
-	dc = 0;
-	dr--;
-	u = p;
+	  p++;
      }
 }
 
-void SLsmg_write_nstring (char *str, unsigned int n)
-{
-   /* Avoid a problem if a user accidently passes a negative value */
-   if ((int) n < 0)
-     return;
-   
-   SLsmg_write_wrapped_string ((unsigned char *)str, This_Row, This_Col, 1, n, 1);
-}
-
 int SLsmg_Tab_Width = 8;
 
 /* Minimum value for which eight bit char is displayed as is. */
@@ -1678,6 +1609,7 @@ void SLsmg_draw_object (int r, int c, SLwchar_Type object)
 void SLsmg_draw_hline (unsigned int n)
 {
    static unsigned char hbuf[16];
+   int count;
    int cmin, cmax;
    int final_col = This_Col + (int) n;
    int save_color;
@@ -1693,6 +1625,7 @@ void SLsmg_draw_hline (unsigned int n)
      }
 
    n = (unsigned int)(cmax - cmin);
+   count = n / 16;
 
    save_color = This_Color;
    This_Color |= SLSMG_ACS_MASK;
