@@ -1,12 +1,26 @@
 /* -*- mode: C; mode: fold; -*- */
 /* Standard intrinsic functions for S-Lang.  Included here are string
    and array operations */
-/* Copyright (c) 1992, 1999 John E. Davis
- * This file is part of the S-Lang library.
- *
- * You may distribute under the terms of either the GNU General Public
- * License or the Perl Artistic License.
- */
+/*
+Copyright (C) 2004, 2005 John E. Davis
+
+This file is part of the S-Lang Library.
+
+The S-Lang Library is free software; you can redistribute it and/or
+modify it under the terms of the GNU General Public License as
+published by the Free Software Foundation; either version 2 of the
+License, or (at your option) any later version.
+
+The S-Lang Library is distributed in the hope that it will be useful,
+but WITHOUT ANY WARRANTY; without even the implied warranty of
+MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+General Public License for more details.
+
+You should have received a copy of the GNU General Public License
+along with this library; if not, write to the Free Software
+Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
+USA.  
+*/
 
 #include "slinclud.h"
 /*{{{ Include Files */
@@ -68,8 +82,8 @@ static int length_cmd (void)
    if (-1 == SLang_pop (&obj))
      return -1;
 
-   cl = _SLclass_get_class (obj.data_type);
-   p = _SLclass_get_ptr_to_value (cl, &obj);
+   cl = _pSLclass_get_class (obj.data_type);
+   p = _pSLclass_get_ptr_to_value (cl, &obj);
 
    len = 1;
    if (cl->cl_length != NULL)
@@ -85,30 +99,42 @@ static int length_cmd (void)
 }
 
 /* convert integer to a string of length 1 */
-static void char_cmd (int *x) /*{{{*/
+static void char_cmd (SLwchar_Type *x) /*{{{*/
 {
-   char ch, buf[2];
+   SLuchar_Type buf[SLUTF8_MAX_MBLEN + 1];
 
-   ch = (char) *x;
-   buf[0] = ch;
-   buf[1] = 0;
-   SLang_push_string (buf);
+   if ((_pSLinterp_UTF8_Mode == 0) || (*x < 0x80))
+     {
+        buf[0] = (SLuchar_Type) *x;
+        buf[1] = 0;
+     }
+   else
+     {
+        SLuchar_Type *p;
+        
+        p = SLutf8_encode (*x, buf, SLUTF8_MAX_MBLEN);
+        if (p == NULL) p = buf;
+        
+        *p = 0;
+     }
+   
+   SLang_push_string ((char *)buf);
 }
 
 /*}}}*/
 
 /* format object into a string and returns slstring */
-char *_SLstringize_object (SLang_Object_Type *obj) /*{{{*/
+char *_pSLstringize_object (SLang_Object_Type *obj) /*{{{*/
 {
    SLang_Class_Type *cl;
-   unsigned char stype;
+   SLtype stype;
    VOID_STAR p;
    char *s, *s1;
 
    stype = obj->data_type;
    p = (VOID_STAR) &obj->v.ptr_val;
 
-   cl = _SLclass_get_class (stype);
+   cl = _pSLclass_get_class (stype);
 
    s = (*cl->cl_string) (stype, p);
    if (s != NULL)
@@ -126,7 +152,8 @@ int SLang_run_hooks(char *hook, unsigned int num_args, ...)
    unsigned int i;
    va_list ap;
 
-   if (SLang_Error) return -1;
+   if (SLang_get_error ())
+     return -1;
 
    if (0 == SLang_is_defined (hook))
      return 0;
@@ -144,7 +171,7 @@ int SLang_run_hooks(char *hook, unsigned int num_args, ...)
    va_end (ap);
    (void) SLang_end_arg_list ();
 
-   if (SLang_Error) return -1;
+   if (_pSLang_Error) return -1;
    return SLang_execute_function (hook);
 }
 
@@ -163,7 +190,7 @@ static void intrin_putenv (void) /*{{{*/
 
    if (putenv (s))
      {
-	SLang_Error = SL_INTRINSIC_ERROR;
+	SLang_set_error (SL_OS_Error);
 	SLfree (s);
      }
 
@@ -174,22 +201,6 @@ static void intrin_putenv (void) /*{{{*/
 
 #endif
 
-static void lang_print_stack (void) /*{{{*/
-{
-   char buf[32];
-   unsigned int n;
-
-   n = (unsigned int) (_SLStack_Pointer - _SLRun_Stack);
-   while (n)
-     {
-	n--;
-	sprintf (buf, "(%u)", n);
-	_SLdump_objects (buf, _SLRun_Stack + n, 1, 1);
-     }
-}
-
-/*}}}*/
-
 static void byte_compile_file (char *f, int *m)
 {
    SLang_byte_compile_file (f, *m);
@@ -209,7 +220,7 @@ static void intrin_type_info1 (void)
 
    SLang_free_object (&obj);
 
-   _SLang_push_datatype (type);
+   SLang_push_datatype (type);
 }
 
 static void intrin_type_info (void)
@@ -219,18 +230,18 @@ static void intrin_type_info (void)
    if (-1 == SLang_pop (&obj))
      return;
 
-   _SLang_push_datatype (obj.data_type);
+   SLang_push_datatype (obj.data_type);
    SLang_free_object (&obj);
 }
 
-void _SLstring_intrinsic (void) /*{{{*/
+void _pSLstring_intrinsic (void) /*{{{*/
 {
    SLang_Object_Type x;
    char *s;
 
    if (SLang_pop (&x)) return;
-   if (NULL != (s = _SLstringize_object (&x)))
-     _SLang_push_slstring (s);
+   if (NULL != (s = _pSLstringize_object (&x)))
+     _pSLang_push_slstring (s);
 
    SLang_free_object (&x);
 }
@@ -239,8 +250,8 @@ void _SLstring_intrinsic (void) /*{{{*/
 
 static void intrin_typecast (void)
 {
-   unsigned char to_type;
-   if (0 == _SLang_pop_datatype (&to_type))
+   SLtype to_type;
+   if (0 == SLang_pop_datatype (&to_type))
      (void) SLclass_typecast (to_type, 0, 1);
 }
 
@@ -262,9 +273,10 @@ static void intrin_int (void) /*{{{*/
 static char *
 intrin_function_name (void)
 {
-   if (NULL == _SLang_Current_Function_Name)
+   char *name;
+   if (NULL == (name = _pSLang_current_function_name ()))
      return "";
-   return _SLang_Current_Function_Name;
+   return name;
 }
 
 static void intrin_message (char *s)
@@ -274,7 +286,7 @@ static void intrin_message (char *s)
 
 static void intrin_error (char *s)
 {
-   SLang_verror (SL_USER_ERROR, "%s", s);
+   SLang_verror (SL_RunTime_Error, "%s", s);
 }
 
 static void intrin_pop_n (int *n)
@@ -296,7 +308,7 @@ static void usage (void)
 {
    char *msg;
 
-   _SLstrops_do_sprintf_n (SLang_Num_Function_Args - 1);   /* do not include format */
+   _pSLstrops_do_sprintf_n (SLang_Num_Function_Args - 1);   /* do not include format */
 
    if (-1 == SLang_pop_slstring (&msg))
      return;
@@ -312,25 +324,52 @@ static int intrin_integer (char *s)
 
    i = SLatoi ((unsigned char *) s);
 
-   if (SLang_Error)
-     SLang_verror (SL_TYPE_MISMATCH, "Unable to convert string to integer");
    return i;
 }
 /*}}}*/
 
 static void guess_type (char *s)
 {
-   _SLang_push_datatype (SLang_guess_type(s));
+   SLang_push_datatype (SLang_guess_type(s));
 }
 
-static int load_file (char *s)
+
+static int load_string_or_file (int (*f) (char *, char *))
 {
-   if (-1 == SLang_load_file (s))
-     return 0;
-   return 1;
+   char *file;
+   char *ns = NULL;
+   int status;
+
+   if (SLang_Num_Function_Args == 2)
+     {
+	if (-1 == SLang_pop_slstring (&ns))
+	  return -1;
+     }
+
+   if (-1 == SLang_pop_slstring (&file))
+     {
+	SLang_free_slstring (ns);
+	return -1;
+     }
+   
+   status = (*f) (file, ns);
+   SLang_free_slstring (file);
+   SLang_free_slstring (ns);
+   return status;
 }
 
-static void get_doc_string (char *file, char *topic)
+static int load_file (void)
+{
+   return (0 == load_string_or_file (SLns_load_file));
+}
+
+static void load_string (void)
+{
+   /* FIXME: This should use the namespace of the currently executing code */
+   (void) load_string_or_file (SLns_load_string);
+}
+
+static int get_doc_string (char *file, char *topic)
 {
    FILE *fp;
    char line[1024];
@@ -339,10 +378,7 @@ static void get_doc_string (char *file, char *topic)
    char ch;
 
    if (NULL == (fp = fopen (file, "r")))
-     {
-	SLang_push_null ();
-	return;
-     }
+     return -1;
 
    topic_len = strlen (topic);
    ch = *topic;
@@ -352,8 +388,7 @@ static void get_doc_string (char *file, char *topic)
 	if (NULL == fgets (line, sizeof(line), fp))
 	  {
 	     fclose (fp);
-	     (void) SLang_push_null ();
-	     return;
+	     return -1;
 	  }
 
 	if ((ch == *line)
@@ -366,8 +401,7 @@ static void get_doc_string (char *file, char *topic)
    if (NULL == (str = SLmake_string (line)))
      {
 	fclose (fp);
-	(void) SLang_push_null ();
-	return;
+	return -1;
      }
    str_len = strlen (str);
 
@@ -395,6 +429,95 @@ static void get_doc_string (char *file, char *topic)
    fclose (fp);
 
    (void) SLang_push_malloced_string (str);
+   return 0;
+}
+
+static _pSLString_List_Type *Doc_Files;
+
+static int add_doc_file (char *file)
+{
+   if (Doc_Files == NULL)
+     {
+	Doc_Files = _pSLstring_list_new (5, 5);
+	if (Doc_Files == NULL)
+	  return -1;
+     }
+
+   if ((file == NULL) || (*file == 0))
+     return 0;
+
+   return _pSLstring_list_append_copy (Doc_Files, file);
+}
+
+static void add_doc_file_intrin (char *file)
+{
+   (void) add_doc_file (file);
+}
+
+static void get_doc_files_intrin (void)
+{
+   if (Doc_Files == NULL)
+     Doc_Files = _pSLstring_list_new (5, 5);
+   (void) _pSLstring_list_push (Doc_Files, 0);
+}
+
+static void set_doc_files_intrin (void)
+{
+   SLang_Array_Type *at;
+   unsigned int i, num;
+   char **data;
+
+   if (-1 == SLang_pop_array_of_type (&at, SLANG_STRING_TYPE))
+     return;
+   
+   _pSLstring_list_delete (Doc_Files);
+   Doc_Files = NULL;
+
+   num = at->num_elements;
+   data = (char **) at->data;
+   for (i = 0; i < num; i++)
+     {
+	if (-1 == add_doc_file (data[i]))
+	  break;
+     }
+   SLang_free_array (at);
+}
+
+static void get_doc_string_intrin (char *topic)
+{
+   char *file;
+   char **files;
+   unsigned int i, num_files;
+
+   if (SLang_Num_Function_Args == 2)
+     {
+	if (-1 == SLang_pop_slstring (&file))
+	  return;
+	
+	if (-1 == get_doc_string (file, topic))
+	  (void) SLang_push_null ();
+	
+	SLang_free_slstring (file);
+	return;
+     }
+   
+   if ((Doc_Files == NULL)
+       || (NULL == (files = Doc_Files->buf)))
+     {
+	SLang_push_null ();
+	return;
+     }
+   num_files = Doc_Files->num;
+   for (i = 0; i < num_files; i++)
+     {
+	file = files[i];
+	if (file == NULL)
+	  continue;
+	
+	if (0 == get_doc_string (file, topic))
+	  return;
+     }
+   (void) SLang_push_null ();
 }
 
 static int push_string_array_elements (SLang_Array_Type *at)
@@ -444,7 +567,7 @@ static void intrin_apropos (void)
 	  goto free_and_return;
      }
 
-   at = _SLang_apropos (namespace_name, pat, flags);
+   at = _pSLang_apropos (namespace_name, pat, flags);
    if (num_args == 3)
      {
 	(void) SLang_push_array (at, 0);
@@ -467,7 +590,7 @@ static void intrin_apropos (void)
 static int intrin_get_defines (void)
 {
    int n = 0;
-   char **s = _SLdefines;
+   char **s = _pSLdefines;
 
    while (*s != NULL)
      {
@@ -484,7 +607,12 @@ static int intrin_get_defines (void)
 
 static void intrin_get_reference (char *name)
 {
-   _SLang_push_ref (1, (VOID_STAR) _SLlocate_name (name));
+   _pSLang_push_ref (1, (VOID_STAR) _pSLlocate_name (name));
+}
+
+static void intrin_get_namespaces (void)
+{
+   SLang_push_array (_pSLns_list_namespaces (), 1);
 }
 
 #ifdef HAVE_SYS_UTSNAME_H
@@ -496,7 +624,7 @@ static void uname_cmd (void)
 #ifdef HAVE_UNAME
    struct utsname u;
    char *field_names [6];
-   unsigned char field_types[6];
+   SLtype field_types[6];
    VOID_STAR field_values [6];
    char *ptrs[6];
    int i;
@@ -525,38 +653,204 @@ static void uname_cmd (void)
 
 static void uninitialize_ref_intrin (SLang_Ref_Type *ref)
 {
-   (void) _SLang_uninitialize_ref (ref);
+   (void) _pSLang_uninitialize_ref (ref);
 }
 
+static int class_type_intrinsic (void)
+{
+   SLtype type;
+
+   if (-1 == SLang_pop_datatype (&type))
+     return -1;
+   return _pSLclass_get_class (type)->cl_class_type;
+}
+
+static int class_id_intrinsic (void)
+{
+   SLtype type;
+
+   if (-1 == SLang_pop_datatype (&type))
+     return -1;
+   return _pSLclass_get_class (type)->cl_data_type;
+}
+
+static int do_obj_cmp_fun (int (*fun)(SLang_Object_Type *, SLang_Object_Type *))
+{
+   int eqs;
+   SLang_Object_Type a, b;
+   
+   if (-1 == SLang_pop (&b))
+     return -1;
+   
+   if (-1 == SLang_pop (&a))
+     {
+	SLang_free_object (&b);
+	return -1;
+     }
+   
+   eqs = (*fun) (&a, &b);
+
+   SLang_free_object (&a);
+   SLang_free_object (&b);
+   return eqs;
+}
+
+static int is_same_intrinsic (void)
+{
+   return do_obj_cmp_fun (_pSLclass_is_same_obj);
+}
+
+static int eqs_intrinsic (void)
+{
+   return do_obj_cmp_fun (_pSLclass_obj_eqs);
+}
+
+
+static int is_callable_intrinsic (void)
+{
+   SLang_Ref_Type *ref;
+   int ret;
+
+   if (SLang_peek_at_stack () != SLANG_REF_TYPE)
+     {
+	(void) SLdo_pop ();
+	return 0;
+     }
+
+   if (-1 == SLang_pop_ref (&ref))
+     return -1;
+   
+   ret = _pSLang_ref_is_callable (ref);
+   SLang_free_ref (ref);
+
+   return ret;
+}
+
+static int is_numeric (SLtype type)
+{
+   /* Version 2: Add attributes to the class tables to simplify this.  
+    * Also clarify exactly what _pSLang_is_arith_type is supposed to return.
+    */
+   if (0 != _pSLang_is_arith_type ((SLtype) type))
+     return 1;
+   if (type == SLANG_COMPLEX_TYPE)
+     return 1;
+
+   return 0;
+}
+
+static int is_numeric_intrinsic (void)
+{
+   int type;
+   
+   if (-1 == (type = SLang_peek_at_stack1 ()))
+     return -1;
+
+   (void) SLdo_pop ();
+   return is_numeric ((SLtype) type);
+}
+
+static int is_datatype_numeric_intrinsic (void)
+{
+   SLtype type;
+
+   if (-1 == SLang_pop_datatype (&type))
+     return -1;
+   
+   return is_numeric (type);
+}
+
+static void lang_print_stack (void)
+{
+   (void) _pSLang_dump_stack ();
+}
+
+static int atoi_intrin (char *s)
+{
+   return atoi (s);
+}
+static long atol_intrin (char *s)
+{
+   return atol (s);
+}
+#ifdef HAVE_LONG_LONG
+static void atoll_intrin (char *s)
+{
+# ifdef HAVE_ATOLL
+   (void) SLang_push_long_long (atoll(s));
+#else
+# ifdef HAVE_STRTOLL
+   (void) SLang_push_long_long (strtoll(s, NULL, 10));
+# else
+   (void) SLang_push_long_long (strtol(s, NULL, 10));
+# endif
+#endif
+}
+#endif
+
+static void autoload_intrinsic (char *a, char *b)
+{
+   SLang_autoload (a, b);
+}
+static int is_defined_intrin (char *s)
+{
+   return SLang_is_defined (s);
+}
+static int system_intrinsic (char *s)
+{
+   return SLsystem (s);
+}
+static int stack_depth_intrin (void)
+{
+   return SLstack_depth ();
+}
+
+static void set_argv_intrinsic (void);
 static SLang_Intrin_Fun_Type SLang_Basic_Table [] = /*{{{*/
 {
-   MAKE_INTRINSIC_1("__is_initialized", _SLang_is_ref_initialized, SLANG_INT_TYPE, SLANG_REF_TYPE),
+   MAKE_INTRINSIC_0("__is_callable", is_callable_intrinsic, SLANG_INT_TYPE),
+   MAKE_INTRINSIC_0("__is_numeric", is_numeric_intrinsic, SLANG_INT_TYPE),
+   MAKE_INTRINSIC_0("__is_datatype_numeric", is_datatype_numeric_intrinsic, SLANG_INT_TYPE),
+   MAKE_INTRINSIC_1("__is_initialized", _pSLang_is_ref_initialized, SLANG_INT_TYPE, SLANG_REF_TYPE),
    MAKE_INTRINSIC_S("__get_reference", intrin_get_reference, SLANG_VOID_TYPE),
    MAKE_INTRINSIC_1("__uninitialize", uninitialize_ref_intrin, SLANG_VOID_TYPE, SLANG_REF_TYPE),
-   MAKE_INTRINSIC_SS("get_doc_string_from_file",  get_doc_string, SLANG_VOID_TYPE),
-   MAKE_INTRINSIC_SS("autoload",  SLang_autoload, SLANG_VOID_TYPE),
-   MAKE_INTRINSIC_S("is_defined",  SLang_is_defined, SLANG_INT_TYPE),
-   MAKE_INTRINSIC_0("string",  _SLstring_intrinsic, SLANG_VOID_TYPE),
+   MAKE_INTRINSIC_0("__is_same", is_same_intrinsic, SLANG_INT_TYPE),
+   MAKE_INTRINSIC_0("__class_type", class_type_intrinsic, SLANG_INT_TYPE),
+   MAKE_INTRINSIC_0("__class_id", class_id_intrinsic, SLANG_INT_TYPE),
+   MAKE_INTRINSIC_0("_eqs", eqs_intrinsic, SLANG_INT_TYPE),
+   MAKE_INTRINSIC_S("get_doc_string_from_file",  get_doc_string_intrin, SLANG_VOID_TYPE),
+   MAKE_INTRINSIC_S("add_doc_file", add_doc_file_intrin, SLANG_VOID_TYPE),
+   MAKE_INTRINSIC_0("get_doc_files", get_doc_files_intrin, SLANG_VOID_TYPE),
+   MAKE_INTRINSIC_0("set_doc_files", set_doc_files_intrin, SLANG_VOID_TYPE),
+   MAKE_INTRINSIC_SS("autoload",  autoload_intrinsic, SLANG_VOID_TYPE),
+   MAKE_INTRINSIC_S("is_defined",  is_defined_intrin, SLANG_INT_TYPE),
+   MAKE_INTRINSIC_0("string",  _pSLstring_intrinsic, SLANG_VOID_TYPE),
    MAKE_INTRINSIC_0("uname", uname_cmd, SLANG_VOID_TYPE),
    MAKE_INTRINSIC_S("getenv",  intrin_getenv_cmd, SLANG_VOID_TYPE),
 #ifdef HAVE_PUTENV
    MAKE_INTRINSIC_0("putenv",  intrin_putenv, SLANG_VOID_TYPE),
 #endif
-   MAKE_INTRINSIC_S("evalfile",  load_file, SLANG_INT_TYPE),
+   MAKE_INTRINSIC_0("evalfile",  load_file, SLANG_INT_TYPE),
    MAKE_INTRINSIC_I("char",  char_cmd, SLANG_VOID_TYPE),
-   MAKE_INTRINSIC_S("eval",  SLang_load_string, SLANG_VOID_TYPE),
+   MAKE_INTRINSIC_0("eval",  load_string, SLANG_VOID_TYPE),
    MAKE_INTRINSIC_0("dup",  do_dup, SLANG_VOID_TYPE),
    MAKE_INTRINSIC_S("integer",  intrin_integer, SLANG_INT_TYPE),
-   MAKE_INTRINSIC_S("system",  SLsystem, SLANG_INT_TYPE),
+   MAKE_INTRINSIC_S("system",  system_intrinsic, SLANG_INT_TYPE),
    MAKE_INTRINSIC_0("_apropos",  intrin_apropos, SLANG_VOID_TYPE),
-   MAKE_INTRINSIC_S("_trace_function",  _SLang_trace_fun, SLANG_VOID_TYPE),
+   MAKE_INTRINSIC_0("_get_namespaces", intrin_get_namespaces, SLANG_VOID_TYPE),
+   MAKE_INTRINSIC_S("_trace_function",  _pSLang_trace_fun, SLANG_VOID_TYPE),
 #if SLANG_HAS_FLOAT
-   MAKE_INTRINSIC_S("atof", _SLang_atof, SLANG_DOUBLE_TYPE),
+   MAKE_INTRINSIC_S("atof", _pSLang_atof, SLANG_DOUBLE_TYPE),
    MAKE_INTRINSIC_0("double", intrin_double, SLANG_VOID_TYPE),
+#endif
+   MAKE_INTRINSIC_S("atoi", atoi_intrin, SLANG_INT_TYPE),
+   MAKE_INTRINSIC_S("atol", atol_intrin, SLANG_LONG_TYPE),
+#ifdef HAVE_LONG_LONG
+   MAKE_INTRINSIC_S("atoll", atoll_intrin, SLANG_VOID_TYPE),
 #endif
    MAKE_INTRINSIC_0("int",  intrin_int, SLANG_VOID_TYPE),
    MAKE_INTRINSIC_0("typecast", intrin_typecast, SLANG_VOID_TYPE),
-   MAKE_INTRINSIC_0("_stkdepth", _SLstack_depth, SLANG_INT_TYPE),
+   MAKE_INTRINSIC_0("_stkdepth", stack_depth_intrin, SLANG_INT_TYPE),
    MAKE_INTRINSIC_I("_stk_reverse", intrin_reverse_stack, SLANG_VOID_TYPE),
    MAKE_INTRINSIC_0("typeof", intrin_type_info, VOID_TYPE),
    MAKE_INTRINSIC_0("_typeof", intrin_type_info1, VOID_TYPE),
@@ -564,22 +858,23 @@ static SLang_Intrin_Fun_Type SLang_Basic_Table [] = /*{{{*/
    MAKE_INTRINSIC_0("_print_stack", lang_print_stack, SLANG_VOID_TYPE),
    MAKE_INTRINSIC_I("_stk_roll", intrin_roll_stack, SLANG_VOID_TYPE),
    MAKE_INTRINSIC_SI("byte_compile_file", byte_compile_file, SLANG_VOID_TYPE),
-   MAKE_INTRINSIC_0("_clear_error", _SLang_clear_error, SLANG_VOID_TYPE),
+   MAKE_INTRINSIC_0("_clear_error", _pSLerr_clear_error, SLANG_VOID_TYPE),
    MAKE_INTRINSIC_0("_function_name", intrin_function_name, SLANG_STRING_TYPE),
 #if SLANG_HAS_FLOAT
-   MAKE_INTRINSIC_S("set_float_format", _SLset_double_format, SLANG_VOID_TYPE),
+   MAKE_INTRINSIC_S("set_float_format", _pSLset_double_format, SLANG_VOID_TYPE),
 #endif
    MAKE_INTRINSIC_S("_slang_guess_type", guess_type, SLANG_VOID_TYPE),
    MAKE_INTRINSIC_S("error", intrin_error, SLANG_VOID_TYPE),
    MAKE_INTRINSIC_S("message", intrin_message, SLANG_VOID_TYPE),
    MAKE_INTRINSIC_0("__get_defined_symbols", intrin_get_defines, SLANG_INT_TYPE),
-   MAKE_INTRINSIC_I("__pop_args", _SLstruct_pop_args, SLANG_VOID_TYPE),
-   MAKE_INTRINSIC_1("__push_args", _SLstruct_push_args, SLANG_VOID_TYPE, SLANG_ARRAY_TYPE),
+   MAKE_INTRINSIC_I("__pop_args", _pSLstruct_pop_args, SLANG_VOID_TYPE),
+   MAKE_INTRINSIC_1("__push_args", _pSLstruct_push_args, SLANG_VOID_TYPE, SLANG_ARRAY_TYPE),
    MAKE_INTRINSIC_0("usage", usage, SLANG_VOID_TYPE),
-   MAKE_INTRINSIC_S("implements", _SLang_implements_intrinsic, SLANG_VOID_TYPE),
-   MAKE_INTRINSIC_S("use_namespace", _SLang_use_namespace_intrinsic, SLANG_VOID_TYPE),
-   MAKE_INTRINSIC_0("current_namespace", _SLang_cur_namespace_intrinsic, SLANG_STRING_TYPE),
+   MAKE_INTRINSIC_S("implements", _pSLang_implements_intrinsic, SLANG_VOID_TYPE),
+   MAKE_INTRINSIC_S("use_namespace", _pSLang_use_namespace_intrinsic, SLANG_VOID_TYPE),
+   MAKE_INTRINSIC_0("current_namespace", _pSLang_cur_namespace_intrinsic, SLANG_STRING_TYPE),
    MAKE_INTRINSIC_0("length", length_cmd, SLANG_INT_TYPE),
+   MAKE_INTRINSIC_0("__set_argc_argv", set_argv_intrinsic, SLANG_VOID_TYPE),
    SLANG_END_INTRIN_FUN_TABLE
 };
 
@@ -591,16 +886,20 @@ char *SLang_Doc_Dir = SLANG_DOC_DIR;
 char *SLang_Doc_Dir = "";
 #endif
 
+#ifdef SLANG_INSTALL_PREFIX
+static char *Install_Prefix = SLANG_INSTALL_PREFIX;
+#else
+static char *Install_Prefix = "";
+#endif
+
 static SLang_Intrin_Var_Type Intrin_Vars[] =
 {
-   MAKE_VARIABLE("_debug_info", &_SLang_Compile_Line_Num_Info, SLANG_INT_TYPE, 0),
-   MAKE_VARIABLE("_auto_declare", &_SLang_Auto_Declare_Globals, SLANG_INT_TYPE, 0),
-   MAKE_VARIABLE("_traceback", &SLang_Traceback, SLANG_INT_TYPE, 0),
-   MAKE_VARIABLE("_slangtrace", &_SLang_Trace, SLANG_INT_TYPE, 0),
-   MAKE_VARIABLE("_slang_version", &SLang_Version, SLANG_INT_TYPE, 1),
-   MAKE_VARIABLE("_slang_version_string", &SLang_Version_String, SLANG_STRING_TYPE, 1),
-   MAKE_VARIABLE("_NARGS", &SLang_Num_Function_Args, SLANG_INT_TYPE, 1),
-   MAKE_VARIABLE("_slang_doc_dir", &SLang_Doc_Dir, SLANG_STRING_TYPE, 1),
+   MAKE_VARIABLE("_debug_info", &_pSLang_Compile_Line_Num_Info, SLANG_INT_TYPE, 0),
+   MAKE_VARIABLE("_boseos_info", &_pSLang_Compile_BOSEOS, SLANG_INT_TYPE, 0),
+   MAKE_VARIABLE("_auto_declare", &_pSLang_Auto_Declare_Globals, SLANG_INT_TYPE, 0),
+   MAKE_VARIABLE("_slangtrace", &_pSLang_Trace, SLANG_INT_TYPE, 0),
+   MAKE_VARIABLE("_slang_utf8_ok", &_pSLinterp_UTF8_Mode, SLANG_INT_TYPE, 1),
+   MAKE_VARIABLE("_slang_install_prefix", &Install_Prefix, SLANG_STRING_TYPE, 1),
    MAKE_VARIABLE("NULL", NULL, SLANG_NULL_TYPE, 1),
    SLANG_END_INTRIN_VAR_TABLE
 };
@@ -639,22 +938,35 @@ int SLang_init_slang (void) /*{{{*/
 	NULL
      };
 
-   if (-1 == _SLregister_types ()) return -1;
+   if (-1 == _pSLerr_init ())
+     return -1;
+
+   if (-1 == _pSLregister_types ()) return -1;
 
    if ((-1 == SLadd_intrin_fun_table(SLang_Basic_Table, NULL))
        || (-1 == SLadd_intrin_var_table (Intrin_Vars, NULL))
-       || (-1 == _SLang_init_slstrops ())
-       || (-1 == _SLang_init_sltime ())
-       || (-1 == _SLstruct_init ())
-#if SLANG_HAS_COMPLEX
-       || (-1 == _SLinit_slcomplex ())
-#endif
+       || (-1 == _pSLang_init_slstrops ())
+       || (-1 == _pSLang_init_sltime ())
+       || (-1 == _pSLang_init_sllist ())
+       || (-1 == _pSLstruct_init ())
 #if SLANG_HAS_ASSOC_ARRAYS
        || (-1 == SLang_init_slassoc ())
 #endif
+#if SLANG_HAS_BOSEOS
+       || (-1 == _pSLang_init_boseos ())
+#endif
+       || (-1 == _pSLang_init_exceptions ())
        )
      return -1;
 
+   /* More nonsense for the windoze loader */
+   if ((-1 == SLadd_intrinsic_variable ("_NARGS", &SLang_Num_Function_Args, SLANG_INT_TYPE, 1))
+       || (-1 == SLadd_intrinsic_variable ("_traceback", &SLang_Traceback, SLANG_INT_TYPE, 0))
+       || (-1 == SLadd_intrinsic_variable ("_slang_version", &SLang_Version, SLANG_INT_TYPE, 1))
+       || (-1 == SLadd_intrinsic_variable ("_slang_version_string", &SLang_Version_String, SLANG_STRING_TYPE, 1))
+       || (-1 == SLadd_intrinsic_variable ("_slang_doc_dir", &SLang_Doc_Dir, SLANG_STRING_TYPE, 1)))
+     return -1;
+
    SLadd_global_variable (SLANG_SYSTEM_NAME);
 
    s = sys_defines;
@@ -678,7 +990,21 @@ int SLang_init_slang (void) /*{{{*/
    SLang_load_string (".(_NARGS 1 - Sprintf error)verror");
    SLang_load_string (".(_NARGS 1 - Sprintf message)vmessage");
 
-   if (SLang_Error)
+#if SLANG_HAS_SIGNALS
+   if (-1 == SLang_add_interrupt_hook (_pSLang_check_signals_hook, NULL))
+     return -1;
+#endif
+
+   if ((SLang_Doc_Dir != NULL) 
+       && (*SLang_Doc_Dir != 0))
+     {
+	char *docfile = SLpath_dircat (SLang_Doc_Dir, "slangfun.txt");
+	/* NULL ok */
+	(void) add_doc_file (docfile);
+	SLfree (docfile);
+     }
+
+   if (_pSLang_Error)
      return -1;
 
    return 0;
@@ -686,6 +1012,52 @@ int SLang_init_slang (void) /*{{{*/
 
 /*}}}*/
 
+static int This_Argc;
+static SLang_Array_Type *This_Argv = NULL;
+
+static int add_argc_argv (SLang_Array_Type *at)
+{
+   This_Argc = at->num_elements;
+   if (-1 == SLadd_intrinsic_variable ("__argc", (VOID_STAR)&This_Argc,
+				       SLANG_INT_TYPE, 1))
+     return -1;
+
+   if (-1 == SLadd_intrinsic_variable ("__argv", (VOID_STAR)at, SLANG_ARRAY_TYPE, 0))
+     return -1;
+   if (This_Argv != NULL)
+     SLang_free_array (This_Argv);
+   This_Argv = at;
+   return 0;
+}
+
+static void set_argv_intrinsic (void)
+{
+   SLang_Array_Type *at;
+   
+   if (-1 == SLang_pop_array_of_type (&at, SLANG_STRING_TYPE))
+     return;
+
+   if (-1 == add_argc_argv (at))
+     SLang_free_array (at);
+}
+   
+int SLang_set_argc_argv (int argc, char **argv)
+{
+   SLang_Array_Type *at = _pSLstrings_to_array (argv, argc);
+
+   if (at == NULL)
+     return -1;
+   
+   if (-1 == add_argc_argv (at))
+     {
+	SLang_free_array (at);
+	return -1;
+     }
+
+   return 0;
+}
+
+#if 0
 int SLang_set_argc_argv (int argc, char **argv)
 {
    static int this_argc;
@@ -722,3 +1094,4 @@ int SLang_set_argc_argv (int argc, char **argv)
 
    return -1;
 }
+#endif
