@@ -1,26 +1,43 @@
-/* Copyright (c) 1998, 1999, 2001 John E. Davis
- * This file is part of the S-Lang library.
- *
- * You may distribute under the terms of either the GNU General Public
- * License or the Perl Artistic License.
- */
+/*
+Copyright (C) 2004, 2005 John E. Davis
+
+This file is part of the S-Lang Library.
+
+The S-Lang Library is free software; you can redistribute it and/or
+modify it under the terms of the GNU General Public License as
+published by the Free Software Foundation; either version 2 of the
+License, or (at your option) any later version.
+
+The S-Lang Library is distributed in the hope that it will be useful,
+but WITHOUT ANY WARRANTY; without even the implied warranty of
+MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+General Public License for more details.
+
+You should have received a copy of the GNU General Public License
+along with this library; if not, write to the Free Software
+Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
+USA.  
+*/
+
 #include "slinclud.h"
 
 #include "slang.h"
 #include "_slang.h"
 
-typedef struct _SLstring_Type
+typedef struct _pSLstring_Type
 {
-   struct _SLstring_Type *next;
+   struct _pSLstring_Type *next;
    unsigned int ref_count;
    char bytes [1];
 }
 SLstring_Type;
 
+#define MAP_HASH_TO_INDEX(hash) ((hash) % SLSTRING_HASH_TABLE_SIZE)
+
 static SLstring_Type *String_Hash_Table [SLSTRING_HASH_TABLE_SIZE];
 static char Single_Char_Strings [256 * 2];
 
-#if _SLANG_OPTIMIZE_FOR_SPEED
+#if SLANG_OPTIMIZE_FOR_SPEED
 #define MAX_FREE_STORE_LEN 32
 static SLstring_Type *SLS_Free_Store [MAX_FREE_STORE_LEN];
 
@@ -28,21 +45,25 @@ static SLstring_Type *SLS_Free_Store [MAX_FREE_STORE_LEN];
 typedef struct 
 {
    unsigned long hash;
+   char *str;
    SLstring_Type *sls;
    unsigned int len;
 }
 Cached_String_Type;
+static char *Deleted_String = "*deleted*";
 static Cached_String_Type Cached_Strings [NUM_CACHED_STRINGS];
 
 #define GET_CACHED_STRING(s) \
    (Cached_Strings + (unsigned int)(((unsigned long) (s)) % NUM_CACHED_STRINGS))
 
+
 _INLINE_
 static void cache_string (SLstring_Type *sls, unsigned int len, unsigned long hash)
 {
    Cached_String_Type *cs;
    
    cs = GET_CACHED_STRING(sls->bytes);
+   cs->str = sls->bytes;
    cs->sls = sls;
    cs->hash = hash;
    cs->len = len;
@@ -54,16 +75,82 @@ static void uncache_string (char *s)
    Cached_String_Type *cs;
    
    cs = GET_CACHED_STRING(s);
-   if ((cs->sls != NULL)
-       && (cs->sls->bytes == s))
-     cs->sls = NULL;
+   if (cs->str == s)
+     {
+	cs->sls = NULL;
+	cs->str = Deleted_String;
+     }
 }
 #endif
 
+#if USE_NEW_HASH_CODE
+/* This hash algorithm comes from:
+ *
+ *   Bob Jenkins, 1996.  bob_jenkins@burtleburtle.net. 
+ *   You may use this code any way you wish, private, educational, or commercial.  It's free.
+ *   See http://burtleburtle.net/bob/hash/evahash.html
+ */
 
+typedef  unsigned long uint32;
+
+#define mix(a,b,c) \
+{ \
+  a -= b; a -= c; a ^= (c>>13); \
+  b -= c; b -= a; b ^= (a<<8); \
+  c -= a; c -= b; c ^= (b>>13); \
+  a -= b; a -= c; a ^= (c>>12);  \
+  b -= c; b -= a; b ^= (a<<16); \
+  c -= a; c -= b; c ^= (b>>5); \
+  a -= b; a -= c; a ^= (c>>3);  \
+  b -= c; b -= a; b ^= (a<<10); \
+  c -= a; c -= b; c ^= (b>>15); \
+}
 
 _INLINE_
-unsigned long _SLstring_hash (unsigned char *s, unsigned char *smax)
+unsigned long _pSLstring_hash (unsigned char *s, unsigned char *smax)
+{
+   register uint32 a,b,c;
+   unsigned int length = (unsigned int)(smax - s);
+   unsigned int len = length;
+   
+   a = b = 0x9e3779b9;  /* the golden ratio; an arbitrary value */
+   c = 0;
+
+   /*---------------------------------------- handle most of the key */
+   while (len >= 12)
+     {
+	a += (s[0] +((uint32)s[1]<<8) +((uint32)s[2]<<16) +((uint32)s[3]<<24));
+	b += (s[4] +((uint32)s[5]<<8) +((uint32)s[6]<<16) +((uint32)s[7]<<24));
+	c += (s[8] +((uint32)s[9]<<8) +((uint32)s[10]<<16)+((uint32)s[11]<<24));
+	mix(a,b,c);
+	s += 12; len -= 12;
+     }
+
+   /*------------------------------------- handle the last 11 bytes */
+   c += length;
+   switch(len)              /* all the case statements fall through */
+   {
+   case 11: c+=((uint32)s[10]<<24);
+   case 10: c+=((uint32)s[9]<<16);
+   case 9 : c+=((uint32)s[8]<<8);
+      /* the first byte of c is reserved for the length */
+   case 8 : b+=((uint32)s[7]<<24);
+   case 7 : b+=((uint32)s[6]<<16);
+   case 6 : b+=((uint32)s[5]<<8);
+   case 5 : b+=s[4];
+   case 4 : a+=((uint32)s[3]<<24);
+   case 3 : a+=((uint32)s[2]<<16);
+   case 2 : a+=((uint32)s[1]<<8);
+   case 1 : a+=s[0];
+     /* case 0: nothing left to add */
+   }
+   mix(a,b,c);
+   /*-------------------------------------------- report the result */
+   return (unsigned long) c;
+}
+#else
+_INLINE_
+unsigned long _pSLstring_hash (unsigned char *s, unsigned char *smax)
 {
    register unsigned long h = 0;
    register unsigned long sum = 0;
@@ -93,19 +180,17 @@ unsigned long _SLstring_hash (unsigned char *s, unsigned char *smax)
 
    return h;
 }
-
-unsigned long _SLcompute_string_hash (char *s)
+#endif
+unsigned long _pSLcompute_string_hash (char *s)
 {
-#if _SLANG_OPTIMIZE_FOR_SPEED
+#if SLANG_OPTIMIZE_FOR_SPEED
    Cached_String_Type *cs;
-   SLstring_Type *sls;
 
    cs = GET_CACHED_STRING(s);
-   if (((sls = cs->sls) != NULL)
-       && (sls->bytes == s))
+   if (cs->str == s)
      return cs->hash;
 #endif
-   return _SLstring_hash ((unsigned char *) s, (unsigned char *) s + strlen (s));
+   return _pSLstring_hash ((unsigned char *) s, (unsigned char *) s + strlen (s));
 }
 
 _INLINE_
@@ -115,7 +200,7 @@ static SLstring_Type *find_string (char *s, unsigned int len, unsigned long hash
    SLstring_Type *sls;
    char ch;
 
-   sls = String_Hash_Table [(unsigned int)(hash % SLSTRING_HASH_TABLE_SIZE)];
+   sls = String_Hash_Table [(unsigned int) MAP_HASH_TO_INDEX(hash)];
 
    if (sls == NULL)
      return NULL;
@@ -130,9 +215,9 @@ static SLstring_Type *find_string (char *s, unsigned int len, unsigned long hash
 	 * this is called from create_nstring, etc...  It is unlikely that the
 	 * pointer is a slstring
 	 */
-	if ((/* (s == bytes) || */ ((ch == bytes[0])
-		 && (0 == strncmp (s, bytes, len))))
-	    && (bytes [len] == 0))
+	if (((ch == bytes[0])
+	     && (0 == strncmp (s, bytes, len))
+	     && (bytes [len] == 0)))
 	  break;
 
 	sls = sls->next;
@@ -147,7 +232,7 @@ static SLstring_Type *find_slstring (char *s, unsigned long hash)
 {
    SLstring_Type *sls;
 
-   sls = String_Hash_Table [(unsigned int)(hash % SLSTRING_HASH_TABLE_SIZE)];
+   sls = String_Hash_Table [MAP_HASH_TO_INDEX(hash)];
    while (sls != NULL)
      {
 	if (s == sls->bytes)
@@ -161,9 +246,9 @@ static SLstring_Type *find_slstring (char *s, unsigned long hash)
 _INLINE_
 static SLstring_Type *allocate_sls (unsigned int len)
 {
+#if SLANG_OPTIMIZE_FOR_SPEED
    SLstring_Type *sls;
    
-#if _SLANG_OPTIMIZE_FOR_SPEED
    if ((len < MAX_FREE_STORE_LEN)
        && (NULL != (sls = SLS_Free_Store [len])))
      {
@@ -178,7 +263,7 @@ static SLstring_Type *allocate_sls (unsigned int len)
 _INLINE_
 static void free_sls (SLstring_Type *sls, unsigned int len)
 {
-#if _SLANG_OPTIMIZE_FOR_SPEED
+#if SLANG_OPTIMIZE_FOR_SPEED
    if ((len < MAX_FREE_STORE_LEN)
        && (SLS_Free_Store[len] == NULL))
      {
@@ -203,7 +288,7 @@ static char *create_long_string (char *s, unsigned int len, unsigned long hash)
 	sls->ref_count++;
 	s = sls->bytes;
 
-#if _SLANG_OPTIMIZE_FOR_SPEED
+#if SLANG_OPTIMIZE_FOR_SPEED
 	cache_string (sls, len, hash); 
 #endif
 	return s;
@@ -217,11 +302,11 @@ static char *create_long_string (char *s, unsigned int len, unsigned long hash)
    sls->bytes[len] = 0;
    sls->ref_count = 1;
 
-#if _SLANG_OPTIMIZE_FOR_SPEED
+#if SLANG_OPTIMIZE_FOR_SPEED
    cache_string (sls, len, hash);
 #endif
 
-   hash = hash % SLSTRING_HASH_TABLE_SIZE;
+   hash = MAP_HASH_TO_INDEX(hash);
    sls->next = String_Hash_Table [(unsigned int)hash];
    String_Hash_Table [(unsigned int)hash] = sls;
 
@@ -246,32 +331,34 @@ static char *create_short_string (char *s, unsigned int len)
 
 /* s cannot be NULL */
 _INLINE_
-static char *create_nstring (char *s, unsigned int len, unsigned long *hash_ptr)
+static SLstr_Type *create_nstring (char *s, unsigned int len, unsigned long *hash_ptr)
 {
    unsigned long hash;
 
    if (len < 2)
      return create_short_string (s, len);
 
-   hash = _SLstring_hash ((unsigned char *) s, (unsigned char *) (s + len));
+   hash = _pSLstring_hash ((unsigned char *) s, (unsigned char *) (s + len));
    *hash_ptr = hash;
 
    return create_long_string (s, len, hash);
 }
 
-char *SLang_create_nslstring (char *s, unsigned int len)
+SLstr_Type *SLang_create_nslstring (char *s, unsigned int len)
 {
    unsigned long hash;
+   if (s == NULL)
+     return NULL;
    return create_nstring (s, len, &hash);
 }
 
-char *_SLstring_make_hashed_string (char *s, unsigned int len, unsigned long *hashptr)
+char *_pSLstring_make_hashed_string (char *s, unsigned int len, unsigned long *hashptr)
 {
    unsigned long hash;
 
    if (s == NULL) return NULL;
 
-   hash = _SLstring_hash ((unsigned char *) s, (unsigned char *) s + len);
+   hash = _pSLstring_hash ((unsigned char *) s, (unsigned char *) s + len);
    *hashptr = hash;
 
    if (len < 2)
@@ -280,12 +367,11 @@ char *_SLstring_make_hashed_string (char *s, unsigned int len, unsigned long *ha
    return create_long_string (s, len, hash);
 }
 
-char *_SLstring_dup_hashed_string (char *s, unsigned long hash)
+char *_pSLstring_dup_hashed_string (char *s, unsigned long hash)
 {
    unsigned int len;
-#if _SLANG_OPTIMIZE_FOR_SPEED
+#if SLANG_OPTIMIZE_FOR_SPEED
    Cached_String_Type *cs;
-   SLstring_Type *sls;
    
    if (s == NULL) return NULL;
    if (s[0] == 0)
@@ -294,10 +380,9 @@ char *_SLstring_dup_hashed_string (char *s, unsigned long hash)
      return create_short_string (s, 1);
      
    cs = GET_CACHED_STRING(s);
-   if (((sls = cs->sls) != NULL)
-       && (sls->bytes == s))
+   if (cs->str == s)
      {
-	sls->ref_count += 1;
+	cs->sls->ref_count += 1;
 	return s;
      }
 #else
@@ -305,59 +390,57 @@ char *_SLstring_dup_hashed_string (char *s, unsigned long hash)
 #endif
 
    len = strlen (s);
-#if !_SLANG_OPTIMIZE_FOR_SPEED
+#if !SLANG_OPTIMIZE_FOR_SPEED
    if (len < 2) return create_short_string (s, len);
 #endif
 
    return create_long_string (s, len, hash);
 }
 
-char *_SLstring_dup_slstring (char *s)
+char *_pSLstring_dup_slstring (char *s)
 {
    SLstring_Type *sls;
    unsigned int len;
    unsigned long hash;
-#if _SLANG_OPTIMIZE_FOR_SPEED
+#if SLANG_OPTIMIZE_FOR_SPEED
    Cached_String_Type *cs;
+#endif
    
+   if (s == NULL)
+     return s;
+#if SLANG_OPTIMIZE_FOR_SPEED
    cs = GET_CACHED_STRING(s);
-   if (((sls = cs->sls) != NULL)
-       && (sls->bytes == s))
+   if (cs->str == s)
      {
-	sls->ref_count += 1;
+	cs->sls->ref_count += 1;
 	return s;
      }
 #endif
    
-   if ((s == NULL) || ((len = strlen (s)) < 2))
+   if ((len = strlen (s)) < 2)
      return s;
 
-   hash = _SLstring_hash ((unsigned char *)s, (unsigned char *)(s + len));
+   hash = _pSLstring_hash ((unsigned char *)s, (unsigned char *)(s + len));
 
    sls = find_slstring (s, hash);
    if (sls == NULL)
      {
-	SLang_Error = SL_INTERNAL_ERROR;
+	SLang_set_error (SL_INTERNAL_ERROR);
 	return NULL;
      }
    
    sls->ref_count++;
-#if _SLANG_OPTIMIZE_FOR_SPEED
+#if SLANG_OPTIMIZE_FOR_SPEED
    cache_string (sls, len, hash);
 #endif
    return s;
 }
 
-static void free_sls_string (SLstring_Type *sls, char *s, unsigned int len, 
-			     unsigned long hash)
+static void free_sls_string (SLstring_Type *sls, unsigned int len, unsigned long hash)
 {
    SLstring_Type *sls1, *prev;
 
-#if _SLANG_OPTIMIZE_FOR_SPEED
-   uncache_string (s);
-#endif
-
-   hash = hash % SLSTRING_HASH_TABLE_SIZE;
+   hash = MAP_HASH_TO_INDEX(hash);
 
    sls1 = String_Hash_Table [(unsigned int) hash];
 
@@ -385,21 +468,22 @@ static void free_long_string (char *s, unsigned int len, unsigned long hash)
 
    if (NULL == (sls = find_slstring (s, hash)))
      {
-	SLang_doerror ("Application internal error: invalid attempt to free string");
+	SLang_verror (SL_APPLICATION_ERROR, "invalid attempt to free string:%s", s);
 	return;
      }
 
    sls->ref_count--;
    if (sls->ref_count != 0)
      {
-#if _SLANG_OPTIMIZE_FOR_SPEED
+#if SLANG_OPTIMIZE_FOR_SPEED
 	/* cache_string (sls, len, hash); */
 #endif
 	return;
      }
-   
-
-   free_sls_string (sls, s, len, hash);
+#if SLANG_OPTIMIZE_FOR_SPEED
+   uncache_string (s);
+#endif
+   free_sls_string (sls, len, hash);
 }
 
 /* This routine may be passed NULL-- it is not an error. */
@@ -407,59 +491,66 @@ void SLang_free_slstring (char *s)
 {
    unsigned long hash;
    unsigned int len;
-#if _SLANG_OPTIMIZE_FOR_SPEED
+#if SLANG_OPTIMIZE_FOR_SPEED
    Cached_String_Type *cs;
-   SLstring_Type *sls;
+#endif
 
+   if (s == NULL) return;
+
+#if SLANG_OPTIMIZE_FOR_SPEED
    cs = GET_CACHED_STRING(s);
-   if (((sls = cs->sls) != NULL)
-       && (sls->bytes == s))
+   if (cs->str == s)
      {
+	SLstring_Type *sls = cs->sls;
 	if (sls->ref_count <= 1)
-	  free_sls_string (sls, s, cs->len, cs->hash);
+	  {
+#if SLANG_OPTIMIZE_FOR_SPEED
+	     cs->sls = NULL;
+	     cs->str = Deleted_String;
+#endif
+	     free_sls_string (sls, cs->len, cs->hash);
+	  }
 	else
 	  sls->ref_count -= 1;
 	return;
      }
 #endif
 
-   if (s == NULL) return;
-
    if ((len = strlen (s)) < 2)
      return;
 
-   hash = _SLstring_hash ((unsigned char *)s, (unsigned char *) s + len);
+   hash = _pSLstring_hash ((unsigned char *)s, (unsigned char *) s + len);
    free_long_string (s, len, hash);
 }
 
 char *SLang_create_slstring (char *s)
 {
    unsigned long hash;
-#if _SLANG_OPTIMIZE_FOR_SPEED
+#if SLANG_OPTIMIZE_FOR_SPEED
    Cached_String_Type *cs;
-   SLstring_Type *sls;
+#endif
 
+   if (s == NULL) return NULL;
+#if SLANG_OPTIMIZE_FOR_SPEED
    cs = GET_CACHED_STRING(s);
-   if (((sls = cs->sls) != NULL)
-       && (sls->bytes == s))
+   if (cs->str == s)
      {
-	sls->ref_count += 1;
+	cs->sls->ref_count += 1;
 	return s;
      }
 #endif
 
-   if (s == NULL) return NULL;
    return create_nstring (s, strlen (s), &hash);
 }
 
-void _SLfree_hashed_string (char *s, unsigned int len, unsigned long hash)
+void _pSLfree_hashed_string (char *s, unsigned int len, unsigned long hash)
 {
    if ((s == NULL) || (len < 2)) return;
    free_long_string (s, len, hash);
 }
 
 
-char *_SLallocate_slstring (unsigned int len)
+char *_pSLallocate_slstring (unsigned int len)
 {
    SLstring_Type *sls = allocate_sls (len);
    if (sls == NULL)
@@ -468,7 +559,7 @@ char *_SLallocate_slstring (unsigned int len)
    return sls->bytes;
 }
 
-void _SLunallocate_slstring (char *s, unsigned int len)
+void _pSLunallocate_slstring (char *s, unsigned int len)
 {
    SLstring_Type *sls;
    
@@ -479,7 +570,7 @@ void _SLunallocate_slstring (char *s, unsigned int len)
    free_sls (sls, len);
 }
 
-char *_SLcreate_via_alloced_slstring (char *s, unsigned int len)
+char *_pSLcreate_via_alloced_slstring (char *s, unsigned int len)
 {   
    unsigned long hash;
    SLstring_Type *sls;
@@ -490,22 +581,22 @@ char *_SLcreate_via_alloced_slstring (char *s, unsigned int len)
    if (len < 2)
      {
 	char *s1 = create_short_string (s, len);
-	_SLunallocate_slstring (s, len);
+	_pSLunallocate_slstring (s, len);
 	return s1;
      }
 
    /* s is not going to be in the cache because when it was malloced, its
     * value was unknown.  This simplifies the coding.
     */
-   hash = _SLstring_hash ((unsigned char *)s, (unsigned char *)s + len);
+   hash = _pSLstring_hash ((unsigned char *)s, (unsigned char *)s + len);
    sls = find_string (s, len, hash);
    if (sls != NULL)
      {
 	sls->ref_count++;
-	_SLunallocate_slstring (s, len);
+	_pSLunallocate_slstring (s, len);
 	s = sls->bytes;
 
-#if _SLANG_OPTIMIZE_FOR_SPEED
+#if SLANG_OPTIMIZE_FOR_SPEED
 	cache_string (sls, len, hash); 
 #endif
 	return s;
@@ -514,11 +605,11 @@ char *_SLcreate_via_alloced_slstring (char *s, unsigned int len)
    sls = (SLstring_Type *) (s - offsetof(SLstring_Type,bytes[0]));
    sls->ref_count = 1;
 
-#if _SLANG_OPTIMIZE_FOR_SPEED
+#if SLANG_OPTIMIZE_FOR_SPEED
    cache_string (sls, len, hash);
 #endif
 
-   hash = hash % SLSTRING_HASH_TABLE_SIZE;
+   hash = MAP_HASH_TO_INDEX(hash);
    sls->next = String_Hash_Table [(unsigned int)hash];
    String_Hash_Table [(unsigned int)hash] = sls;
 
@@ -534,13 +625,18 @@ char *SLang_concat_slstrings (char *a, char *b)
    lena = strlen (a);
    len = lena + strlen (b);
 
-   c = _SLallocate_slstring (len);
+   c = _pSLallocate_slstring (len);
    if (c == NULL)
      return NULL;
 
    strcpy (c, a);
    strcpy (c + lena, b);
 
-   return _SLcreate_via_alloced_slstring (c, len);
+   return _pSLcreate_via_alloced_slstring (c, len);
+}
+
+unsigned int _pSLstring_bytelen (SLstr_Type *s)
+{
+   return strlen (s);
 }
 
