@@ -1,7 +1,7 @@
 /* -*- mode: C; mode: fold; -*- */
 /* string manipulation functions for S-Lang. */
 /*
-Copyright (C) 2004, 2005, 2006, 2007 John E. Davis
+Copyright (C) 2004, 2005 John E. Davis
 
 This file is part of the S-Lang Library.
 
@@ -102,7 +102,7 @@ static unsigned int do_trim (SLuchar_Type **beg, int do_beg,
    else white = NULL;
 
    a = *beg;
-   len = _pSLstring_bytelen ((char *)a);
+   len = strlen ((char *)a);
    b = a + len;
 
    if (do_beg)
@@ -156,6 +156,15 @@ static void free_3_strings (char *a, char *b, char *c)
    SLfree (c);
 }
 
+static int SLang_pop_wchar (SLwchar_Type *wc)
+{
+   unsigned long w;
+   if (-1 == SLang_pop_ulong (&w))
+     return -1;
+   *wc = (SLwchar_Type)w;
+   return 0;
+}
+
 static void strcat_cmd (void) /*{{{*/
 {
    char *c, *c1;
@@ -186,7 +195,7 @@ static void strcat_cmd (void) /*{{{*/
 	if (-1 == SLang_pop_slstring (&s))
 	  goto free_and_return;
 	ptrs[i] = s;
-	len += _pSLstring_bytelen (s);
+	len += strlen (s);
      }
    if (NULL == (c = _pSLallocate_slstring (len)))
      goto free_and_return;
@@ -195,12 +204,12 @@ static void strcat_cmd (void) /*{{{*/
    for (i = 0; i < nargs; i++)
      {
 	strcpy (c1, ptrs[i]);
-	c1 += _pSLstring_bytelen (ptrs[i]);   /* c1 does not take advantage of the slstring cache */
+	c1 += strlen (c1);
      }
-
+   
    free_and_return:
    for (i = 0; i < nargs; i++)
-     _pSLang_free_slstring (ptrs[i]);
+     SLang_free_slstring (ptrs[i]);
    if (ptrs != ptrs_buf)
      SLfree ((char *) ptrs);
 
@@ -244,7 +253,7 @@ static void strtrim_cmd_internal (char *str, int do_beg, int do_end)
    
    (void) _pSLang_push_nstring ((char *)beg, len);
    if (free_str)
-     _pSLang_free_slstring (str);
+     SLang_free_slstring (str);
 }
 
    
@@ -515,7 +524,7 @@ static void strtok_cmd (char *str)
      goto the_return;
 
    s = (SLuchar_Type *)str;
-   smax = s + _pSLstring_bytelen ((SLstr_Type *)s);
+   smax = s + strlen ((char *)s);
    while (s < smax)
      {
 	SLuchar_Type *s0;
@@ -538,7 +547,7 @@ static void strtok_cmd (char *str)
 	  }
 	if (-1 == _pSLstring_list_append (&sl, new_s))
 	  {
-	     _pSLang_free_slstring (new_s);
+	     SLang_free_slstring (new_s);
 	     _pSLstring_list_delete (&sl);
 	     goto the_return;
 	  }
@@ -550,7 +559,7 @@ static void strtok_cmd (char *str)
    the_return:
    if (white != NULL)
      {
-	_pSLang_free_slstring (str);
+	SLang_free_slstring (str);
 	SLwchar_free_lut (lut);
      }
 }
@@ -773,16 +782,18 @@ static int issubstr_cmd (char *a, char *b) /*{{{*/
 
 /*}}}*/
 
-static void subbytes_cmd (SLstr_Type *a, int *n_ptr, int *m_ptr) /*{{{*/
+static void subbytes_cmd (char *a, int *n_ptr, int *m_ptr) /*{{{*/
 {
    int m;
    unsigned int n;
    unsigned int lena;
+   char *amax;
 
    n = (unsigned int) (*n_ptr - 1);
    m = *m_ptr;
 
-   lena = _pSLstring_bytelen (a);
+   lena = strlen (a);
+   amax = a + lena;
 
    if (n > lena) 
      n = lena;
@@ -795,12 +806,12 @@ static void subbytes_cmd (SLstr_Type *a, int *n_ptr, int *m_ptr) /*{{{*/
 
 /*}}}*/
 
-static void substr_cmd (SLstr_Type *a, int *n_ptr, int *m_ptr) /*{{{*/
+static void substr_cmd (char *a, int *n_ptr, int *m_ptr) /*{{{*/
 {
    int n, m;
    int lena;
    int ignore_combining;
-   char *a1;
+   char *a1, *amax;
    
    if (_pSLinterp_UTF8_Mode == 0)
      {
@@ -813,6 +824,9 @@ static void substr_cmd (SLstr_Type *a, int *n_ptr, int *m_ptr) /*{{{*/
    n = *n_ptr;
    m = *m_ptr;
 
+   lena = strlen (a);
+   amax = a + lena;
+
    lena = SLutf8_strlen ((SLuchar_Type *)a, ignore_combining);
 	
    if (n > lena) n = lena + 1;
@@ -825,8 +839,7 @@ static void substr_cmd (SLstr_Type *a, int *n_ptr, int *m_ptr) /*{{{*/
    n--;
    if (m < 0) m = lena;
    if (n + m > lena) m = lena - n;
-
-   /* FIXME: Are the strlens necessary here?  */
+   
    a = (char *)SLutf8_skip_chars ((SLuchar_Type *)a, (SLuchar_Type*)a + strlen(a),
 				  (unsigned int)n, NULL, ignore_combining);
 
@@ -956,7 +969,7 @@ static int isdigit_cmd (void) /*{{{*/
 	  }
 	else wch = s[0];
 
-	_pSLang_free_slstring (s);
+	SLang_free_slstring (s);
      }
    else if (-1 == SLang_pop_wchar (&wch))
      return -1;
@@ -1180,24 +1193,20 @@ static int strcmp_cmd (char *a, char *b) /*{{{*/
 
 /*}}}*/
 
-static int do_strncmp_cmd (SLstr_Type *a, SLstr_Type *b, int n, int skip_combining) /*{{{*/
+static int do_strncmp_cmd (char *a, char *b, int n, int skip_combining) /*{{{*/
 {
    char *p;
    unsigned int na, nb;
-   unsigned int lena, lenb;
    int cmp;
 
    if (_pSLinterp_UTF8_Mode == 0)
      return strncmp(a, b, (unsigned int) n);
    
-   lena = _pSLstring_bytelen (a);
-   lenb = _pSLstring_bytelen (b);
-
-   p = (char *)SLutf8_skip_chars ((SLuchar_Type *)a, (SLuchar_Type *)(a + lena),
+   p = (char *)SLutf8_skip_chars ((SLuchar_Type *)a, (SLuchar_Type *)(a + strlen (a)),
                           (unsigned int) n, NULL, skip_combining);
    na = (unsigned int) (p - a);
 
-   p = (char *)SLutf8_skip_chars ((SLuchar_Type *)b, (SLuchar_Type *)(b + lenb),
+   p = (char *)SLutf8_skip_chars ((SLuchar_Type *)b, (SLuchar_Type *)(b + strlen (b)),
                           (unsigned int) n, NULL, skip_combining);
    nb = (unsigned int) (p - b);
    
@@ -1222,58 +1231,60 @@ static int do_strncmp_cmd (SLstr_Type *a, SLstr_Type *b, int n, int skip_combini
 
 /*}}}*/
 
-static int strncmp_cmd (SLstr_Type *a, SLstr_Type *b, int *n)
+static int strncmp_cmd (char *a, char *b, int *n)
 {
    return do_strncmp_cmd (a, b, *n, 1);
 }
 
-static int strnbytecmp_cmd (SLstr_Type *a, SLstr_Type *b, int *n)
+static int strnbytecmp_cmd (char *a, char *b, int *n)
 {
    return (int) strncmp (a, b, *n);
 }
 
-static int strncharcmp_cmd (SLstr_Type *a, SLstr_Type *b, int *n)
+static int strncharcmp_cmd (char *a, char *b, int *n)
 {
    return do_strncmp_cmd (a, b, *n, 0);
 }
 
-static int do_strlen_combining (SLstr_Type *s, int ignore_combining) /*{{{*/
+static int do_strlen_combining (char *s, int ignore_combining) /*{{{*/
 {
    if (_pSLinterp_UTF8_Mode == 0)
-     return (int) _pSLstring_bytelen (s);
+     return (int) strlen (s);
    
    return (int) SLutf8_strlen ((SLuchar_Type *)s, ignore_combining);
 }
 /*}}}*/
 
 
-static int strlen_cmd (SLstr_Type *s) /*{{{*/
+static int strlen_cmd (char *s) /*{{{*/
 {
    return do_strlen_combining (s, 1);
 }
 
-static int strcharlen_cmd (SLstr_Type *s)
+static int strcharlen_cmd (char *s)
 {
    return do_strlen_combining (s, 0);
 }
 
-static int strbytelen_cmd (SLstr_Type *s)
+static int strbytelen_cmd (char *s)
 {
-   return (int) _pSLstring_bytelen (s);
+   return (int) strlen (s);
 }
 
 /*}}}*/
 
 
+/* sprintf functionality for S-Lang */
+
 static char *SLdo_sprintf (char *fmt) /*{{{*/
 {
    register char *p = fmt, ch;
    char *out = NULL, *outp = NULL;
    char dfmt[1024];	       /* used to hold part of format */
    char *f;
-   char *str;
+   VOID_STAR varp;
    unsigned int want_width, width, precis;
-   int int_var, use_string;
+   int use_varp, int_var;
    long long_var;
 #ifdef HAVE_LONG_LONG
    long long llong_var;
@@ -1285,7 +1296,6 @@ static char *SLdo_sprintf (char *fmt) /*{{{*/
    int use_double;
    double x;
 #endif
-   unsigned char uch;
    int use_long = 0;
 
    while (1)
@@ -1404,13 +1414,13 @@ static char *SLdo_sprintf (char *fmt) /*{{{*/
 #ifdef HAVE_LONG_LONG
 	llong_var = 0;
 #endif
-	str = NULL;
+	varp = NULL;
 	guess_size = 32;
 #if SLANG_HAS_FLOAT
 	use_double = 0;
 #endif
 	use_long = 0;
-	use_string = 0;
+	use_varp = 0;
 	do_free = 0;
 
 	if (ch == 'l')
@@ -1433,52 +1443,36 @@ static char *SLdo_sprintf (char *fmt) /*{{{*/
 	     ch = 's';
 	     /* drop */
 	   case 's':
-	     if (-1 == SLang_pop_slstring(&str))
-	       return (out);
+	     if (SLang_pop_slstring((char **) &varp)) return (out);
 	     do_free = 1;
-	     guess_size = strlen(str);
-	     use_string = 1;
+	     guess_size = strlen((char *) varp);
+	     use_varp = 1;
 	     break;
 
 	   case '%':
 	     guess_size = 1;
 	     do_free = 0;
-	     use_string = 1;
-	     str = "%";
+	     use_varp = 1;
+	     varp = (VOID_STAR) "%";
 	     break;
 
 	   case 'c':
-#if 0
 	     if (use_long)
-#endif
 	       {
 		  SLwchar_Type wc;
 		  SLuchar_Type utf8_buf[SLUTF8_MAX_MBLEN+1];
 
-		  if (-1 == SLang_pop_wchar (&wc))
-		    return out;
-		  if ((_pSLinterp_UTF8_Mode == 0) && (wc <= 0xFF))
-		    {
-		       utf8_buf[0] = (unsigned char)wc;
-		       utf8_buf[1] = 0;
-		    }
-		  else if (NULL == SLutf8_encode_null_terminate (wc, utf8_buf))
+		  if ((-1 == SLang_pop_wchar (&wc))
+		      || (NULL == (_pSLinterp_encode_wchar (wc, utf8_buf, &guess_size))))
 		    return out;
 		  ch = 's';
-		  str = (char *)utf8_buf;
-		  use_string = 1;
+		  varp = (char *)utf8_buf;
+		  use_varp = 1;
+		  break;
 	       }
-	     break;
-	     
-	   case 'b':
-	     use_long = 0;
 	     guess_size = 1;
-	     if (-1 == SLang_pop_uchar (&uch))
-	       return out;
-	     int_var = (int) uch;
-	     ch = 'c';
-	     break;
-
+	     use_long = 0;
+	     /* drop */
 	   case 'd':
 	   case 'i':
 	   case 'o':
@@ -1521,8 +1515,8 @@ static char *SLdo_sprintf (char *fmt) /*{{{*/
 	     /* Pointer type?? Why?? */
 	     if (-1 == SLdo_pop ())
 	       return out;
-	     str = (char *) _pSLang_get_run_stack_pointer ();
-	     use_string = 1;
+	     varp = (VOID_STAR) _pSLang_get_run_stack_pointer ();
+	     use_varp = 1;
 	     use_long = 0;
 	     break;
 
@@ -1548,10 +1542,10 @@ static char *SLdo_sprintf (char *fmt) /*{{{*/
 	     malloc_len = len + guess_size;
 	  }
 
-	if (use_string)
+	if (use_varp)
 	  {
-	     sprintf(outp, dfmt, str);
-	     if (do_free) _pSLang_free_slstring (str);
+	     sprintf(outp, dfmt, varp);
+	     if (do_free) SLang_free_slstring ((char *)varp);
 	  }
 #if SLANG_HAS_FLOAT
 	else if (use_double) sprintf(outp, dfmt, x);
@@ -1599,7 +1593,7 @@ int _pSLstrops_do_sprintf_n (int n) /*{{{*/
      return -1;
 
    p = SLdo_sprintf (fmt);
-   _pSLang_free_slstring (fmt);
+   SLang_free_slstring (fmt);
 
    SLdo_pop_n (_pSLang_get_run_stack_pointer () - ptr);
 
@@ -1906,7 +1900,7 @@ static void create_delimited_string_cmd (int *nptr)
    str = create_delimited_string (strings + 1, (n - 1), strings[0]);
    /* drop */
    return_error:
-   for (i = 0; i < n; i++) _pSLang_free_slstring (strings[i]);
+   for (i = 0; i < n; i++) SLang_free_slstring (strings[i]);
    SLfree ((char *)strings);
 
    (void) SLang_push_malloced_string (str);   /* NULL Ok */
@@ -2018,25 +2012,14 @@ static void glob_to_regexp (char *glob)
    char *pat, *p;
    char ch;
 
-   len = _pSLstring_bytelen (glob);
-   pat = SLmalloc (2*len + 8);
+   len = strlen (glob);
+   pat = SLmalloc (2*len + 3);
    if (pat == NULL)
      return;
-
+   
    p = pat;
    *p++ = '^';
 
-   /* If the first character of a file is '.', it must be explicitly matched. */
-   if ((*glob == '?') || (*glob == '*'))
-     {
-	*p++ = '[';
-	*p++ = '^';
-	*p++ = '.';
-	*p++ = ']';
-	if (*glob == '?')
-	  glob++;
-     }
-
    while (0 != (ch = *glob++))
      {
 	if ((ch == '.') || (ch == '$') || (ch == '+') || (ch == '\\'))
@@ -2062,33 +2045,25 @@ static void glob_to_regexp (char *glob)
 	     continue;
 	  }
 	
-	/* Otherwise ch = '[' */
 	if (*glob != 0)
 	  {
 	     char *g = glob;
-	     int is_complement = 0;
-
+	     
 	     ch = *g;
-	     if ((ch == '^') || (ch == '!'))
+	     if (ch == '^')
 	       {
-		  is_complement = 1;
 		  g++;
 		  ch = *g;
 	       }
 	     if (ch == ']')
 	       g++;
-
+	     
 	     while (((ch = *g) != 0) && (ch != ']'))
 	       g++;
 	     
 	     if (ch == ']')
 	       {
 		  *p++ = '[';
-		  if (is_complement)
-		    {
-		       *p++ = '^';
-		       glob++;
-		    }
 		  while (glob <= g)
 		    *p++ = *glob++;
 		  
