@@ -1,6 +1,6 @@
 /* Structure type implementation */
 /*
-Copyright (C) 2004, 2005, 2006, 2007 John E. Davis
+Copyright (C) 2004, 2005 John E. Davis
 
 This file is part of the S-Lang Library.
 
@@ -27,7 +27,7 @@ USA.
 #include "_slang.h"
 
 
-void SLang_free_struct (_pSLang_Struct_Type *s)
+void _pSLstruct_delete_struct (_pSLang_Struct_Type *s)
 {
    _pSLstruct_Field_Type *field, *field_max;
 
@@ -117,15 +117,12 @@ static int push_struct_of_type (SLtype type, _pSLang_Struct_Type *s)
    return -1;
 }
 
-int SLang_push_struct (_pSLang_Struct_Type *s)
+int _pSLang_push_struct (_pSLang_Struct_Type *s)
 {
-   if (s == NULL)
-     return SLang_push_null ();
-
    return push_struct_of_type (SLANG_STRUCT_TYPE, s);
 }
 
-int SLang_pop_struct (_pSLang_Struct_Type **sp)
+int _pSLang_pop_struct (_pSLang_Struct_Type **sp)
 {
    SLang_Object_Type obj;
    SLang_Class_Type *cl;
@@ -156,7 +153,7 @@ int SLang_pop_struct (_pSLang_Struct_Type **sp)
 static void struct_destroy (SLtype type, VOID_STAR vs)
 {
    (void) type;
-   SLang_free_struct (*(_pSLang_Struct_Type **) vs);
+   _pSLstruct_delete_struct (*(_pSLang_Struct_Type **) vs);
 }
 
 static int struct_push (SLtype type, VOID_STAR ptr)
@@ -183,17 +180,6 @@ static _pSLstruct_Field_Type *find_field (_pSLang_Struct_Type *s, char *name)
    return NULL;
 }
 
-/* This function is used by the qualifier-code */
-SLang_Object_Type *_pSLstruct_get_field_value (SLang_Struct_Type *s, char *name)
-{
-   _pSLstruct_Field_Type *f = find_field (s, name);
-
-   if (f == NULL)
-     return NULL;
-
-   return &f->obj;
-}
-
 static _pSLstruct_Field_Type *pop_field (_pSLang_Struct_Type *s, char *name,
 					_pSLstruct_Field_Type *(*find)(_pSLang_Struct_Type *, char *))
 {
@@ -258,7 +244,7 @@ static _pSLang_Struct_Type *
    return s;
 
    return_error:
-   SLang_free_struct (s);
+   _pSLstruct_delete_struct (s);
    return NULL;
 }
 
@@ -272,10 +258,10 @@ int SLstruct_create_struct (unsigned int nfields,
    if (NULL == (s = create_struct (nfields, field_names, field_types, field_values)))
      return -1;
 
-   if (0 == SLang_push_struct (s))
+   if (0 == _pSLang_push_struct (s))
      return 0;
 
-   SLang_free_struct (s);
+   _pSLstruct_delete_struct (s);
    return -1;
 }
 
@@ -306,7 +292,7 @@ static _pSLang_Struct_Type *struct_from_struct_fields (int nfields)
 	nfields--;
 	if (-1 == SLang_pop_slstring (&name))
 	  {
-	     SLang_free_struct (s);
+	     _pSLstruct_delete_struct (s);
 	     return NULL;
 	  }
 
@@ -320,7 +306,7 @@ static _pSLang_Struct_Type *struct_from_struct_fields (int nfields)
 	     SLang_verror (SL_DuplicateDefinition_Error,
 			   "Field %s used more than once in the struct",
 			   name);
-	     SLang_free_struct (s);
+	     _pSLstruct_delete_struct (s);
 	     return NULL;
 	  }
      }
@@ -339,69 +325,14 @@ int _pSLstruct_define_struct (void)
    if (NULL == (s = struct_from_struct_fields (nfields)))
      return -1;
 
-   if (-1 == SLang_push_struct (s))
+   if (-1 == _pSLang_push_struct (s))
      {
-	SLang_free_struct (s);
+	_pSLstruct_delete_struct (s);
 	return -1;
      }
    return 0;
 }
 
-static int pop_to_struct_field (_pSLang_Struct_Type *s, char *name)
-{
-   _pSLstruct_Field_Type *f;
-   SLang_Object_Type obj;
-   
-   if ((NULL == (f = pop_field (s, name, find_field)))
-       || (-1 == SLang_pop (&obj)))
-     return -1;
-
-   SLang_free_object (&f->obj);
-   f->obj = obj;
-   
-   return 0;
-}
-
-/* This function is used for structure definitions with embedded assignments */
-int _pSLstruct_define_struct2 (void)
-{
-   _pSLang_Struct_Type *s;
-   int nfields;
-   int nassigns;
-
-   if (-1 == SLang_pop_integer (&nassigns))
-     return -1;
-
-   if (-1 == SLang_pop_integer (&nfields))
-     return -1;
-
-   if (NULL == (s = struct_from_struct_fields (nfields)))
-     return -1;
-
-   while (nassigns > 0)
-     {
-	char *name;
-	if (-1 == SLang_pop_slstring (&name))
-	  goto return_error;
-	
-	if (-1 == pop_to_struct_field (s, name))
-	  {
-	     SLang_free_slstring (name);
-	     goto return_error;
-	  }
-	SLang_free_slstring (name);
-	nassigns--;
-     }
-
-   if (0 == SLang_push_struct (s))
-     return 0;
-   
-   return_error:
-   
-   SLang_free_struct (s);
-   return -1;
-}
-
 static int init_struct_with_user_methods (SLtype, _pSLang_Struct_Type *);
 /* Simply make a struct that contains the same fields as struct s.  Do not
  * duplicate the field values.
@@ -423,7 +354,7 @@ static _pSLang_Struct_Type *make_struct_shell (_pSLang_Struct_Type *s, SLtype ty
      {
 	if (NULL == (new_f[i].name = SLang_create_slstring (old_f[i].name)))
 	  {
-	     SLang_free_struct (new_s);
+	     _pSLstruct_delete_struct (new_s);
 	     return NULL;
 	  }
      }
@@ -459,7 +390,7 @@ typedefed_struct_datatype_deref (SLtype type)
 
    if (-1 == push_struct_of_type (type, s))
      {
-	SLang_free_struct (s);
+	_pSLstruct_delete_struct (s);
 	return -1;
      }
 
@@ -490,7 +421,7 @@ static _pSLang_Struct_Type *duplicate_struct (_pSLang_Struct_Type *s, SLtype typ
 	     if ((-1 == _pSLpush_slang_obj (obj))
 		 || (-1 == SLang_pop (&new_f->obj)))
 	       {
-		  SLang_free_struct (new_s);
+		  _pSLstruct_delete_struct (new_s);
 		  return NULL;
 	       }
 	  }
@@ -510,7 +441,7 @@ static int struct_dereference (SLtype type, VOID_STAR addr)
 
    if (-1 == push_struct_of_type (type, s))
      {
-	SLang_free_struct (s);
+	_pSLstruct_delete_struct (s);
 	return -1;
      }
 
@@ -534,7 +465,7 @@ struct_foreach_open (SLtype type, unsigned int num)
 
    (void) type;
 
-   if (-1 == SLang_pop_struct (&s))
+   if (-1 == _pSLang_pop_struct (&s))
      return NULL;
 
    switch (num)
@@ -558,14 +489,14 @@ struct_foreach_open (SLtype type, unsigned int num)
 
    if (next_name == NULL)
      {
-	SLang_free_struct (s);
+	_pSLstruct_delete_struct (s);
 	return NULL;
      }
 
    c = (SLang_Foreach_Context_Type *)SLmalloc (sizeof (SLang_Foreach_Context_Type));
    if (c == NULL)
      {
-	SLang_free_struct (s);
+	_pSLstruct_delete_struct (s);
 	SLang_free_slstring (next_name);
 	return NULL;
      }
@@ -583,7 +514,7 @@ static void struct_foreach_close (SLtype type, SLang_Foreach_Context_Type *c)
    if (c == NULL) return;
 
    SLang_free_slstring (c->next_field_name);
-   if (c->s != NULL) SLang_free_struct (c->s);
+   if (c->s != NULL) _pSLstruct_delete_struct (c->s);
    SLfree ((char *) c);
 }
 
@@ -600,7 +531,7 @@ static int struct_foreach (SLtype type, SLang_Foreach_Context_Type *c)
    if (c->s == NULL)
      return 0;			       /* done */
 
-   if (-1 == SLang_push_struct (c->s))
+   if (-1 == _pSLang_push_struct (c->s))
      return -1;
 
    /* Now get the next one ready for the next foreach loop */
@@ -622,7 +553,7 @@ static int struct_foreach (SLtype type, SLang_Foreach_Context_Type *c)
 	  }
      }
 
-   SLang_free_struct (c->s);
+   _pSLstruct_delete_struct (c->s);
    c->s = next_s;
 
    /* keep going */
@@ -968,9 +899,13 @@ static int this_binary_this (int op,
 
 static int register_unary_ops (Struct_Info_Type *si, SLtype t)
 {
+   SLang_Class_Type *cl;
+
    if (si->unary_registered)
      return 0;
    
+   cl = _pSLclass_get_class (t);
+
    if (-1 == SLclass_add_unary_op (t, struct_unary, struct_unary_result))
      return -1;
 
@@ -980,8 +915,12 @@ static int register_unary_ops (Struct_Info_Type *si, SLtype t)
 
 static int register_binary_ops (Struct_Info_Type *si, SLtype t)
 {
+   SLang_Class_Type *cl;
+
    if (si->binary_registered)
      return 0;
+   
+   cl = _pSLclass_get_class (t);
 
    if ((-1 == SLclass_add_binary_op (t, SLANG_VOID_TYPE,
 				     this_binary_any, this_binary_any_result))
@@ -1199,7 +1138,7 @@ static void add_destroy_method (void)
 	return;
      }
    
-   if (-1 == SLang_pop_struct (&s))
+   if (-1 == _pSLang_pop_struct (&s))
      {
 	SLang_free_function (f);
 	return;
@@ -1208,7 +1147,7 @@ static void add_destroy_method (void)
    if (s->destroy_method != NULL)
      SLang_free_function (s->destroy_method);
    s->destroy_method = SLang_copy_function (f);
-   SLang_free_struct (s);
+   _pSLstruct_delete_struct (s);
 }
 
 static void add_string_method (SLtype *typep, SLang_Ref_Type *ref)
@@ -1244,18 +1183,24 @@ static int init_struct_with_user_methods (SLtype type, _pSLang_Struct_Type *s)
 static int struct_sput (SLtype type, char *name)
 {
    _pSLang_Struct_Type *s;
+   _pSLstruct_Field_Type *f;
+   SLang_Object_Type obj;
 
    (void) type;
 
-   if (-1 == SLang_pop_struct (&s))
+   if (-1 == _pSLang_pop_struct (&s))
      return -1;
 
-   if (-1 == pop_to_struct_field (s, name))
+   if ((NULL == (f = pop_field (s, name, find_field)))
+       || (-1 == SLang_pop (&obj)))
      {
-	SLang_free_struct (s);
+	_pSLstruct_delete_struct (s);
 	return -1;
      }
-   SLang_free_struct (s);
+
+   SLang_free_object (&f->obj);
+   f->obj = obj;
+   _pSLstruct_delete_struct (s);
    return 0;
 }
 
@@ -1267,17 +1212,17 @@ static int struct_sget (SLtype type, char *name)
 
    (void) type;
 
-   if (-1 == SLang_pop_struct (&s))
+   if (-1 == _pSLang_pop_struct (&s))
      return -1;
 
    if (NULL == (f = pop_field (s, name, find_field)))
      {
-	SLang_free_struct (s);
+	_pSLstruct_delete_struct (s);
 	return -1;
      }
 
    ret = _pSLpush_slang_obj (&f->obj);
-   SLang_free_struct (s);
+   _pSLstruct_delete_struct (s);
    return ret;
 }
 
@@ -1322,10 +1267,10 @@ static char *string_method (SLtype type, VOID_STAR p)
 	return SLmake_string (buf);
      }
    
-   if ((-1 == SLang_start_arg_list ())
-       || (-1 == SLang_push_struct (s))
-       || (-1 == SLang_end_arg_list ())
-       || (-1 == SLexecute_function (f)))
+   if (-1 == _pSLang_push_struct (s))
+     return NULL;
+   
+   if (-1 == SLexecute_function (f))
      return NULL;
 
    if (-1 == SLpop_string (&str))
@@ -1385,7 +1330,7 @@ int _pSLstruct_define_typedef (void)
    if (-1 == SLang_pop_slstring (&type_name))
      return -1;
 
-   if (-1 == SLang_pop_struct (&s))
+   if (-1 == _pSLang_pop_struct (&s))
      {
 	SLang_free_slstring (type_name);
 	return -1;
@@ -1394,16 +1339,16 @@ int _pSLstruct_define_typedef (void)
    if (NULL == (s1 = make_struct_shell (s, SLANG_STRUCT_TYPE)))
      {
 	SLang_free_slstring (type_name);
-	SLang_free_struct (s);
+	_pSLstruct_delete_struct (s);
 	return -1;
      }
 
-   SLang_free_struct (s);
+   _pSLstruct_delete_struct (s);
 
    if (NULL == (cl = SLclass_allocate_class (type_name)))
      {
 	SLang_free_slstring (type_name);
-	SLang_free_struct (s1);
+	_pSLstruct_delete_struct (s1);
 	return -1;
      }
    SLang_free_slstring (type_name);
@@ -1565,7 +1510,7 @@ static void struct_set_field (void)
 	return;
      }
 
-   if (-1 == SLang_pop_struct (&s))
+   if (-1 == _pSLang_pop_struct (&s))
      {
 	SLang_free_slstring (name);
 	SLang_free_object (&obj);
@@ -1574,7 +1519,7 @@ static void struct_set_field (void)
 
    if (NULL == (f = pop_field (s, name, find_field)))
      {
-	SLang_free_struct (s);
+	_pSLstruct_delete_struct (s);
 	SLang_free_slstring (name);
 	SLang_free_object (&obj);
 	return;
@@ -1583,7 +1528,7 @@ static void struct_set_field (void)
    SLang_free_object (&f->obj);
    f->obj = obj;
 
-   SLang_free_struct (s);
+   _pSLstruct_delete_struct (s);
    SLang_free_slstring (name);
 }
 
@@ -1600,7 +1545,7 @@ static void set_struct_fields (void)
      return;
 
    n--;
-   if (-1 == SLang_pop_struct (&s))
+   if (-1 == _pSLang_pop_struct (&s))
      {
 	SLdo_pop_n (n);
 	return;
@@ -1610,7 +1555,7 @@ static void set_struct_fields (void)
      {
 	SLdo_pop_n (n);
 	SLang_verror (SL_INVALID_PARM, "Too many values for structure");
-	SLang_free_struct (s);
+	_pSLstruct_delete_struct (s);
 	return;
      }
 
@@ -1629,7 +1574,7 @@ static void set_struct_fields (void)
 	n--;
      }
 
-   SLang_free_struct (s);
+   _pSLstruct_delete_struct (s);
 }
 
 static void get_struct_field (char *name)
@@ -1763,7 +1708,7 @@ void _pSLstruct_pop_args (int *np)
 
 	s = data[i];
 	if (s != NULL)
-	  SLang_free_struct (s);
+	  _pSLstruct_delete_struct (s);
      }
 
    SLfree ((char *) data);
@@ -1848,7 +1793,7 @@ int SLang_pop_cstruct (VOID_STAR cs, SLang_CStruct_Field_Type *cfields)
    if ((cfields == NULL) || (cs == NULL))
      return -1;
 
-   if (-1 == SLang_pop_struct (&s))
+   if (-1 == _pSLang_pop_struct (&s))
      return -1;
 
    cfield = cfields;
@@ -1869,7 +1814,7 @@ int SLang_pop_cstruct (VOID_STAR cs, SLang_CStruct_Field_Type *cfields)
 	cfield++;
      }
 
-   SLang_free_struct (s);
+   _pSLstruct_delete_struct (s);
    return 0;
 
    return_error:
@@ -1878,7 +1823,7 @@ int SLang_pop_cstruct (VOID_STAR cs, SLang_CStruct_Field_Type *cfields)
 	free_cstruct_field (cfield, cs);
 	cfield--;
      }
-   SLang_free_struct (s);
+   _pSLstruct_delete_struct (s);
    return -1;
 }
 
@@ -1938,10 +1883,10 @@ int SLang_push_cstruct (VOID_STAR cs, SLang_CStruct_Field_Type *cfields)
    if (NULL == (s = create_cstruct (cs, cfields)))
      return -1;
    
-   if (0 == SLang_push_struct (s))
+   if (0 == _pSLang_push_struct (s))
      return 0;
    
-   SLang_free_struct (s);
+   _pSLstruct_delete_struct (s);
    return -1;
 }
 
@@ -1955,73 +1900,23 @@ int SLang_assign_cstruct_to_ref (SLang_Ref_Type *ref, VOID_STAR cs, SLang_CStruc
    if (0 == SLang_assign_to_ref (ref, SLANG_STRUCT_TYPE, (VOID_STAR) &s))
      return 0;
 
-   SLang_free_struct (s);
+   _pSLstruct_delete_struct (s);
    return -1;
 }
 
-/* Struct Field Reference */
-typedef struct
+/* Public functions */
+extern void SLang_free_struct (SLang_Struct_Type *s)
 {
-   SLang_Struct_Type *s;
-   char *field_name;
-}
-Struct_Field_Ref_Type;
-
-static int struct_field_deref_assign (VOID_STAR vdata)
-{
-   Struct_Field_Ref_Type *data = (Struct_Field_Ref_Type *)vdata;
-   return pop_to_struct_field (data->s, data->field_name);
+   _pSLstruct_delete_struct (s);
 }
 
-static int struct_field_deref (VOID_STAR vdata)
+extern int SLang_push_struct (SLang_Struct_Type *s)
 {
-   Struct_Field_Ref_Type *frt = (Struct_Field_Ref_Type *)vdata;
-   _pSLstruct_Field_Type *f;
-
-   if (NULL == (f = pop_field (frt->s, frt->field_name, find_field)))
-     return -1;
-
-   return _pSLpush_slang_obj (&f->obj);
+   return _pSLang_push_struct (s);
 }
 
-static void struct_field_ref_destroy (VOID_STAR vdata)
+extern int SLang_pop_struct (SLang_Struct_Type **s)
 {
-   Struct_Field_Ref_Type *frt = (Struct_Field_Ref_Type *)vdata;
-
-   SLang_free_slstring (frt->field_name);
-   SLang_free_struct (frt->s);
-}
-
-/* Stack: struct */
-int _pSLstruct_push_field_ref (char *name)
-{
-   SLang_Struct_Type *s;
-   Struct_Field_Ref_Type *frt;
-   SLang_Ref_Type *ref;
-   int ret;
-
-   if (-1 == SLang_pop_struct (&s))
-     return -1;
-
-   if (NULL == (name = SLang_create_slstring (name)))
-     {
-	SLang_free_struct (s);
-	return -1;
-     }
-   if (NULL == (ref = _pSLang_new_ref (sizeof (Struct_Field_Ref_Type))))
-     {
-	SLang_free_struct (s);
-	SLang_free_slstring (name);
-     }
-   frt = (Struct_Field_Ref_Type *) ref->data;
-   frt->s = s;
-   frt->field_name = name;
-   ref->deref = struct_field_deref;
-   ref->deref_assign = struct_field_deref_assign;
-   ref->destroy = struct_field_ref_destroy;
-
-   ret = SLang_push_ref (ref);
-   SLang_free_ref (ref);
-   return ret;
+   return _pSLang_pop_struct (s);
 }
 
