@@ -1,20 +1,37 @@
 /* time related system calls */
-/* Copyright (c) 1992, 1999, 2001 John E. Davis
- * This file is part of the S-Lang library.
- *
- * You may distribute under the terms of either the GNU General Public
- * License or the Perl Artistic License.
- */
+/*
+Copyright (C) 2004, 2005 John E. Davis
+
+This file is part of the S-Lang Library.
+
+The S-Lang Library is free software; you can redistribute it and/or
+modify it under the terms of the GNU General Public License as
+published by the Free Software Foundation; either version 2 of the
+License, or (at your option) any later version.
+
+The S-Lang Library is distributed in the hope that it will be useful,
+but WITHOUT ANY WARRANTY; without even the implied warranty of
+MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+General Public License for more details.
+
+You should have received a copy of the GNU General Public License
+along with this library; if not, write to the Free Software
+Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
+USA.  
+*/
 
 #include "slinclud.h"
 
 #include <sys/types.h>
 #include <time.h>
+#ifdef HAVE_SYS_TIME_H
+# include <sys/time.h>
+#endif
 
 #if defined(__BORLANDC__)
 # include <dos.h>
 #endif
-#if defined(__GO32__) || defined(__WATCOMC__)
+#if defined(__GO32__) || (defined(__WATCOMC__) && !defined(__QNX__))
 # include <dos.h>
 # include <bios.h>
 #endif
@@ -37,10 +54,12 @@
 # endif
 #endif
 
+/* Turns on/off intrinsics that _may_ get added in the future */
+#define USE_FROM_FUTURE		0
 
 #if defined(IBMPC_SYSTEM)
-/* For other system (Unix and VMS), _SLusleep is in sldisply.c */
-int _SLusleep (unsigned long s)
+/* For other system (Unix and VMS), _pSLusleep is in sldisply.c */
+int _pSLusleep (unsigned long s)
 {
    sleep (s/1000000L);
    s = s % 1000000L;
@@ -69,7 +88,7 @@ unsigned int sleep (unsigned int seconds)
 }
 #endif
 
-static char *ctime_cmd (unsigned long *tt)
+static char *ctime_cmd (long *tt)
 {
    char *t;
 
@@ -85,7 +104,7 @@ static void sleep_cmd (void)
    unsigned long usecs;
    double x;
 
-   if (-1 == SLang_pop_double (&x, NULL, NULL))
+   if (-1 == SLang_pop_double (&x))
      return;
 
    if (x < 0.0) 
@@ -94,7 +113,7 @@ static void sleep_cmd (void)
    sleep (secs);
    x -= (double) secs;
    usecs = (unsigned long) (1e6 * x);
-   if (usecs > 0) _SLusleep (usecs);
+   if (usecs > 0) _pSLusleep (usecs);
 #else
    if (-1 == SLang_pop_uinteger (&secs))
      return;
@@ -102,9 +121,9 @@ static void sleep_cmd (void)
 #endif
 }
 
-static unsigned long _time_cmd (void)
+static long _time_cmd (void)
 {
-   return (unsigned long) time (NULL);
+   return (long) time (NULL);
 }
 
 #if defined(__GO32__)
@@ -171,37 +190,25 @@ char *SLcurrent_time_string (void) /*{{{*/
 
 /*}}}*/
 
+static SLang_CStruct_Field_Type TM_Struct [] =
+{
+   MAKE_CSTRUCT_INT_FIELD(struct tm, tm_sec, "tm_sec", 0),
+   MAKE_CSTRUCT_INT_FIELD(struct tm, tm_min, "tm_min", 0),
+   MAKE_CSTRUCT_INT_FIELD(struct tm, tm_hour, "tm_hour", 0),
+   MAKE_CSTRUCT_INT_FIELD(struct tm, tm_mday, "tm_mday", 0),
+   MAKE_CSTRUCT_INT_FIELD(struct tm, tm_mon, "tm_mon", 0),
+   MAKE_CSTRUCT_INT_FIELD(struct tm, tm_year, "tm_year", 0),
+   MAKE_CSTRUCT_INT_FIELD(struct tm, tm_wday, "tm_wday", 0),
+   MAKE_CSTRUCT_INT_FIELD(struct tm, tm_yday, "tm_yday", 0),
+   MAKE_CSTRUCT_INT_FIELD(struct tm, tm_isdst, "tm_isdst", 0),
+   SLANG_END_CSTRUCT_TABLE
+};
+
 static int push_tm_struct (struct tm *tms)
 {
-   char *field_names [9];
-   unsigned char field_types[9];
-   VOID_STAR field_values [9];
-   int int_values [9];
-   unsigned int i;
-
-   if (tms == NULL)
-     return SLang_push_null ();
-
-   field_names [0] = "tm_sec"; int_values [0] = tms->tm_sec;
-   field_names [1] = "tm_min"; int_values [1] = tms->tm_min;
-   field_names [2] = "tm_hour"; int_values [2] = tms->tm_hour;
-   field_names [3] = "tm_mday"; int_values [3] = tms->tm_mday;
-   field_names [4] = "tm_mon"; int_values [4] = tms->tm_mon;
-   field_names [5] = "tm_year"; int_values [5] = tms->tm_year;
-   field_names [6] = "tm_wday"; int_values [6] = tms->tm_wday;
-   field_names [7] = "tm_yday"; int_values [7] = tms->tm_yday;
-   field_names [8] = "tm_isdst"; int_values [8] = tms->tm_isdst;
-
-   for (i = 0; i < 9; i++)
-     {
-	field_types [i] = SLANG_INT_TYPE;
-	field_values [i] = (VOID_STAR) (int_values + i);
-     }
-
-   return SLstruct_create_struct (9, field_names, field_types, field_values);
+   return SLang_push_cstruct ((VOID_STAR) tms, TM_Struct);
 }
 
-
 static void localtime_cmd (long *t)
 {
    time_t tt = (time_t) *t;
@@ -218,6 +225,18 @@ static void gmtime_cmd (long *t)
 #endif
 }
 
+#ifdef HAVE_MKTIME
+static long mktime_cmd (void)
+{
+   struct tm t;
+
+   if (-1 == SLang_pop_cstruct (&t, TM_Struct))
+     return (long)-1;
+   
+   return (long) mktime (&t);
+}
+#endif
+
 #ifdef HAVE_TIMES
 
 # ifdef HAVE_SYS_TIMES_H
@@ -236,44 +255,52 @@ static void gmtime_cmd (long *t)
 # endif
 #endif
 
+typedef struct
+{ 
+   double tms_utime;
+   double tms_stime;
+   double tms_cutime;
+   double tms_cstime;
+}
+TMS_Type;
+
+static SLang_CStruct_Field_Type TMS_Struct [] =
+{
+   MAKE_CSTRUCT_FIELD(TMS_Type, tms_utime, "tms_utime", SLANG_DOUBLE_TYPE, 0),
+   MAKE_CSTRUCT_FIELD(TMS_Type, tms_utime, "tms_stime", SLANG_DOUBLE_TYPE, 0),
+   MAKE_CSTRUCT_FIELD(TMS_Type, tms_utime, "tms_cutime", SLANG_DOUBLE_TYPE, 0),
+   MAKE_CSTRUCT_FIELD(TMS_Type, tms_utime, "tms_cstime", SLANG_DOUBLE_TYPE, 0),
+   SLANG_END_CSTRUCT_TABLE
+};
+
 static void times_cmd (void)
 {
-   double dvals[4];
+   TMS_Type d;
    struct tms t;
-   VOID_STAR field_values[4];
-   char *field_names[4];
-   unsigned int i;
-   unsigned char field_types[4];
-   
+
    (void) times (&t);
 
-   field_names[0] = "tms_utime";   dvals[0] = (double)t.tms_utime; 
-   field_names[1] = "tms_stime";   dvals[1] = (double)t.tms_stime; 
-   field_names[2] = "tms_cutime";  dvals[2] = (double)t.tms_cutime;
-   field_names[3] = "tms_cstime";  dvals[3] = (double)t.tms_cstime;
-
-   for (i = 0; i < 4; i++)
-     {
-	dvals[i] *= SECS_PER_TICK;
-	field_values[i] = (VOID_STAR) &dvals[i];
-	field_types[i] = SLANG_DOUBLE_TYPE;
-     }
-   (void) SLstruct_create_struct (4, field_names, field_types, field_values);
+   d.tms_utime = SECS_PER_TICK * (double)t.tms_utime; 
+   d.tms_stime = SECS_PER_TICK * (double)t.tms_stime;
+   d.tms_cutime = SECS_PER_TICK * (double)t.tms_cutime;
+   d.tms_cstime = SECS_PER_TICK * (double)t.tms_cstime;
+   (void) SLang_push_cstruct ((VOID_STAR)&d, TMS_Struct);
 }
 
 static struct tms Tic_TMS;
 
-static void tic_cmd (void)
+static void _tic_cmd (void)
 {
-   (void) times (&Tic_TMS);
+   times (&Tic_TMS);
 }
 
-static double toc_cmd (void)
+static double _toc_cmd (void)
 {
    struct tms t;
    double d;
 
    (void) times (&t);
+   
    d = ((t.tms_utime - Tic_TMS.tms_utime)
 	+ (t.tms_stime - Tic_TMS.tms_stime)) * SECS_PER_TICK;
    Tic_TMS = t;
@@ -282,25 +309,92 @@ static double toc_cmd (void)
 
 #endif				       /* HAVE_TIMES */
 
+#ifdef __WIN32__
+static LARGE_INTEGER Start_Time, Frequency;
+#else
+# ifdef HAVE_GETTIMEOFDAY
+static struct timeval Start_Time;
+# else
+static time_t Start_Time;
+# endif
+#endif
+
+static void tic_cmd (void)
+{
+#ifdef __WIN32__
+   QueryPerformanceCounter(&Start_Time);
+   QueryPerformanceFrequency(&Frequency);
+#else
+# ifdef HAVE_GETTIMEOFDAY
+   (void) gettimeofday (&Start_Time, NULL);
+# else
+   Start_Time = time (NULL);
+# endif
+#endif
+}
+
+static double toc_cmd (void)
+{
+#ifdef __WIN32__
+   LARGE_INTEGER t;
+
+   QueryPerformanceCounter(&t);
+   return (t.QuadPart-Start_Time.QuadPart)/(double)Frequency.QuadPart;
+#else
+# ifdef HAVE_GETTIMEOFDAY
+   struct timeval tv;
+   (void) gettimeofday (&tv, NULL);
+   return (double) tv.tv_sec - (double) Start_Time.tv_sec
+     + ((double) tv.tv_usec - (double) Start_Time.tv_usec) * 1e-6;
+# else
+   return (double) time(NULL) - Start_Time;
+# endif
+#endif
+}
+     
+#ifdef SLANG_HAS_FLOAT
+# if USE_FROM_FUTURE
+static double ftime_cmd (void)
+{
+#  ifdef HAVE_GETTIMEOFDAY
+   struct timeval tv;
+   (void) gettimeofday (&tv, NULL);
+   return (double) tv.tv_sec + (double) tv.tv_usec*1e-6;
+#  else
+   return (double) _time_cmd ();
+#  endif
+}
+# endif
+#endif
 
 static SLang_Intrin_Fun_Type Time_Funs_Table [] =
 {
-   MAKE_INTRINSIC_1("ctime", ctime_cmd, SLANG_STRING_TYPE, SLANG_ULONG_TYPE),
+   MAKE_INTRINSIC_1("ctime", ctime_cmd, SLANG_STRING_TYPE, SLANG_LONG_TYPE),
    MAKE_INTRINSIC_0("sleep", sleep_cmd, SLANG_VOID_TYPE),
-   MAKE_INTRINSIC_0("_time", _time_cmd, SLANG_ULONG_TYPE),
+   MAKE_INTRINSIC_0("_time", _time_cmd, SLANG_LONG_TYPE),
    MAKE_INTRINSIC_0("time", SLcurrent_time_string, SLANG_STRING_TYPE),
    MAKE_INTRINSIC_1("localtime", localtime_cmd, SLANG_VOID_TYPE, SLANG_LONG_TYPE),
    MAKE_INTRINSIC_1("gmtime", gmtime_cmd, SLANG_VOID_TYPE, SLANG_LONG_TYPE),
+#ifdef HAVE_MKTIME
+   MAKE_INTRINSIC_0("mktime", mktime_cmd, SLANG_LONG_TYPE),
+#endif
 
 #ifdef HAVE_TIMES
    MAKE_INTRINSIC_0("times", times_cmd, SLANG_VOID_TYPE),
+   MAKE_INTRINSIC_0("_tic", _tic_cmd, SLANG_VOID_TYPE),
+   MAKE_INTRINSIC_0("_toc", _toc_cmd, SLANG_DOUBLE_TYPE),
+#endif
    MAKE_INTRINSIC_0("tic", tic_cmd, SLANG_VOID_TYPE),
    MAKE_INTRINSIC_0("toc", toc_cmd, SLANG_DOUBLE_TYPE),
+#if SLANG_HAS_FLOAT
+# if USE_FROM_FUTURE
+   MAKE_INTRINSIC_0("ftime", ftime_cmd, SLANG_DOUBLE_TYPE),
+# endif
 #endif
    SLANG_END_INTRIN_FUN_TABLE
 };
 
-int _SLang_init_sltime (void)
+int _pSLang_init_sltime (void)
 {
 #ifdef HAVE_TIMES
    (void) tic_cmd ();
